package org.artifactory.ph;

import com.jfrog.ph.acl.exception.PackageVirtualTraversalException;
import com.jfrog.ph.acl.listener.PackageArtifactEventListener;
import com.jfrog.ph.acl.listener.PackageConfigChangedEventListener;
import com.jfrog.ph.acl.listener.PackageConfigChangedEventType;
import com.jfrog.ph.acl.listener.PackageRepoConfigListenerContext;
import com.jfrog.ph.acl.model.PackageArtifact;
import com.jfrog.ph.acl.model.PackageAuthenticationRequestContext;
import com.jfrog.ph.acl.service.PackageAnalyticsService;
import com.jfrog.ph.acl.service.PackageArtifactService;
import com.jfrog.ph.acl.service.PackageArtifactoryConfigService;
import com.jfrog.ph.acl.service.PackageAuthenticationAdapter;
import com.jfrog.ph.acl.service.PackageEventService;
import com.jfrog.ph.acl.service.PackageNugetCache;
import com.jfrog.ph.acl.service.PackageOAuthHandler;
import com.jfrog.ph.acl.service.PackageRepoIndexCalculatorService;
import com.jfrog.ph.acl.service.PackageRepoService;
import com.jfrog.ph.acl.service.PackageResourceServiceAdapter;
import com.jfrog.ph.acl.service.PackageSearchArtifactService;
import com.jfrog.ph.acl.service.PackageSecurityService;
import com.jfrog.ph.acl.service.PackageSpecificationAdapter;
import com.jfrog.ph.acl.service.PackageTaskExecutorSupplier;
import com.jfrog.ph.acl.service.PackageUploadService;
import com.jfrog.ph.acl.service.PackageXrayService;
import com.jfrog.ph.acl.virtual.resolution.PackageMetadataRow;
import com.jfrog.ph.acl.virtual.resolution.PackageMetadataVersionsSet;
import com.jfrog.ph.acl.virtual.resolution.PackageRepo;
import com.jfrog.ph.acl.virtual.resolution.PackageRequestContext;
import com.jfrog.ph.acl.virtual.resolution.PackageVirtualResolutionAdapter;
import lombok.NonNull;
import org.apache.http.HttpStatus;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.oauth.OAuthHandler;
import org.artifactory.api.properties.PropertiesService;
import org.artifactory.api.request.DownloadService;
import org.artifactory.api.request.UploadService;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.aql.AqlService;
import org.artifactory.authentication.ArtifactoryAuthenticationContextFilter;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.common.repo.PackageType;
import org.artifactory.common.repo.RepoInfo;
import org.artifactory.common.resource.BaseResource;
import org.artifactory.config.ConfigurationChangesInterceptor;
import org.artifactory.confprop.ConfProps;
import org.artifactory.event.queue.QueueService;
import org.artifactory.ph.selector.IndexingSelector;
import org.artifactory.repo.interceptor.storage.StorageInterceptorAdapter;
import org.artifactory.repo.service.InternalRepositoryService;
import org.artifactory.repo.service.trash.NonTrashableItem;
import org.artifactory.schedule.CachedThreadPoolTaskExecutor;
import org.artifactory.security.InternalSecurityService;
import org.artifactory.security.access.AccessService;
import org.artifactory.security.props.auth.OauthManager;
import org.artifactory.spring.ConditionalOnMissingBean;
import org.artifactory.storage.security.service.UserGroupStoreService;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.Lazy;
import org.springframework.core.Ordered;
import org.springframework.security.authentication.AuthenticationManager;

import javax.annotation.Nonnull;
import javax.annotation.Priority;
import javax.servlet.http.HttpServletRequest;
import javax.ws.rs.core.Response;
import java.io.InputStream;
import java.util.LinkedHashSet;
import java.util.Map;

/**
 * Beans configuration class
 */
@SuppressWarnings("SpringJavaInjectionPointsAutowiringInspection")
@Configuration
public class ArtifactoryPackageBeansConfig {

    /**
     * Expose {@link org.artifactory.common.property.ArtifactorySystemProperties} as spring bean type of
     * {@link ConfProps}
     */
    @Bean
    public ConfProps getConfProps() {
        return ArtifactoryHome.get().getArtifactoryProperties();
    }


    @Bean
    public PackageRepoService packageRepoService(InternalRepositoryService repositoryService,
            PropertiesService propertiesService) {
        return new PackageRepoServiceImpl(repositoryService, propertiesService);
    }

    @Bean
    public PackageArtifactService packageDownloadService(DownloadService downloadService,
            InternalRepositoryService repositoryService) {
        return new PackageArtifactServiceImpl(downloadService, repositoryService);
    }

    @Bean
    public PackageUploadService packageUploadService(InternalRepositoryService repositoryService,
            UploadService uploadService) {
        return new PackageUploadServiceImpl(repositoryService, uploadService);
    }

    @Bean
    public StorageInterceptorAdapter packageStorageInterceptorAdapter(PackageArtifactEventListener packagesListener,
            InternalRepositoryService repositoryService, PackageSpecificationAdapter packageSpecificationAdapter) {
        return new PackageStorageInterceptorImpl(packagesListener, repositoryService, packageSpecificationAdapter);
    }

    @Bean
    public ConfigurationChangesInterceptor packageConfigurationChangesInterceptor(
            PackageConfigChangedEventListener packageConfigChangedEventListener,
            PackageRepoService packageRepoService) {
        return new PackageConfigurationChangesInterceptorImpl(packageConfigChangedEventListener,
                packageRepoService);
    }

    @Bean
    public NonTrashableItem PackageMetadataNonTrashableInterceptor(
            PackageSpecificationAdapter packageSpecificationAdapter) {
        return new PackageMetadataNonTrashableItem(packageSpecificationAdapter);
    }

    /**
     * We are using the Qualifier here because spring will load RealmAwareAuthenticationManager otherwise.
     * RT using PasswordDecryptingManager when using beansForType(AuthenticationManager.class) due the logic in
     * ArtifactoryApplicationContext
     */
    @Bean
    public PackageSecurityService packageSecurityService(InternalSecurityService securityService,
            AccessService accessService,
            @Qualifier("passwordDecryptingManager") AuthenticationManager authenticationManager,
            UserGroupStoreService userGroupStoreService, OauthManager oauthManager) {
        return new PackageSecurityServiceImpl(securityService, accessService, authenticationManager,
                userGroupStoreService, oauthManager);
    }

    @Bean
    public PackageSearchArtifactService packageSearchArtifactService(AqlService aqlService,
            AuthorizationService authorizationService, PackageRepoService repoService) {
        return new PackageSearchArtifactServiceImpl(aqlService, authorizationService, repoService);
    }

    @Bean
    public PackageSpecification packageSpecification(PackageSpecificationAdapter packageSpecificationAdapter) {
        return new PackageSpecificationImpl(packageSpecificationAdapter);
    }

    @Bean
    public PackageResourceService packageResourceService(PackageResourceServiceAdapter packageResourceServiceAdapter,
            PackageSpecificationAdapter packageSpecificationAdapter) {
        return new PackageResourceServiceImpl(packageResourceServiceAdapter, packageSpecificationAdapter);
    }

    @Bean
    public ArtifactoryAuthenticationContextFilter packageAuthenticationFilter(
            PackageAuthenticationAdapter packageAuthenticationAdapter) {
        return new PackageAuthenticationFilterImpl(packageAuthenticationAdapter);
    }

    @Bean
    @ConditionalOnMissingBean(PackageAuthenticationAdapter.class)
    public PackageAuthenticationAdapter packageAuthenticationAdapter() {
        return request -> PackageAuthenticationRequestContext.builder().build();
    }

    @Bean
    public PackageArtifactoryConfigService packageRepoDescriptorService(
            InternalRepositoryService internalRepositoryService) {
        return new PackageArtifactoryConfigServiceImpl(internalRepositoryService);
    }

    @Bean
    public PackageVirtualResolutionService<PackageMetadataRow> packageVirtualResolutionService(
            PackageVirtualResolutionAdapter<PackageMetadataRow> packageVirtualResolutionAdapter) {
        return new PackageVirtualResolutionServiceImpl(packageVirtualResolutionAdapter);
    }

    @Bean
    PackageEventService packageEventService(QueueService queueService) {
        return new PackageEventServiceImpl(queueService);
    }

    @Bean
    PackageTaskExecutorSupplier packageTaskExecutorSupplier(CachedThreadPoolTaskExecutor executor) {
        return new PackageTaskExecutorSupplierImpl(executor);
    }

    @Bean
    PackageXrayService packageXrayService(@Lazy AddonsManager addonsManager) {
        return new PackageXrayServiceImpl(addonsManager);
    }

    @Bean
    PackageOAuthHandler packageOAuthHandler(OAuthHandler oAuthHandler) {
        return new PackageOAuthHandlerImpl(oAuthHandler);
    }

    @Bean
    PackageAnalyticsService packageAnalyticsService(AddonsManager addonsManager) {
        return new PackageAnalyticsServiceImpl(addonsManager);
    }

    @Bean
    IndexingSelector indexingSelector(PackageRepoService packageRepoService,
            PackageSpecificationAdapter packageSpecificationAdapter, AddonsManager addonsManager,
            PackageRepoIndexCalculatorService repoIndexCalculatorService, AuthorizationService authorizationService) {
        return new PackageIndexingSelectorImpl(packageRepoService, packageSpecificationAdapter, addonsManager,
                repoIndexCalculatorService, authorizationService);
    }

    /**
     * Configuration class for default beans that should be loaded only in None PRO run
     * Must have the lowest priority since has condition on missing bean.
     * We want to wait for the real beans a chance to be loaded, and if not, we load the default beans.
     */
    @Configuration
    @Priority(Ordered.LOWEST_PRECEDENCE)
    public static class PackageOssConfig {

        /**
         * Note: The method name must be the same as the method name of the real impl.
         * See
         * (com.jfrog.ph.config.PackageBeansConfig)}
         */
        @Bean
        @ConditionalOnMissingBean(PackageArtifactEventListener.class)
        public PackageArtifactEventListener packageArtifactEventListener() {
            return new PackageArtifactEventListener() {
            };
        }

        @Bean
        public PackageNugetCache packageNugetCache() {
            return new DummyPackageNugetCacheImpl();
        }

        /**
         * Note: The method name must be the same as the method name of the real impl.
         * See
         * (com.jfrog.ph.config.PackageBeansConfig)}
         */
        @Bean
        @ConditionalOnMissingBean(PackageConfigChangedEventListener.class)
        public PackageConfigChangedEventListener packageConfigChangedEventListener() {
            return new PackageConfigChangedEventListener() {
                @Override
                public void onGlobalConfigChange(@NonNull PackageConfigChangedEventType packageConfigChangedEventType) {

                }

                @Override
                public void onRepoConfigChange(@NonNull PackageRepoConfigListenerContext context) {

                }
            };
        }

        /**
         * Note: The method name must be the same as the method name of the real impl.
         * See
         * (com.jfrog.ph.service.InternalPackageHandlerService,
         * com.jfrog.ph.acl.service.PackageRepoService,
         * com.jfrog.ph.acl.service.PackageDownloadService,
         * com.jfrog.ph.acl.service.PackageUploadService,
         * com.jfrog.ph.service.PackageHandlerConflictsGuardService,
         * org.artifactory.confprop.ConfProps)}
         */
        @Bean
        @ConditionalOnMissingBean(PackageVirtualResolutionAdapter.class)
        public PackageVirtualResolutionAdapter<PackageMetadataRow> packageVirtualAdapter() {
            return new PackageVirtualResolutionAdapter<>() {
                @Override
                public boolean shouldCacheMergedMetadata(String virtualRepoKey, String path) {
                    return false;
                }

                @Override
                public boolean shouldUsePackageVirtualResolution(String virtualRepoKey) {
                    return false;
                }

                @Override
                public BaseResource getItemInfoFromVirtual(PackageRepo virtualRepo,
                        PackageRequestContext context) {
                    return null;
                }

                @Nonnull
                @Override
                public Map<String, LinkedHashSet<PackageMetadataRow>> extract(BaseResource res, String virtualRepoKey,
                        String originalPath, PackageRepo sourceRepo)
                        throws PackageVirtualTraversalException {
                    String message = "The 'extract' method is not supported in OSS release";
                    throw new PackageVirtualTraversalException(HttpStatus.SC_INTERNAL_SERVER_ERROR,
                            "Unsupported request", message);
                }

                @Nonnull
                @Override
                public InputStream toInputStream(
                        Map<String, PackageMetadataVersionsSet<PackageMetadataRow>> metadata,
                        PackageRequestContext requestContext) throws PackageVirtualTraversalException {
                    String message = "The 'toInputStream' method is not supported in OSS release";
                    throw new PackageVirtualTraversalException(HttpStatus.SC_INTERNAL_SERVER_ERROR,
                            "Unsupported request", message);
                }
            };
        }

        /**
         * Note: The method name must be the same as the method name of the real impl.
         * See
         * (com.jfrog.ph.service.InternalPackageHandlerService)}
         */
        @Bean
        @ConditionalOnMissingBean(PackageRepoIndexCalculatorService.class)
        public PackageRepoIndexCalculatorService packageRepoIndexCalculatorService() {
            return new PackageRepoIndexCalculatorService() {
                @Override
                public void reindex(PackageType packageType, String repoKey) {

                }

                @Override
                public void reindexSpecificPath(PackageType packageType, String repoKey, String path) {
                }
            };
        }

        /**
         * Note: The method name must be the same as the method name of the real impl.
         * See
         * (com.jfrog.ph.service.InternalPackageHandlerService)}
         */
        @Bean
        @ConditionalOnMissingBean(PackageSpecificationAdapter.class)
        public PackageSpecificationAdapter packageSpecificationAdapter() {
            return new PackageSpecificationAdapter() {

                @Override
                public boolean isPackageArtifact(@Nonnull String repoKey, @Nonnull String path) {
                    return false;
                }

                @Override
                public boolean isExpirable(@Nonnull PackageType packageType, @Nonnull String path) {
                    return false;
                }

                @Override
                public boolean isLocalGenerated(@Nonnull PackageType packageType, @Nonnull String path) {
                    return false;
                }

                @Override
                public boolean isLocallyGeneratedProperty(@Nonnull PackageType packageType,
                        @Nonnull String propertyKey) {
                    return false;
                }

                @Override
                public boolean isSupportedPackage(PackageType packageType) {
                    return false;
                }

                @Override
                public boolean isSupportedPackage(String repoKey) {
                    return false;
                }

                @Override
                public boolean hasReindex(RepoInfo repoInfo) {
                    return false;
                }

                @Override
                public boolean isViewable(String repoKey, String path) {
                    return false;
                }

                @Override
                public boolean shouldVerifyRemoteArtifact(PackageArtifact artifact) {
                    return false;
                }

                @Override
                public void verifyRemoteArtifactNotCorrupted(PackageArtifact artifact, InputStream is) {
                }

                @Override
                public boolean shouldMarkOriginalPropertiesInBeforeEvents(String repoKey, String artifactPath) {
                    return false;
                }
            };
        }

        /**
         * Note: The method name must be the same as the method name of the real impl.
         * See
         * (com.jfrog.ph.service.InternalPackageHandlerService)}
         */
        @Bean
        @ConditionalOnMissingBean(PackageResourceServiceAdapter.class)
        public PackageResourceServiceAdapter packageResourceServiceAdapter() {
            return new PackageResourceServiceAdapter() {
                @Override
                public Response reindex(@NonNull PackageType packageType, @NonNull String repoKey) {
                    return null;
                }

                @Override
                public Object getVersionedSubResource(@NonNull PackageType packageType, String version,
                        @NonNull String repoKey, HttpServletRequest request) {
                    return null;
                }

                @Override
                public Object getNonVersionedSubResource(@NonNull PackageType packageType, @NonNull String context,
                        HttpServletRequest request) {
                    return null;
                }

                @Override
                public Response getPackageInfo(String repoKey, String path) {
                    return null;
                }

                @Override
                public Response getVersionNativeReadme(String type, String fullRepoPath) {
                    return null;
                }
            };
        }
    }
}
