package org.artifactory.ph;

import com.google.common.collect.Lists;
import com.jfrog.ph.acl.exception.PackageException;
import com.jfrog.ph.acl.model.PackageTokenType;
import com.jfrog.ph.acl.model.SimpleUser;
import com.jfrog.ph.acl.service.PackageSecurityService;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang.StringUtils;
import org.apache.http.HttpStatus;
import org.artifactory.api.security.SecurityService;
import org.artifactory.api.security.UserInfoBuilder;
import org.artifactory.api.security.access.UserTokenSpec;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.crypto.CryptoHelper;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.security.AuthenticationHelper;
import org.artifactory.security.InternalSecurityService;
import org.artifactory.security.MutableUserInfo;
import org.artifactory.security.UserInfo;
import org.artifactory.security.access.AccessService;
import org.artifactory.security.access.AccessTokenAuthentication;
import org.artifactory.security.props.auth.TokenManager;
import org.artifactory.security.props.auth.model.AuthenticationModel;
import org.artifactory.security.props.auth.model.OauthErrorModel;
import org.artifactory.security.props.auth.model.OauthModel;
import org.artifactory.security.props.auth.model.TokenKeyValue;
import org.artifactory.security.props.auth.model.TokenManagerKey;
import org.artifactory.security.props.auth.model.TokenType;
import org.artifactory.storage.security.service.UserGroupStoreService;
import org.artifactory.util.date.DateUtils;
import org.jfrog.access.client.AccessClientException;
import org.springframework.security.authentication.AuthenticationManager;
import org.springframework.security.authentication.BadCredentialsException;
import org.springframework.security.authentication.UsernamePasswordAuthenticationToken;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.security.core.userdetails.UserDetails;

import javax.annotation.Nullable;
import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.Response;
import java.util.concurrent.Callable;

@Slf4j
public class PackageSecurityServiceImpl implements PackageSecurityService {

    public static final String FAILED_TO_AUTHENTICATE_LOCALLY = "Failed to authenticate locally.";
    public static final String LOCAL_TOKEN_CREATION_FAILED = "Local token creation failed";
    private final InternalSecurityService securityService;
    private final AccessService accessService;
    private final AuthenticationManager authenticationManager;
    private final UserGroupStoreService userGroupStoreService;
    private final TokenManager tokenManager;

    public PackageSecurityServiceImpl(InternalSecurityService securityService, AccessService accessService,
            AuthenticationManager authenticationManager,
            UserGroupStoreService userGroupStoreService, TokenManager tokenManager) {
        this.securityService = securityService;
        this.accessService = accessService;
        this.authenticationManager = authenticationManager;
        this.userGroupStoreService = userGroupStoreService;
        this.tokenManager = tokenManager;
    }

    @Override
    public boolean canRead(String repoKey, String path) {
        return securityService.canRead(RepoPathFactory.create(repoKey, path));
    }

    @Override
    public boolean canWrite(String repoKey, String path) {
        return securityService.canDeploy(RepoPathFactory.create(repoKey, path));
    }

    @Override
    public boolean canAnnotate(String repoKey, String path) {
        return securityService.canAnnotate(RepoPathFactory.create(repoKey, path));
    }

    @Override
    public boolean canDelete(String repoKey, String path) {
        return securityService.canDelete(RepoPathFactory.create(repoKey, path));
    }

    @Override
    public boolean canManage(String repoKey, String path) {
        return securityService.canManage(RepoPathFactory.create(repoKey, path));
    }

    @Override
    public <V> V callAsSystem(Callable<V> callable) throws PackageException {
        try {
            return securityService.callAsSystem(callable);
        } catch (PackageException e) {
            throw e;
        } catch (Exception e) {
            throw new PackageException(HttpStatus.SC_INTERNAL_SERVER_ERROR, e.getMessage(), e.getMessage(), e);
        }
    }

    @Override
    public boolean isAnonymous() {
        return securityService.isAnonymous();
    }

    @Override
    public boolean isAnonAccessEnabled() {
        return securityService.isAnonAccessEnabled();
    }

    @Override
    @NonNull
    public String getCurrentUserName() {
        return securityService.currentUsername();
    }

    @Override
    @NonNull
    public String getSystemUserName() {
        return SecurityService.USER_SYSTEM;
    }

    @Override
    public boolean isAuthenticated() {
        return securityService.isAuthenticated();
    }

    @Override
    @NonNull
    public String loginAndGenerateAccessTokenWithBasicAuth(String userName, String password) throws PackageException {
        Authentication originalAuth = SecurityContextHolder.getContext().getAuthentication();
        UsernamePasswordAuthenticationToken authenticationToken =
                new UsernamePasswordAuthenticationToken(userName, password);
        try {
            Authentication authentication = authenticationManager.authenticate(authenticationToken);
            if (authentication instanceof AccessTokenAuthentication accessTokenAuthentication) {
                // user came with access token - we should use the same token
                return accessTokenAuthentication.getAccessTokenValue();
            }
            // Adding the authentication so Access will get it for token
            SecurityContextHolder.getContext().setAuthentication(authentication);
            return generateAccessTokenForUser(userName);
        } catch (BadCredentialsException e) {
            log.warn(FAILED_TO_AUTHENTICATE_LOCALLY + ":{}", e.getMessage());
            log.debug(FAILED_TO_AUTHENTICATE_LOCALLY + " for user: '{}'", userName, e);
            throw new PackageException(HttpStatus.SC_UNAUTHORIZED, FAILED_TO_AUTHENTICATE_LOCALLY,
                    FAILED_TO_AUTHENTICATE_LOCALLY);
        } catch (Exception e) {
            log.error(FAILED_TO_AUTHENTICATE_LOCALLY + ":{}", e.getMessage());
            log.debug(FAILED_TO_AUTHENTICATE_LOCALLY + " for user: '{}'", userName, e);
            throw new PackageException(HttpStatus.SC_BAD_REQUEST, FAILED_TO_AUTHENTICATE_LOCALLY,
                    FAILED_TO_AUTHENTICATE_LOCALLY);
        } finally {
            SecurityContextHolder.getContext().setAuthentication(originalAuth);
        }
    }

    @Override
    public void revokeToken(String token) throws PackageException {
        try {
            accessService.revokeToken(token);
        } catch (AccessClientException accessClientException) {
            log.debug(accessClientException.getMessage());
        } catch (Exception e) {
            String msg = "Logout locally failed";
            log.warn(msg + ": error on token revoke.", e.getMessage());
            log.debug(msg + ": error on token revoke of: {}", token, e);
            throw new PackageException(HttpStatus.SC_BAD_REQUEST, msg, msg);
        }
    }

    @Override
    public SimpleUser getCurrentUserDetails() throws PackageException {
        Authentication authentication = AuthenticationHelper.getAuthentication();
        if ((authentication == null) || !authentication.isAuthenticated()) {
            String msg = "User is unauthenticated";
            throw new PackageException(HttpStatus.SC_BAD_REQUEST, msg, msg);
        }
        String authUsername = ((UserDetails) authentication.getPrincipal()).getUsername();
        String password = (String) authentication.getCredentials();
        if (StringUtils.isNotBlank(password)) {
            // Check if the user is an Access Token user
            if (authentication instanceof AccessTokenAuthentication accessTokenAuthentication) {
                String username = accessService
                        .extractSubjectUsername(accessTokenAuthentication.getAccessToken());
                MutableUserInfo transientUser = new UserInfoBuilder(username).transientUser().build();
                return new SimpleUser(transientUser.getUsername(), password, transientUser.getEmail());
            }
            UserInfo user = userGroupStoreService.findUser(authUsername);
            if (user != null) {
                if (!CryptoHelper.isApiKey(password)) {
                    password = securityService.createEncryptedPasswordIfNeeded(user, password);
                }
                return new SimpleUser(authUsername, password, user.getEmail());
            }
        }
        String msg = "The user: '" + authentication.getPrincipal() + "' couldn't be found";
        throw new PackageException(HttpStatus.SC_BAD_REQUEST, msg, msg);
    }

    @Nullable
    @Override
    public SimpleUser findUserByProperty(String key, String val, boolean exactKeyMatch) {
        UserInfo userInfo = userGroupStoreService.findUserByProperty(key, val, exactKeyMatch);
        if (userInfo != null) {
            return new SimpleUser(userInfo.getUsername(), userInfo.getPassword(), userInfo.getEmail());
        }
        return null;
    }

    @Override
    public boolean addUserProperty(String username, String key, String value) {
        return userGroupStoreService.addUserProperty(username, key, value);
    }
    @Override
    public boolean deleteUserProperty(String username, String propKey) {
        return userGroupStoreService.deleteUserProperty(username, propKey);
    }

    @Override
    public String encryptIfNeeded(String password) {
        return CryptoHelper.encryptIfNeeded(ArtifactoryHome.get(), password);
    }

    @Override
    public String decryptIfNeeded(String password) {
        return CryptoHelper.decryptIfNeeded(ArtifactoryHome.get(), password);
    }

    @Override
    public String createTokenAsSystem(String userName, PackageTokenType packageTokenType) throws Exception {
        TokenType tokenType = mapToTokenType(packageTokenType);
        TokenKeyValue tokenKeyValue =
                securityService.callAsSystem(() -> tokenManager.createToken(new TokenManagerKey(userName, tokenType)));
        OauthModel oauthModel =
                new AuthenticationModel(tokenKeyValue.getToken(), DateUtils.formatBuildDate(System.currentTimeMillis()),
                        tokenKeyValue.getExpiresIn());
        return tokenFromOAuthModel(oauthModel);
    }

    @NonNull
    private TokenType mapToTokenType(PackageTokenType packageTokenType) throws IllegalArgumentException {
        return TokenType.valueOf(packageTokenType.name());
    }

    @NonNull
    private String tokenFromOAuthModel(OauthModel oauthModel) throws WebApplicationException {
        if (oauthModel instanceof AuthenticationModel authenticationModel) {
            return authenticationModel.getToken();
        } else if (oauthModel instanceof OauthErrorModel oauthErrorModel) {
            log.debug("An error occurred while trying to generate token");
            throw new WebApplicationException(errorResponse(oauthErrorModel));
        }
        log.debug("Unexpected OauthModel: {}", oauthModel == null ? null : oauthModel.getClass());
        throw new WebApplicationException(HttpStatus.SC_INTERNAL_SERVER_ERROR);
    }

    private Response errorResponse(OauthErrorModel errorModel) {
        int status = errorModel.getStatusCode();
        String message = errorModel.getMessage();
        return Response.status(status).entity(message).build();
    }

    /**
     * Generating Access token for the user with following specs: Expired in
     * {@link ConstantValues#genericTokensExpirationTimeSecs}, non-Refreshable, no groups' limitation.
     * We set the scope to be "member-of-groups:*" since it will allow the user to obtain groups access according
     * to his group memberships AT THE TIME OF EACH LOGIN.
     */
    private String generateAccessTokenForUser(String username) throws PackageException {
        // Creating the token to be valid + with no specific groups limitation
        try {
            UserTokenSpec tokenSpec = UserTokenSpec.create(username)
                    .expiresIn(ConstantValues.genericTokensExpirationTimeSecs.getLong())
                    .refreshable(false)
                    .scope(Lists.newArrayList("member-of-groups:*"));
            return accessService.createToken(tokenSpec, true, false).getTokenValue();
        } catch (Exception e) {
            log.warn(LOCAL_TOKEN_CREATION_FAILED + ": {}", e.getMessage());
            log.debug("Local token creation failed for user: '{}'", username, e);
            throw new PackageException(HttpStatus.SC_BAD_REQUEST, LOCAL_TOKEN_CREATION_FAILED,
                    LOCAL_TOKEN_CREATION_FAILED);
        }
    }
}
