package org.artifactory.projects.interceptor;

import org.artifactory.projects.ProjectsService;
import org.artifactory.projects.exception.AccessClientStateException;
import org.artifactory.repo.RepoConfigChange;
import org.artifactory.repo.RepoDetailsType;
import org.artifactory.repo.interceptor.RepoConfigChangeInterceptor;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.stream.Stream;

import static org.artifactory.repo.RepoConfigChange.ConfigAction.REPO_DELETE;
import static org.artifactory.repo.RepoDetailsType.FEDERATED;
import static org.artifactory.repo.RepoDetailsType.LOCAL;
import static org.artifactory.repo.RepoDetailsType.REMOTE;
import static org.artifactory.repo.RepoDetailsType.VIRTUAL;

/**
 * Handles repository removal from Projects (managed in Access) for local, remote and virtual repositories
 *
 * @author andreik
 */
@Component
public class ProjectRepoConfigurationInterceptor implements RepoConfigChangeInterceptor {

    private static final Logger log = LoggerFactory.getLogger(ProjectRepoConfigurationInterceptor.class);

    private final ProjectsService projectsService;

    @Autowired
    public ProjectRepoConfigurationInterceptor(ProjectsService projectsService) {
        this.projectsService = projectsService;
    }

    @Override
    public void onBeforeSave(Map<RepoDetailsType, List<RepoConfigChange>> mutableRepoConfigChanges) {
        List<RepoConfigChange> localRepoChanges = mutableRepoConfigChanges.getOrDefault(LOCAL, List.of());
        List<RepoConfigChange> federatedRepoChanges = mutableRepoConfigChanges.getOrDefault(FEDERATED, List.of());
        List<RepoConfigChange> remoteRepoChanges = mutableRepoConfigChanges.getOrDefault(REMOTE, List.of());
        List<RepoConfigChange> virtualRepoChanges = mutableRepoConfigChanges.getOrDefault(VIRTUAL, List.of());

        Stream.of(localRepoChanges, federatedRepoChanges, remoteRepoChanges, virtualRepoChanges)
                .flatMap(Collection::stream)
                .filter(repoConfigChange -> REPO_DELETE.equals(repoConfigChange.getConfigAction()))
                .forEach(repoConfigChange -> removeDeletedRepositoriesFromProjects(repoConfigChange.getRepoKey(),
                        repoConfigChange.getOldRepoConfig().getRepoType()));
    }

    private void removeDeletedRepositoriesFromProjects(String repoKey, RepoDetailsType repoType) {
        try {
            // for every project remove the repo
            log.debug("Repo '{}' from type '{}' was deleted, removing from Access", repoKey, repoType);
            projectsService.removeRepositoryFromCurrentProject(repoKey);
        } catch (AccessClientStateException e) {
            log.warn("Failed to remove repo '{}' from a project association: {}. Will be handled by " +
                    "SyncResourcesWithAccessJob", repoKey, e.getMessage());
        } catch (Exception e) {
            log.error("Failed to remove repo '{}' from a project association: {}", repoKey, e.getMessage());
            log.debug("Failed to remove repo '{}' from a project association: {}", repoKey, e.getMessage(),
                    e);
        }
    }
}
