package org.artifactory.repo.cache;

import com.google.common.annotations.VisibleForTesting;
import com.google.common.base.Joiner;
import com.google.common.cache.CacheLoader;
import lombok.extern.slf4j.Slf4j;
import org.artifactory.model.RepoConfig;
import org.checkerframework.checker.nullness.qual.Nullable;

import java.util.Map;
import java.util.Set;
import java.util.function.Function;
import java.util.function.Supplier;
import java.util.stream.StreamSupport;

/**
 * Cache loader for repo configs. Based on {@link CacheLoader}.
 * Has optimization for batch load repo configs by keys according to an actual state of repo count and app settings,
 * since sometimes it is faster to load all data than do find by keys via IN.
 */
@Slf4j
public class RepoConfigCacheLoader extends CacheLoader<String, RepoConfig> {
    private final Supplier<Integer> allReposCountFunction;
    private final Supplier<Map<String, RepoConfig>> getAllRepoConfigsFunction;
    private final Function<String, RepoConfig> getRepoConfigByKeyFunction;
    private final Function<Set<String>, Map<String, RepoConfig>> getRepoConfigsByKeysFunction;
    private final int batchOperationSize;
    private final double optimizationRatio;

    /**
     * @param batchOperationSize           - DB batch operation size for load all query
     * @param optimizationRatio            - optimization ratio to define when optimization should be applied.
     *                                     If less than 0 - will not be applied.
     *                                     If equals to 0 - will be applied always.
     * @param allReposCountFunction        - function to get current number of repos
     * @param getAllRepoConfigsFunction    - function to load all repo configs from DB
     * @param getRepoConfigByKeyFunction   - function to load a repo config by key from DB
     * @param getRepoConfigsByKeysFunction - function to load repo configs by keys from DB
     */
    public RepoConfigCacheLoader(int batchOperationSize, double optimizationRatio,
            Supplier<Integer> allReposCountFunction,
            Supplier<Map<String, RepoConfig>> getAllRepoConfigsFunction,
            Function<String, RepoConfig> getRepoConfigByKeyFunction,
            Function<Set<String>, Map<String, RepoConfig>> getRepoConfigsByKeysFunction) {
        this.batchOperationSize = batchOperationSize;
        this.optimizationRatio = optimizationRatio;
        this.allReposCountFunction = allReposCountFunction;
        this.getAllRepoConfigsFunction = getAllRepoConfigsFunction;
        this.getRepoConfigByKeyFunction = getRepoConfigByKeyFunction;
        this.getRepoConfigsByKeysFunction = getRepoConfigsByKeysFunction;
    }

    @Override
    public @Nullable RepoConfig load(String key) {
        try {
            log.trace("Loading repo config key: {} from DB...", key);
            return getRepoConfigByKeyFunction.apply(key);
        } finally {
            log.trace("Loading repo config key: {} from DB has been done");
        }
    }

    @Override
    public Map<String, RepoConfig> loadAll(Iterable<? extends String> keys) throws Exception {
        try {
            int keysCount = (int) StreamSupport.stream(keys.spliterator(), false).count();
            if (shouldGetAllOptimizationBeApplied(keysCount)) {
                log.trace("Loading all repo configs from DB because of optimization...");
                // it is ok to return all of loaded repo configs, since they are loaded anyway. All of them
                // will be updated in cache
                return getAllRepoConfigsFunction.get();
            }

            if (log.isTraceEnabled()) {
                log.trace("Loading repo configs by keys: [{}] from DB...", Joiner.on(",").join(keys));
            }
            return getRepoConfigsByKeysFunction.apply((Set<String>) keys);
        } finally {
            if (log.isTraceEnabled()) {
                log.trace("Loading repo configs by keys: [{}] from DB has been done", Joiner.on(",").join(keys));
            }
        }
    }

    /**
     * Defines should getAll optimization be applied. Since some times, it is faster to do getAll instead of loading
     * repo configs by keys via IN. For example:
     * ---
     * keys to fetch = 10, total keys = 100, batch = 5
     * 2 batches - useful, 18 batches - unuseful
     * ratio = 2/18 = 0.11, optimization ratio = 1, getAll is not effective
     * ---
     * keys to fetch = 60, total keys = 100, batch = 5
     * 12 batches - useful, 8 batches - unuseful
     * ratio = 12/8 = 1.5, optimization ratio = 1, getAll is effective
     */
    @VisibleForTesting
    boolean shouldGetAllOptimizationBeApplied(int keysToLoad) {
        int allRepoKeysCount = allReposCountFunction.get();
        log.trace("Keys to load : {}, all keys count: {}, batch size: {}, optimization ratio: {}", keysToLoad,
                allRepoKeysCount, batchOperationSize, optimizationRatio);
        // if allowed ratio < 0 - do not do optimization
        if (optimizationRatio < 0) {
            log.trace("Skip optimization because optimization ratio less than 0");
            return false;
        }
        // if allowed ratio is 0 - always do optimization
        if (optimizationRatio == 0) {
            log.trace("Always do optimization because optimization ratio is 0");
            return true;
        }
        // no need to get all if number of keys less than batch size
        if (keysToLoad <= batchOperationSize) {
            log.trace("Skip optimization because keys to load less than batch size");
            return false;
        }
        // just a validation - almost impossible. Get all repos if client requests more repo keys than repo keys
        // set has
        if (keysToLoad >= allRepoKeysCount) {
            log.trace("Do optimization because keys to load more than all repo keys");
            return true;
        }
        // use ceil since there is impossible to execute half of the batch, if it is required to get 1 from the
        // next batch - whole batch is useful.
        double usefulRepoConfigBatchQueries = Math.ceil((double) keysToLoad / batchOperationSize);
        double allRepoConfigBatchQueries = Math.ceil((double) allRepoKeysCount / batchOperationSize);
        // great almost impossible
        if (usefulRepoConfigBatchQueries >= allRepoConfigBatchQueries) {
            log.trace("Do optimization because useful batch queries {} more than all batch queries {}",
                    usefulRepoConfigBatchQueries, allRepoConfigBatchQueries);
            return true;
        }
        double unusefulRepoConfigBatchQueries = allRepoConfigBatchQueries - usefulRepoConfigBatchQueries;
        log.trace("Unuseful batch queries: {}", unusefulRepoConfigBatchQueries);
        double ration = usefulRepoConfigBatchQueries / unusefulRepoConfigBatchQueries;
        boolean isOptimizationHelpful = ration >= optimizationRatio;
        log.trace("Calculated ratio: {}, is optimization helpful: {}", ration, isOptimizationHelpful);
        return isOptimizationHelpful;
    }
}