package org.artifactory.repo.migration.mappers;

import lombok.extern.slf4j.Slf4j;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.descriptor.property.PropertySet;
import org.artifactory.model.FederatedRepoConfig;
import org.artifactory.model.FederatedRepositoryRepoTypeConfig;
import org.artifactory.repo.CdnRedirectRepoConfig;
import org.artifactory.repo.DownloadRedirectRepoConfig;
import org.artifactory.repo.federated.FederatedMember;
import org.artifactory.repo.local.LocalRepoChecksumPolicyType;
import org.artifactory.repo.packagetype.SnapshotVersionBehavior;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

@Slf4j
public class FederatedRepoConfigMapper {
    private FederatedRepoConfigMapper() {
    }

    public static FederatedRepoConfig mapFederatedRepoConfig(FederatedRepoConfig federatedRepoConfig,
            Map<String, Object> newValues, CentralConfigDescriptor descriptor) {
        RepoConfigMigrationPatchUtils.buildBaseConfig(federatedRepoConfig.getBaseConfig(), newValues, descriptor);
        buildRepoTypeConfig(federatedRepoConfig.getRepoTypeConfig(), newValues, descriptor);
        RepoConfigMigrationPatchUtils.buildLocalPackageTypeConfig(federatedRepoConfig.getPackageTypeConfig(),
                federatedRepoConfig.getPackageType(),
                newValues);
        RepoConfigMigrationPatchUtils.buildSecurityConfig(federatedRepoConfig.getSecurityConfig(), newValues,
                descriptor);
        return federatedRepoConfig;
    }

    private static void buildRepoTypeConfig(FederatedRepositoryRepoTypeConfig repositoryRepoTypeConfig,
            Map<String, Object> newValues, CentralConfigDescriptor descriptor) {
        for (Map.Entry<String, Object> newValue : newValues.entrySet()) {
            try {
                if ("snapshotVersionBehavior".equals(newValue.getKey())) {
                    repositoryRepoTypeConfig.setSnapshotVersionBehavior(
                            SnapshotVersionBehavior.valueOf(String.valueOf(newValue.getValue())));
                } else if ("maxUniqueSnapshots".equals(newValue.getKey())) {
                    repositoryRepoTypeConfig.setMaxUniqueSnapshots((int) newValue.getValue());
                } else if ("handleReleases".equals(newValue.getKey())) {
                    repositoryRepoTypeConfig.setHandleReleases((boolean) newValue.getValue());
                } else if ("handleSnapshots".equals(newValue.getKey())) {
                    repositoryRepoTypeConfig.setHandleSnapshots((boolean) newValue.getValue());
                } else if ("cdnRedirectRepoConfig".equals(newValue.getKey())) {
                    CdnRedirectRepoConfig cdnRedirectRepoConfig = new CdnRedirectRepoConfig();
                    cdnRedirectRepoConfig.setEnabled(
                            (boolean) ((Map<String, Object>) newValue.getValue()).get("enabled"));
                    repositoryRepoTypeConfig.setCdnRedirectRepoConfig(cdnRedirectRepoConfig);
                } else if ("downloadRedirect".equals(newValue.getKey())) {
                    DownloadRedirectRepoConfig downloadRedirectRepoConfig = new DownloadRedirectRepoConfig();
                    downloadRedirectRepoConfig.setEnabled(
                            (boolean) ((Map<String, Object>) newValue.getValue()).get("enabled"));
                    repositoryRepoTypeConfig.setDownloadRedirectConfig(downloadRedirectRepoConfig);
                } else if ("blackedOut".equals(newValue.getKey())) {
                    repositoryRepoTypeConfig.setBlackedOut((boolean) newValue.getValue());
                } else if ("archiveBrowsingEnabled".equals(newValue.getKey())) {
                    repositoryRepoTypeConfig.setArchiveBrowsingEnabled((boolean) newValue.getValue());
                } else if ("checksumPolicyType".equals(newValue.getKey())) {
                    repositoryRepoTypeConfig.setChecksumPolicyType(
                            LocalRepoChecksumPolicyType.fromString(
                                    (String) newValue.getValue()));
                } else if ("priorityResolution".equals(newValue.getKey())) {
                    repositoryRepoTypeConfig.setPriorityResolution((boolean) newValue.getValue());
                } else if ("propertySets".equals(newValue.getKey())) {
                    Set<String> existingPropertySetKeys = descriptor.getPropertySets().stream()
                            .map(PropertySet::getName)
                            .collect(Collectors.toSet());
                    Map<String, Object> properties = (Map<String, Object>) newValue.getValue();
                    List<String> propertiesList = properties.entrySet().stream()
                            .sorted(Map.Entry.comparingByKey())
                            .map(Map.Entry::getValue)
                            .map(String.class::cast)
                            .filter(propertySetKey -> {
                                if (!existingPropertySetKeys.contains(propertySetKey)) {
                                    log.debug("Property set '{}' does not exist in the config descriptor",
                                            propertySetKey);
                                    return false;
                                }
                                return true;
                            })
                            .toList();
                    repositoryRepoTypeConfig.setPropertySetRefs(propertiesList);
                } else if ("federatedMembers".equals(newValue.getKey())) {
                    Map<String, Object> federatedMembers = (Map<String, Object>) newValue.getValue();
                    List<FederatedMember> federatedMembersConfig = new ArrayList<>();
                    List<Map<String, Object>> federatedMembersList = federatedMembers.entrySet().stream()
                            .sorted(Map.Entry.comparingByKey())
                            .map(Map.Entry::getValue)
                            .map(x -> (Map<String, Object>) x)
                            .toList();
                    federatedMembersList.forEach(federatedMembers1 -> {
                        FederatedMember federatedMemberInner = new FederatedMember();
                        for (Map.Entry<String, Object> federatedMember : federatedMembers1.entrySet()) {
                            if ("enabled".equals(federatedMember.getKey())) {
                                federatedMemberInner.setEnabled((boolean) federatedMember.getValue());
                            } else if ("mode".equals(federatedMember.getKey())) {
                                federatedMemberInner.setMode(String.valueOf(federatedMember.getValue()));
                            } else if ("url".equals(federatedMember.getKey())) {
                                federatedMemberInner.setUrl(String.valueOf(federatedMember.getValue()));
                            }
                        }
                        federatedMembersConfig.add(federatedMemberInner);
                    });
                    repositoryRepoTypeConfig.setFederatedMembers(federatedMembersConfig);

                } else if ("modificationDate".equals(newValue.getKey())) {
                    repositoryRepoTypeConfig.setModificationDate(
                            Integer.valueOf(String.valueOf(newValue.getValue())).longValue());
                } else if ("proxy".equals(newValue.getKey())) {
                    String proxyRef = String.valueOf(newValue.getValue());
                    if (descriptor.getProxy(proxyRef) == null) {
                        log.debug("Proxy '{}' does not exist in the config descriptor", proxyRef);
                    } else {
                        repositoryRepoTypeConfig.setProxyRef(proxyRef);
                    }
                } else if ("disableProxy".equals(newValue.getKey())) {
                    repositoryRepoTypeConfig.setDisableProxy((boolean) newValue.getValue());
                } else if ("maxNumberOfThreads".equals(newValue.getKey())) {
                    repositoryRepoTypeConfig.setMaxNumberOfThreads(
                            Integer.parseInt(String.valueOf(newValue.getValue())));
                }
            } catch (ClassCastException e) {
                throw new IllegalArgumentException("No values under " + newValue.getKey());
            }
        }
    }

    private static FederatedMember cast(Map<String, Object> m) {
        return new FederatedMember();
    }
}