package org.artifactory.repo.migration.mappers;

import lombok.extern.slf4j.Slf4j;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.descriptor.property.PropertySet;
import org.artifactory.model.LocalRepoConfig;
import org.artifactory.model.LocalRepositoryRepoTypeConfig;
import org.artifactory.repo.CdnRedirectRepoConfig;
import org.artifactory.repo.DownloadRedirectRepoConfig;
import org.artifactory.repo.local.LocalRepoChecksumPolicyType;
import org.artifactory.repo.packagetype.SnapshotVersionBehavior;

import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

@Slf4j
public class LocalRepoConfigMapper {

    public static LocalRepoConfig mapLocalRepoConfig(LocalRepoConfig localRepoConfig, Map<String, Object> newValues,
            CentralConfigDescriptor descriptor) {
        RepoConfigMigrationPatchUtils.buildBaseConfig(localRepoConfig.getBaseConfig(), newValues, descriptor);
        buildRepoTypeConfig(localRepoConfig.getRepoTypeConfig(), newValues, descriptor);
        RepoConfigMigrationPatchUtils.buildLocalPackageTypeConfig(localRepoConfig.getPackageTypeConfig(),
                localRepoConfig.getPackageType(),
                newValues);
        RepoConfigMigrationPatchUtils.buildSecurityConfig(localRepoConfig.getSecurityConfig(), newValues, descriptor);
        return localRepoConfig;
    }

    private static void buildRepoTypeConfig(LocalRepositoryRepoTypeConfig repositoryRepoTypeConfig,
            Map<String, Object> newValues, CentralConfigDescriptor descriptor) {
        for (Map.Entry<String, Object> newValue : newValues.entrySet()) {
            try {
                if ("snapshotVersionBehavior".equals(newValue.getKey())) {
                    repositoryRepoTypeConfig.setSnapshotVersionBehavior(SnapshotVersionBehavior.valueOf(
                            ((String) newValue.getValue()).toUpperCase(Locale.ROOT)));
                } else if ("maxUniqueSnapshots".equals(newValue.getKey())) {
                    repositoryRepoTypeConfig.setMaxUniqueSnapshots((int) newValue.getValue());
                } else if ("handleReleases".equals(newValue.getKey())) {
                    repositoryRepoTypeConfig.setHandleReleases((boolean) newValue.getValue());
                } else if ("handleSnapshots".equals(newValue.getKey())) {
                    repositoryRepoTypeConfig.setHandleSnapshots((boolean) newValue.getValue());
                } else if ("cdnRedirectRepoConfig".equals(newValue.getKey())) {
                    CdnRedirectRepoConfig cdnRedirectRepoConfig = new CdnRedirectRepoConfig();
                    cdnRedirectRepoConfig.setEnabled(
                            (boolean) ((Map<String, Object>) newValue.getValue()).get("enabled"));
                    repositoryRepoTypeConfig.setCdnRedirectRepoConfig(cdnRedirectRepoConfig);
                } else if ("downloadRedirect".equals(newValue.getKey())) {
                    DownloadRedirectRepoConfig downloadRedirectRepoConfig = new DownloadRedirectRepoConfig();
                    downloadRedirectRepoConfig.setEnabled(
                            (boolean) ((Map<String, Object>) newValue.getValue()).get("enabled"));
                    repositoryRepoTypeConfig.setDownloadRedirectConfig(downloadRedirectRepoConfig);
                } else if ("blackedOut".equals(newValue.getKey())) {
                    repositoryRepoTypeConfig.setBlackedOut((boolean) newValue.getValue());
                } else if ("archiveBrowsingEnabled".equals(newValue.getKey())) {
                    repositoryRepoTypeConfig.setArchiveBrowsingEnabled((boolean) newValue.getValue());
                } else if ("checksumPolicyType".equals(newValue.getKey())) {
                    repositoryRepoTypeConfig.setChecksumPolicyType(
                            LocalRepoChecksumPolicyType.fromString(
                                    (String) newValue.getValue()));
                } else if ("priorityResolution".equals(newValue.getKey())) {
                    repositoryRepoTypeConfig.setPriorityResolution((boolean) newValue.getValue());
                } else if ("propertySets".equals(newValue.getKey())) {
                    Set<String> existingPropertySetKeys = descriptor.getPropertySets().stream()
                            .map(PropertySet::getName)
                            .collect(Collectors.toSet());
                    Map<String, Object> properties = (Map<String, Object>) newValue.getValue();
                    List<String> propertiesList = properties.entrySet().stream()
                            .sorted(Map.Entry.comparingByKey())
                            .map(Map.Entry::getValue)
                            .map(String.class::cast)
                            .filter(propertySetKey -> {
                                if (!existingPropertySetKeys.contains(propertySetKey)) {
                                    log.debug("Property set '{}' does not exist in the config descriptor",
                                            propertySetKey);
                                    return false;
                                }
                                return true;
                            })
                            .toList();
                    repositoryRepoTypeConfig.setPropertySetRefs(propertiesList);
                }
            } catch (ClassCastException e) {
                throw new IllegalArgumentException("No values under " + newValue.getKey());
            }
        }
    }
}
