package org.artifactory.repo.service;

import com.google.common.annotations.VisibleForTesting;
import com.google.common.collect.Lists;
import com.google.common.collect.Sets;
import lombok.NonNull;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.builder.DiffResult;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.HaAddon;
import org.artifactory.addon.curation.CurationAddon;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.config.InternalRepositoryConfigService;
import org.artifactory.api.config.RepositoryConfigService;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.api.security.AuthorizationException;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.MutableStatusHolder;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.config.ArtifactoryConfigDiffUtil;
import org.artifactory.config.ConfigurationException;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.descriptor.config.MutableCentralConfigDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.metrics.MetricsInfo;
import org.artifactory.metrics.MetricsThreadLocal;
import org.artifactory.model.AggregatedReposConfig;
import org.artifactory.model.BasicRepoConfig;
import org.artifactory.model.FederatedRepoConfig;
import org.artifactory.model.LocalCacheRepoConfig;
import org.artifactory.model.LocalConfig;
import org.artifactory.model.LocalRepoConfig;
import org.artifactory.model.RealRepoConfig;
import org.artifactory.model.RemoteRepoConfig;
import org.artifactory.model.RepoConfig;
import org.artifactory.model.RepoConfigPropagationModel;
import org.artifactory.model.SpecialAction;
import org.artifactory.model.VirtualRepoConfig;
import org.artifactory.repo.ExtendedRepositoryConfig;
import org.artifactory.repo.ReloadableRepositoryConfigService;
import org.artifactory.repo.RepoConfigChange;
import org.artifactory.repo.RepoConfigReload;
import org.artifactory.repo.RepoDetailsType;
import org.artifactory.repo.cache.RepoConfigCache;
import org.artifactory.repo.cache.RepoConfigCacheDiffCalculator;
import org.artifactory.repo.converter.RepositoriesConfigSubConverter;
import org.artifactory.repo.converter.RepositoriesConverterManager;
import org.artifactory.repo.interceptor.RepoConfigChangeInterceptor;
import org.artifactory.repo.migration.mappers.FederatedRepoConfigMapper;
import org.artifactory.repo.migration.mappers.LocalRepoConfigMapper;
import org.artifactory.repo.migration.mappers.RemoteRepoConfigMapper;
import org.artifactory.repo.migration.mappers.VirtualRepoConfigMapper;
import org.artifactory.repo.packagetype.ExternalDependenciesConfig;
import org.artifactory.repo.packagetype.PomCleanupPolicy;
import org.artifactory.repo.packagetype.configuration.MavenConfiguration;
import org.artifactory.repo.propagation.RepoConfigPropagationModelFilter;
import org.artifactory.repo.service.validation.ArtifactoryVersionValidator;
import org.artifactory.repo.service.validation.ProjectAdminRepoConfigValidator;
import org.artifactory.repo.service.validation.RepositoryNameYamlConfigurationValidator;
import org.artifactory.repo.system.suppliers.BuildInfoRepo;
import org.artifactory.repo.virtual.VirtualRepositoryConfigResolverImpl;
import org.artifactory.rest.exception.NotFoundException;
import org.artifactory.sapi.common.ExportSettings;
import org.artifactory.sapi.common.ImportSettings;
import org.artifactory.security.SystemAuthenticationToken;
import org.artifactory.service.ConfigValidationService;
import org.artifactory.spring.Reloadable;
import org.artifactory.spring.RepoConfigChangeReloadableListener;
import org.artifactory.storage.StorageException;
import org.artifactory.storage.db.repo.entity.DbRepositoryConfig;
import org.artifactory.storage.db.repo.service.RepositoryConfigStorageService;
import org.artifactory.storage.db.repo.service.mapper.RepositoryConfigAttributesMapper;
import org.artifactory.storage.db.repo.service.mapper.RepositoryConfigAttributesMapperImpl;
import org.artifactory.storage.db.repo.service.mapper.RepositoryConfigMapper;
import org.artifactory.storage.db.repo.service.mapper.RepositoryConfigMapperImpl;
import org.artifactory.util.CollectionUtils;
import org.artifactory.util.ExternalDependenciesPackageTypeConfigMapper;
import org.artifactory.util.RepoPathUtils;
import org.artifactory.util.bundles.BundleConfigurationUtil;
import org.artifactory.version.CompoundVersionDetails;
import org.jetbrains.annotations.NotNull;
import org.jfrog.common.JsonUtils;
import org.jfrog.common.config.diff.DataDiff;
import org.slf4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.stereotype.Service;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.EnumMap;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.locks.ReentrantLock;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static java.util.stream.Collectors.groupingBy;
import static java.util.stream.Collectors.toMap;
import static java.util.stream.Collectors.toSet;
import static org.artifactory.descriptor.repo.RepoType.ReleaseBundles;
import static org.artifactory.repo.RepoConfigChange.ConfigAction.REPO_CREATE;
import static org.artifactory.repo.RepoConfigChange.ConfigAction.REPO_DELETE;
import static org.artifactory.repo.RepoConfigChange.ConfigAction.REPO_UPDATE;
import static org.artifactory.repo.RepoDetailsType.FEDERATED;
import static org.artifactory.repo.RepoDetailsType.LOCAL;
import static org.artifactory.repo.RepoDetailsType.RELEASE_BUNDLE;
import static org.artifactory.repo.RepoDetailsType.REMOTE;
import static org.artifactory.repo.RepoDetailsType.VIRTUAL;
import static org.artifactory.repo.RepoPath.REMOTE_CACHE_SUFFIX;
import static org.artifactory.repo.config.RepoConfigDefaultValues.EXAMPLE_REPO_KEY;
import static org.artifactory.repo.system.SystemRepos.isSupportBundle;
import static org.artifactory.util.RepoPathUtils.isCache;
import static org.artifactory.util.RepoPathUtils.isTrash;

/**
 * @author Tamar Rosen
 */
@Service
@Reloadable(beanClass = ReloadableRepositoryConfigService.class, initAfter = {RepositoryConfigBootstrapService.class})
public class RepositoryConfigServiceImpl
        implements InternalRepositoryConfigService, ReloadableRepositoryConfigService, RepoConfigChangesMapFunction {
    private static final String AUTH_ERROR_MESSAGE = "Only an admin user can save the artifactory configuration.";
    public static final String YML_INHERITANCE_MARKER = "»";
    public static final String YML_REPRESENTATION_FOR_MISSING_DATA = "$$deleted";
    private static final Logger log = org.slf4j.LoggerFactory.getLogger(RepositoryConfigServiceImpl.class);
    public static final String REMOTE_REPOSITORIES = "remoteRepositories";
    public static final String LOCAL_REPOSITORIES = "localRepositories";
    public static final String VIRTUAL_REPOSITORIES = "virtualRepositories";
    public static final String FEDERATED_REPOSITORIES = "federatedRepositories";
    public static final String RELEASE_BUNDLE_REPOSITORIES = "releaseBundleRepositories";
    private final RepositoryConfigStorageService storageService;
    private final Set<RepositoriesConfigSubConverter> subConverters;
    private final RepositoryConfigMapper repositoryConfigMapper;
    private final RepositoryConfigAttributesMapper repoConfigAttributesMapper;
    private Set<RepoConfigChangeReloadableListener> listeners = new HashSet<>();
    private Set<RepoConfigChangeInterceptor> interceptors = new HashSet<>();
    private ConfigValidationService configValidationService;
    private volatile RepoConfigCache repoConfigCache;
    private final ReentrantLock repoCacheCreateLock = new ReentrantLock();

    @Autowired
    public RepositoryConfigServiceImpl(RepositoryConfigStorageService storageService,
            Set<RepositoriesConfigSubConverter> subConverters) {
        this.storageService = storageService;
        this.subConverters = subConverters;
        this.repositoryConfigMapper = new RepositoryConfigMapperImpl();
        this.repoConfigAttributesMapper = new RepositoryConfigAttributesMapperImpl();
        createRepoConfigCache();
    }

    @Override
    public void onBeforeContextCreated() {
        if (countAll() == 0) {
            createDefaultRepoConfigsIfDoesNotExist();
        }
    }

    @Override
    public void init() {
        listeners =
                new HashSet<>(ContextHelper.get().beansForType(RepoConfigChangeReloadableListener.class).values());
        interceptors = new HashSet<>(ContextHelper.get().beansForType(RepoConfigChangeInterceptor.class).values());
    }

    private RepoConfigCache createRepoConfigCache() {
        RepoConfigCache oldCache = this.repoConfigCache;
        try {
            log.debug("Creating repo cache. Acquire lock...");
            repoCacheCreateLock.lock();
            log.debug("Lock has been acquired. Creating repo cache instance...");
            this.repoConfigCache = new RepoConfigCache(
                    ConstantValues.repoConfigCacheSyncUpMillis.getLong(),
                    ConstantValues.repoConfigExpirationMillis.getLong(),
                    ConstantValues.repoConfigDbBatchSize.getInt(),
                    Double.parseDouble(ConstantValues.repoConfigCacheGetAllOptimizationRatio.getString()),
                    this::getAllRepoConfigsWithoutCache,
                    this::getRepoByKeyWithoutCache,
                    this::getRepoConfigsByKeysWithoutCache,
                    this::callOnRepoConfigReloadAsync,
                    this::getAllBasicRepoConfigs);
            if (oldCache != null) {
                oldCache.destroy();
            }
        } finally {
            log.debug("Repo cache has been created. Releasing lock...");
            repoCacheCreateLock.unlock();
            log.debug("Lock has been released");
        }
        return oldCache;
    }

    @Override
    public void destroy() {
        repoConfigCache.destroy();
    }

    @Autowired
    public void setConfigValidationService(ConfigValidationService configValidationService) {
        this.configValidationService = configValidationService;
    }

    @Override
    public List<RepoConfig> findRepositoriesByType(RepoDetailsType repoType) {
        boolean isBundlePackageTypeEnabled = BundleConfigurationUtil.isReleaseBundleHandlerEnabled();
        Set<BasicRepoConfig> allBasicRepoConfigs = repoConfigCache.getAllBasicConfigs();
        Set<String> filteredRepoKeys;
        if (isBundlePackageTypeEnabled || RELEASE_BUNDLE == repoType) {
            filteredRepoKeys = allBasicRepoConfigs.stream()
                    .filter(repoConfig -> repoType.equals(repoConfig.getRepoType()))
                    .map(BasicRepoConfig::getRepoKey)
                    .collect(toSet());
        } else {
            // in case release bundles handler is not enabled and the repository type is not release bundle v1,
            // then filter out all the repositories with package type "ReleaseBundles".
            filteredRepoKeys = allBasicRepoConfigs.stream()
                    .filter(repoConfig -> repoType.equals(repoConfig.getRepoType()))
                    .filter(repo -> ReleaseBundles != repo.getPackageType())
                    .map(BasicRepoConfig::getRepoKey)
                    .collect(toSet());
        }
        return new ArrayList<>(repoConfigCache.getRepositoryConfigsByKeys(filteredRepoKeys));
    }

    @Override
    public List<RepoConfig> findRepositoriesByPackageType(RepoType packageType) {
        Set<BasicRepoConfig> allBasicRepoConfigs = repoConfigCache.getAllBasicConfigs();
        Set<String> filteredRepoKeys = allBasicRepoConfigs.stream()
                .filter(repoConfig -> packageType.equals(repoConfig.getPackageType()))
                .map(BasicRepoConfig::getRepoKey)
                .collect(toSet());
        return new ArrayList<>(repoConfigCache.getRepositoryConfigsByKeys(filteredRepoKeys));
    }

    @Override
    public List<RepoConfig> findRepositoriesByTypeAndPackageType(Set<RepoDetailsType> type, RepoType packageType) {
        return new ArrayList<>(
                repoConfigCache.getRepositoryConfigsByKeys(getRepoKeysByTypeAndPackageType(type, packageType)));
    }

    @Override
    public Set<String> getRepoKeysByTypeAndPackageType(Set<RepoDetailsType> type, RepoType packageType) {
        Set<BasicRepoConfig> allBasicRepoConfigs = repoConfigCache.getAllBasicConfigs();
        return allBasicRepoConfigs.stream()
                .filter(repoConfig -> packageType.equals(repoConfig.getPackageType()) &&
                        type.contains(repoConfig.getRepoType()))
                .map(BasicRepoConfig::getRepoKey)
                .collect(toSet());
    }

    @Override
    public void insertRepository(RepoConfig repositoryConfig) {
        if (repositoryConfig == null) {
            log.warn("Cannot insert repository: repository config cannot be null");
            log.debug("Cannot insert repository: repository config cannot be null", new NullPointerException());
            return;
        }
        if (ContextHelper.get().isReposImportInProgress()) {
            log.info("Cannot insert repository config '{}', import is still in progress", repositoryConfig.getKey());
            return;
        }
        validate(Set.of(repositoryConfig));
        if (repositoryConfig instanceof RemoteRepoConfig remoteRepoConfig) {
            insertRemoteRepoConfig(remoteRepoConfig);
            return;
        }
        insertRepositoryInternal(repositoryConfig);
    }

    @Override
    public void upsertRepository(RepoConfig repositoryConfig) {
        if (repositoryConfig == null) {
            log.warn("Cannot upsert repository: repository config cannot be null");
            log.debug("Cannot upsert repository: repository config cannot be null", new NullPointerException());
            return;
        }
        if (ContextHelper.get().isReposImportInProgress()) {
            log.info("Cannot upsert repository config '{}', import is still in progress",
                    repositoryConfig.getKey());
            return;
        }
        validate(Set.of(repositoryConfig));
        RepoConfig oldRepoConfig = getRepoConfigByKey(repositoryConfig.getKey());
        if (repositoryConfig instanceof RemoteRepoConfig remoteRepoConfig) {
            assertCurationIsValid(remoteRepoConfig);
            if (oldRepoConfig == null) {
                insertRemoteRepoConfig(remoteRepoConfig);
            } else {
                updateRemoteRepoConfig(remoteRepoConfig);
            }
            return;
        }
        RepoConfigChange repoConfigChange =
                oldRepoConfig == null ? getRepoConfigCreateChange(repositoryConfig) :
                        getRepoConfigUpdateChange(repositoryConfig, oldRepoConfig);
        flushRepoConfigChange(repositoryConfig, repoConfigChange);
    }

    private void assertCurationIsValid(RemoteRepoConfig config) {
        if (!isCurationConfigValid(config)) {
            throw new IllegalArgumentException(String.format(
                    "Cannot upsert '%s' repo due to invalid curation settings", config.getKey()));
        }
    }

    private boolean isCurationConfigValid(RemoteRepoConfig config) {
        return ContextHelper.get().beanForType(AddonsManager.class).addonByType(CurationAddon.class)
                .isCurationConfigValid(config);
    }

    @Override
    public void updateRepository(RepoConfig repositoryConfig) {
        if (repositoryConfig == null) {
            log.warn("Cannot update repository: repository config cannot be null");
            log.debug("Cannot update repository: repository config cannot be null", new NullPointerException());
            return;
        }
        if (ContextHelper.get().isReposImportInProgress()) {
            log.info("Cannot update repository config '{}', import is still in progress", repositoryConfig.getKey());
            return;
        }
        validate(Set.of(repositoryConfig));
        if (repositoryConfig instanceof RemoteRepoConfig remoteRepoConfig) {
            updateRemoteRepoConfig(remoteRepoConfig);
            return;
        }
        updateRepositoryInternal(repositoryConfig);
    }

    @Override
    public RepoConfig deleteRepository(String repoKey) {
        if (StringUtils.isBlank(repoKey)) {
            log.warn("Cannot delete repository: repository key cannot be null");
            log.debug("Cannot delete repository: repository key cannot be null", new NullPointerException());
            return null;
        }

        if (ContextHelper.get().isReposImportInProgress()) {
            log.info("Cannot delete repository config '{}', import is still in progress", repoKey);
            return null;
        }
        if (isInvalid(repoKey)) {
            return null;
        }
        // TODO[vovav]: move this code into interceptor
        if (!ContextHelper.get().beanForType(AuthorizationService.class).isAnyProjectAdmin()) {
            throw new AuthorizationException(AUTH_ERROR_MESSAGE);
        }
        RepoConfig oldRepoConfig = getRepoConfigByKey(repoKey);
        if (oldRepoConfig == null) {
            log.debug("Cannot delete repository '{}': repository config does not exist.", repoKey);
            return null;
        }
        bulkDeleteRepositoryConfigs(List.of(repoKey), true, true);
        return oldRepoConfig;
    }

    @Override
    public void bulkInsertRepositoryConfigs(List<? extends RepoConfig> repositoryConfigs,
            boolean reload, boolean propagateChanges, boolean doBackups) {
        if (CollectionUtils.isNullOrEmpty(repositoryConfigs)) {
            return;
        }
        if (ContextHelper.get().isReposImportInProgress()) {
            log.info("Cannot insert repository configs, import is still in progress");
            return;
        }
        validate(repositoryConfigs);
        Collection<RepoConfig> existingRepos = getExistingRepoConfigs(repositoryConfigs, false);
        Map<RepoDetailsType, List<RepoConfigChange>> typeToChanges =
                convertRepoConfigsToCreateUpdateChanges(repositoryConfigs, existingRepos);
        interceptors.forEach(listener -> listener.onBeforeSave(typeToChanges));
        if (doBackups) {
            setBackups(repositoryConfigs.stream()
                    .filter(repoConfig -> RepoDetailsType.REMOTE.equals(repoConfig.getRepoType()))
                    .map(RemoteRepoConfig.class::cast)
                    .map(RemoteRepoConfig::getKey)
                    .toList());
        }
        Set<DbRepositoryConfig> reposToCreate = typeToChanges.values().stream()
                .flatMap(Collection::stream)
                .filter(change -> REPO_CREATE.equals(change.getConfigAction()))
                .map(change -> {
                    change.getNewRepoConfig().setModification(System.currentTimeMillis());
                    return repositoryConfigMapper.repositoryConfigToDbRepositoryConfig(change.getNewRepoConfig());
                })
                .collect(Collectors.toSet());
        storageService.bulkInsertRepositoryConfigs(reposToCreate);
        if (!reposToCreate.isEmpty()) {
            repoConfigCache.addRepoConfigKeys(dbInfoBlobListToRepoKeyAndModificationMap(reposToCreate));
            //Map<RepoDetailsType, List<RepoConfigReload>> repoConfigReloads =
            if (propagateChanges) {
                propagateCreateRepoConfig(typeToChanges);
            }
        }
        if (reload) {
            callOnRepoConfigReload(convertToRepoConfigReloadByType(typeToChanges));
            interceptors.forEach(interceptor -> interceptor.onAfterSave(typeToChanges));
        }
    }

    @Override
    public void bulkUpsertRepositoryConfigs(Collection<? extends RepoConfig> repositoryConfigs, boolean reload) {
        bulkUpsertRepositoryConfigs(repositoryConfigs, reload, false);
    }

    private void bulkUpsertRepositoryConfigs(Collection<? extends RepoConfig> repositoryConfigs, boolean reload,
            boolean isImport) {
        if (repositoryConfigs.isEmpty()) {
            return;
        }
        if (!isImport && ContextHelper.get().isReposImportInProgress()) {
            log.info("Cannot upsert repository configs, import is still in progress");
            return;
        }
        validate(repositoryConfigs);
        Collection<RepoConfig> existingRepos = getExistingRepoConfigs(repositoryConfigs, isImport);
        Map<RepoDetailsType, List<RepoConfigChange>> mutableRepoConfigChangesByType =
                convertRepoConfigsToCreateUpdateChanges(repositoryConfigs, existingRepos);
        interceptors.forEach(listener -> listener.onBeforeSave(mutableRepoConfigChangesByType));
        flushRepoConfigChanges(mutableRepoConfigChangesByType, true);
        if (reload) {
            callOnRepoConfigReload(convertToRepoConfigReloadByType(mutableRepoConfigChangesByType));
            interceptors.forEach(interceptor -> interceptor.onAfterSave(mutableRepoConfigChangesByType));
        }
    }

    @Override
    public void bulkDeleteRepositoryConfigs(List<String> repositoryConfigKeys, boolean reload,
            boolean propagateChanges) {
        bulkDeleteRepositoryConfigs(repositoryConfigKeys, reload, propagateChanges, false);
    }

    private void bulkDeleteRepositoryConfigs(List<String> repositoryConfigKeys, boolean reload,
            boolean propagateChanges, boolean isImport) {
        List<RepoConfig> repoConfigs = new ArrayList<>();
        for (String repoKey : repositoryConfigKeys) {
            if (isInvalid(repoKey)) {
                continue;
            }
            RepoConfig oldRepoConfig = getRepoConfigByKey(repoKey, isImport);
            if (oldRepoConfig != null) {
                repoConfigs.add(oldRepoConfig);
            }
        }
        bulkDeleteRepositoryConfigsInternal(repoConfigs, reload, propagateChanges, isImport);
    }

    private void bulkDeleteRepositoryConfigsInternal(Collection<RepoConfig> repoConfigs, boolean reload,
            boolean propagateChanges, boolean isImport) {
        if (!isImport && ContextHelper.get().isReposImportInProgress()) {
            log.info("Cannot delete repository configs, import is still in progress");
            return;
        }
        if (repoConfigs.isEmpty()) {
            return;
        }
        validate(repoConfigs);
        List<RepoConfigChange> repoConfigChanges = new ArrayList<>();
        for (RepoConfig oldRepoConfig : repoConfigs) {
            if (oldRepoConfig != null) {
                RepoConfigChange repoConfigChange = getRepoConfigDeleteChange(oldRepoConfig);
                repoConfigChanges.add(repoConfigChange);
            }
        }

        List<VirtualRepoConfig> virtualRepoConfigs =
                deleteReferenceFromVirtualRepos(repoConfigs.stream().toList(), isImport);
        if (!virtualRepoConfigs.isEmpty()) {
            Collection<RepoConfig> existingRepos = getExistingRepoConfigs(virtualRepoConfigs, isImport);
            Map<RepoDetailsType, List<RepoConfigChange>> changedVirtualRepoConfigs =
                    convertRepoConfigsToUpdateChanges(virtualRepoConfigs, existingRepos);
            if (changedVirtualRepoConfigs.get(VIRTUAL) != null) {
                repoConfigChanges.addAll(changedVirtualRepoConfigs.get(VIRTUAL));
            }
        }

        Map<RepoDetailsType, List<RepoConfigChange>> mutableRepoConfigChangesByType = repoConfigChanges.stream()
                .filter(change -> change.getOldRepoConfig() != null)
                .collect(groupingBy(change -> change.getOldRepoConfig().getRepoType()));
        interceptors.forEach(listener -> listener.onBeforeSave(mutableRepoConfigChangesByType));
        if (reload) {
            // Force access to the repositories cache to avoid a collapsing effect of any previous changes with new
            // changes during the reload flow
            String repoKey = "DUMMY_REPO";
            listeners.stream()
                    .filter(InternalRepositoryService.class::isInstance)
                    .findFirst()
                    .map(InternalRepositoryService.class::cast)
                    .ifPresent(repoService -> repoService.repositoryByKey(repoKey));
        }
        dereferenceRepositoryFromConfigDescriptor(repoConfigs.stream().toList());
        flushRepoConfigChanges(mutableRepoConfigChangesByType, propagateChanges);
        if (reload) {
            callOnRepoConfigReload(convertToRepoConfigReloadByType(mutableRepoConfigChangesByType));
            interceptors.forEach(interceptor -> interceptor.onAfterSave(mutableRepoConfigChangesByType));
        }
    }

    private Map<RepoDetailsType, List<RepoConfigReload>> convertToRepoConfigReloadByType(
            Map<RepoDetailsType, List<RepoConfigChange>> repoConfigChangesByType) {
        Map<RepoDetailsType, List<RepoConfigReload>> repoConfigReloadsByType = new HashMap<>();
        repoConfigChangesByType.forEach((type, repoConfigChanges) -> repoConfigReloadsByType.put(type,
                repoConfigChanges.stream().map(this::repoConfigChangeToRepoConfigReload).toList()));
        return repoConfigReloadsByType;
    }

    private RepoConfigReload repoConfigChangeToRepoConfigReload(RepoConfigChange repoConfigChange) {
        RepoConfig repoConfig =
                repoConfigChange.getConfigAction() == REPO_DELETE ? repoConfigChange.getOldRepoConfig() :
                        repoConfigChange.getNewRepoConfig();
        RepoConfigReload repoConfigReload = new RepoConfigReload(repoConfigChange.getRepoKey(),
                repoConfig.getRepoType(),
                repoConfig.getPackageType(),
                repoConfigChange.getConfigAction(),
                repoConfig.getModification());
        if (repoConfigChange.getConfigAction() == REPO_UPDATE) {
            if (repoConfigChange.getNewRepoConfig().getRepoType() == VIRTUAL &&
                    isPomCleanUpPolicyChanged(repoConfigChange.getNewRepoConfig(),
                            repoConfigChange.getOldRepoConfig())) {
                repoConfigReload.setSpecialAction(SpecialAction.PERFORM_POM_CLEANUP_POLICY);
            }
            if (repoConfigChange.getNewRepoConfig().getRepoType() == REMOTE &&
                    isRemoteRepoConfigChanged(
                            (RemoteRepoConfig) repoConfigChange.getNewRepoConfig(),
                            (RemoteRepoConfig) repoConfigChange.getOldRepoConfig())) {
                repoConfigReload.setSpecialAction(SpecialAction.PERFORM_CONFIGURATION_REPLICATOR);
            }
        }
        return repoConfigReload;
    }

    @Override
    public void insertPackageConfigByKey(String repoKey, Map<String, String> packageConfig) {
        if (ContextHelper.get().isReposImportInProgress()) {
            log.info("Cannot update repository config '{}', import is still in progress", repoKey);
            return;
        }
        RepoConfig repositoryConfig = repoConfigCache.getByKey(repoKey);
        RepoConfig oldRepoConfig = getRepoConfigByKey(repositoryConfig.getKey());
        validate(Set.of(oldRepoConfig));
        packageConfig.forEach((k, v) -> repositoryConfig.getPackageTypeConfig().put(k, v));
        RepoConfigChange repoConfigChange =
                getRepoConfigUpdateChange(repositoryConfig, oldRepoConfig);
        flushRepoConfigChange(repositoryConfig, repoConfigChange);
    }

    @Override
    public RepoConfig getRepoConfigByKey(String repoKey) {
        return getRepoConfigByKey(repoKey, false);
    }

    private RepoConfig getRepoConfigByKey(String repoKey, boolean isImport) {
        if (StringUtils.isBlank(repoKey)) {
            return null;
        }
        if (isTrash(repoKey) || isSupportBundle(repoKey) || isCache(repoKey)) {
            throw new IllegalArgumentException(String.format(
                    "Cannot retrieve '%s' repo config from the RepositoryConfigService. Please use the " +
                            "RepositoryService instead.", repoKey));
        }
        return isImport
                ? dbInfoBlobToRepositoryConfig(storageService.findRepositoryByKey(repoKey))
                : repoConfigCache.getByKey(repoKey);
    }

    @Override
    public RealRepoConfig getRealRepoConfigByKey(String repoKey) {
        if (isInvalid(repoKey)) {
            return null;
        }
        RepoConfig repoConfig = getRepoConfigByKey(repoKey);
        if (repoConfig == null) {
            return null;
        }
        if (repoConfig.isReal()) {
            return (RealRepoConfig) repoConfig;
        }
        return null;
    }

    @Override
    public VirtualRepoConfig getVirtualRepoConfigByKey(String repoKey) {
        if (isInvalid(repoKey)) {
            return null;
        }
        RepoConfig repoConfig = getRepoConfigByKey(repoKey);
        if (repoConfig != null && RepoDetailsType.VIRTUAL.equals(repoConfig.getRepoType())) {

            return (VirtualRepoConfig) repoConfig;
        }
        return null;
    }

    @Override
    public RemoteRepoConfig getRemoteRepoConfigByKey(String repoKey) {
        if (isInvalid(repoKey)) {
            return null;
        }
        RepoConfig repoConfig = getRepoConfigByKey(repoKey);
        if (repoConfig != null && REMOTE.equals(repoConfig.getRepoType())) {
            return (RemoteRepoConfig) repoConfig;
        }
        return null;
    }

    @Override
    public boolean exists(String repoKey) {
        return repoConfigCache.existByKey(repoKey);
    }

    @Override
    public boolean existsRepositoryIgnoreCase(String repoKey) {
        return repoConfigCache.getAllRepoKeys().stream()
                .anyMatch(key -> key.compareToIgnoreCase(repoKey) == 0);
    }

    @Override
    public Collection<RepoConfig> getAllRepositoryConfigs() {
        Collection<RepoConfig> allRepoConfigs = repoConfigCache.getAll();
        // in case release bundles handler is not enabled and the repository type is not release bundle v1,
        // then filter out all the repositories with package type "ReleaseBundles".
        return BundleConfigurationUtil.isReleaseBundleHandlerEnabled() ? allRepoConfigs
                : allRepoConfigs.stream()
                .filter(repo -> RELEASE_BUNDLE == repo.getRepoType() || ReleaseBundles != repo.getPackageType())
                .toList();
    }

    @Override
    public Collection<String> getAllRepositoryKeys() {
        return repoConfigCache.getAllRepoKeys();
    }

    @Override
    public Set<RepoConfig> findRepositoryConfigsByKeys(Set<String> repoKeys) {
        if (repoKeys.isEmpty()) {
            log.warn(
                    "Attempt to search repository by keys, but no repository keys were provided. Returning empty list" +
                            "...");
            return Set.of();
        }
        Set<String> copyOfRequestedKeys = new HashSet<>(repoKeys);
        copyOfRequestedKeys.retainAll(repoConfigCache.getAllRepoKeys());
        if (copyOfRequestedKeys.isEmpty()) {
            return Set.of();
        }
        return new HashSet<>(repoConfigCache.getRepositoryConfigsByKeys(copyOfRequestedKeys));
    }

    @Override
    public LinkedHashSet<RepoConfig> findRepositoryConfigsByKeysInSpecificOrder(VirtualRepoConfig virtualRepoConfig) {
        LinkedHashSet<String> repositoryRefs =
                new LinkedHashSet<>(virtualRepoConfig.getRepoTypeConfig().getRepositoryRefs());
        final List<String> order = virtualRepoConfig.getRepoTypeConfig().getRepositoryRefs();
        if (repositoryRefs.isEmpty()) {
            log.info(
                    "Attempt to search aggregated repositories, but virtual repo '{}' does not contain any. Returning" +
                            " empty list...", virtualRepoConfig.getKey());
            return Sets.newLinkedHashSet();
        }
        Set<RepoConfig> repositoryConfigsByKeys = findRepositoryConfigsByKeys(repositoryRefs);
        return repositoryConfigsByKeys.stream()
                .sorted(Comparator.comparing((RepoConfig config) -> order.indexOf(config.getKey()))
                        .thenComparing(RepoConfig::getKey))
                .collect(Collectors.toCollection(LinkedHashSet::new));
    }

    @Override
    public <T> ExtendedRepositoryConfig<T> getExtendedRepositoryByKey(String repoKey, Class<T> type) {
        RepoConfig repo = getRepoConfigByKey(repoKey);
        return new ExtendedRepositoryConfig<>(repo, getDeserializedType(type, repo));
    }

    @Override
    public Map<String, String> getPackageConfigByRepoKey(String repoKey) {
        RepoConfig repoConfig = repoConfigCache.getByKey(repoKey);
        if (repoConfig == null) {
            return Map.of();
        }
        return repoConfig.getPackageTypeConfig();
    }

    @Override
    public LocalConfig getLocalOrFederatedRepoConfigByKey(String repoKey) {
        if (isInvalid(repoKey)) {
            return null;
        }
        var repoConfig = getRepoConfigByKey(repoKey);
        if (repoConfig == null) {
            return null;
        }
        if (repoConfig instanceof LocalConfig localConfig) {
            return localConfig;
        }
        log.debug("No matching local repository. Found repository is: {}", repoConfig.getClass().getSimpleName());
        return null;
    }

    @Override
    public FederatedRepoConfig getFederatedRepoConfigByKey(String repoKey) {
        if (isInvalid(repoKey)) {
            return null;
        }
        var repoConfig = getRepoConfigByKey(repoKey);
        if (repoConfig == null) {
            return null;
        }
        if (repoConfig instanceof FederatedRepoConfig federatedRepoConfig) {
            return federatedRepoConfig;
        }
        log.debug("No matching federated repository. Found repository is: {}", repoConfig.getClass().getSimpleName());
        return null;
    }

    @Override
    public <T> T getPackageConfigByRepoKey(String repoKey, Class<T> packageTypeModel) {
        RepoConfig repoConfig = repoConfigCache.getByKey(repoKey);
        if (repoConfig == null) {
            return null;
        }
        return getDeserializedType(packageTypeModel, repoConfig);
    }

    @NotNull
    public static <T> T getDeserializedType(Class<T> packageTypeModel, RepoConfig repositoryConfig) {
        JsonUtils jsonMapper = JsonUtils.getInstance();
        byte[] packageTypeConfigByKey =
                jsonMapper.valueToByteArray(repositoryConfig.getPackageTypeConfig());
        return jsonMapper.readValue(packageTypeConfigByKey, packageTypeModel);
    }


    @Override
    public void convert(CompoundVersionDetails source, CompoundVersionDetails target) {
        // Raw - do not deserialize
        Set<byte[]> repoInfoBlobs =
                storageService.getAllRepositories().stream().map(DbRepositoryConfig::getRepositoryInfoBlob)
                        .collect(toSet());

        RepositoriesConverterManager repositoriesConverterManager = new RepositoriesConverterManager(subConverters);

        Set<DbRepositoryConfig> convertedDbRepos = repoInfoBlobs.stream().map(repositoriesConverterManager::convert)
                .map(repoConfig -> {
                    repoConfig.setModification(System.currentTimeMillis());
                    return repositoryConfigMapper.repositoryConfigToDbRepositoryConfig(repoConfig);
                })
                .collect(Collectors.toSet());

        storageService.bulkUpdateRepositoryConfigs(convertedDbRepos);
        createRepoConfigCache();
    }

    @Override
    public void callOnRepoConfigReload(Map<RepoDetailsType, List<RepoConfigReload>> typeToChanges) {
        log.info("Calling onRepoConfigChange...");
        if (log.isDebugEnabled()) {
            typeToChanges.values().stream()
                    .flatMap(Collection::stream)
                    .forEach(repo -> log.debug("Changed repo: {}, action: {}", repo.getRepoKey(),
                            repo.getConfigAction()));
        }
        ContextHelper.get().beanForType(InternalRepositoryService.class).onRepoConfigReload(typeToChanges);
        listeners.stream()
                .filter(listener -> !(listener instanceof InternalRepositoryService))
                .forEach(listener -> listener.onRepoConfigReload(typeToChanges));
    }

    @Override
    public void callOnRepoConfigReloadAsync(Map<RepoDetailsType, List<RepoConfigReload>> typeToChanges) {
        callOnRepoConfigReload(typeToChanges);
    }

    @Override
    public int countAll() {
        return repoConfigCache.countAll();
    }

    @Override
    public int deleteAll(boolean reload, boolean propagateChanges) {
        Collection<RepoConfig> repoConfigs = getAllRepositoryConfigs();
        bulkDeleteRepositoryConfigsInternal(repoConfigs, reload, propagateChanges, false);
        return repoConfigs.size();
    }

    @Override
    public List<RepoConfig> getRepoKeysByProjectAndType(Function<String, Boolean> projectFilter,
            Set<RepoDetailsType> types) {
        return getRepoConfigsByTypes(types)
                .stream()
                .filter(repo -> projectFilter.apply(RepoPathUtils.removeDashCache(repo.getKey())))
                .toList();
    }

    @Override
    public void reloadRepoConfigurations() {
        RepoConfigCache oldCache = createRepoConfigCache();
        ContextHelper.get().beanForType(RepositoryConfigService.class)
                .calculateCacheDiffAndReloadAsync(oldCache.getCachedBasicRepoConfigs(),
                        repoConfigCache.getCachedBasicRepoConfigs());
    }

    @Override
    public void calculateCacheDiffAndReloadAsync(Map<String, BasicRepoConfig> oldBasicRepoConfigs,
            Map<String, BasicRepoConfig> newBasicRepoConfigs) {
        reloadRepoConfigurations(
                RepoConfigCacheDiffCalculator.calculateCacheDiff(oldBasicRepoConfigs, newBasicRepoConfigs));
    }

    @Override
    public void repoConfigSaveNewData(Set<DataDiff<?>> newDataDiffs) {
        Map<String, Object> migratedDiffs = new HashMap<>();
        newDataDiffs.forEach(e -> {
            checkNoEmptyRepoConfigIsPassed(e);
            migratedDiffs.put(e.getFieldName(), e.getNewValue());

        });
        List<String> repoConfigDataByKeys = new ArrayList<>(migratedDiffs.keySet());
        Collections.sort(repoConfigDataByKeys);
        String key = "";
        Map<String, RepoDetailsType> duplicateKey = new HashMap<>();
        Map<String, Map<String, Object>> main = new HashMap<>();
        Map<String, Object> second = new HashMap<>();
        Map<String, RepoConfig> repoConfigs = new HashMap<>();
        String type = "";
        for (String repoConfigDataByKey : repoConfigDataByKeys) {
            String[] split = repoConfigDataByKey.split(YML_INHERITANCE_MARKER);
            checkForDuplicate(split[1], duplicateKey, split[0]);
            if (key.equals("")) {
                type = split[0];
                key = split[1];
            } else if (!key.equals(split[1])) {
                main.put(key, second);
                addRepoConfigByType(repoConfigs, key, type, main.get(key).get("type"));
                key = split[1];
                type = split[0];
                second = new HashMap<>();
            }
            if (split.length == 3) {
                second.put(split[2], migratedDiffs.get(repoConfigDataByKey));
            } else if (split.length == 4) {
                if (second.get(split[2]) != null) {
                    ((Map<String, Object>) second.get(split[2])).put(split[3], migratedDiffs.get(repoConfigDataByKey));
                } else {
                    Map<String, Object> inner = new HashMap<>();
                    inner.put(split[3], migratedDiffs.get(repoConfigDataByKey));
                    second.put(split[2], inner);
                }
            } else if (split.length == 5) {
                if (second.get(split[2]) != null) {
                    Map<String, Object> thirdLvlMap = (Map<String, Object>) second.get(split[2]);
                    if (thirdLvlMap.get(split[3]) != null) {
                        ((Map<String, Object>) thirdLvlMap.get(split[3])).put(split[4],
                                migratedDiffs.get(repoConfigDataByKey));
                    } else {
                        HashMap<String, Object> fourthLvlMap = new HashMap<>();
                        fourthLvlMap.put(split[4], migratedDiffs.get(repoConfigDataByKey));
                        thirdLvlMap.put(split[3], fourthLvlMap);
                    }
                } else {
                    Map<String, Object> thirdLvlMap = new HashMap<>();
                    HashMap<String, Object> fourthLvlMap = new HashMap<>();
                    fourthLvlMap.put(split[4], migratedDiffs.get(repoConfigDataByKey));
                    thirdLvlMap.put(split[3], fourthLvlMap);
                    second.put(split[2], thirdLvlMap);
                }
            }
        }
        main.put(key, second);
        addRepoConfigByType(repoConfigs, key, type, main.get(key).get("type"));
        validateRepoNames(main);
        Set<RepoConfig> repositoryConfigsByKeys = findRepositoryConfigsByKeys(main.keySet());
        List<RepoConfig> repoConfigsToUpsert = new ArrayList<>();
        List<String> repoConfigsToDelete = new ArrayList<>();
        if (!repositoryConfigsByKeys.isEmpty()) {
            repositoryConfigsByKeys.forEach(
                    dbRepoConfig -> {
                        checkForDuplicate(dbRepoConfig.getKey(), duplicateKey, dbRepoConfig.getRepoType());
                        Map<String, Object> newData = main.remove(dbRepoConfig.getKey());
                        if (newData.entrySet().size() == 1 &&
                                YML_REPRESENTATION_FOR_MISSING_DATA.equals(
                                        newData.entrySet().stream().findFirst().get().getKey())) {
                            repoConfigsToDelete.add(dbRepoConfig.getKey());
                        } else {
                            createRepoByRepoType(dbRepoConfig, newData).ifPresent(repoConfigsToUpsert::add);
                        }
                    });
        }
        main.keySet().forEach(e -> {
            if (repoConfigs.get(e) != null) {
                if (main.get(e).size() == 1 &&
                        YML_REPRESENTATION_FOR_MISSING_DATA.equals(
                                main.get(e).entrySet().stream().findFirst().get().getKey())) {
                    //do nothing
                } else {
                    createRepoByRepoType(repoConfigs.get(e), main.get(e)).ifPresent(repoConfigsToUpsert::add);
                }
            }
        });
        bulkUpsertRepositoryConfigs(repoConfigsToUpsert, true);
        bulkDeleteRepositoryConfigs(repoConfigsToDelete, true, true);
    }

    public Set<String> findAllMissingRepoRootPaths() {
        return storageService.findAllMissingRepoRootPaths();
    }

    private void checkForDuplicate(String key, Map<String, RepoDetailsType> duplicateName, RepoDetailsType type) {
        if (!duplicateName.computeIfAbsent(key, k -> type).equals(type)) {
            log.error("duplicate repo key {} for repo types {} {} ", key, duplicateName.get(key), type);
            throw new StorageException(
                    "duplicate repo key - " + key + " for repo types " + duplicateName.get(key) + " and " + type);
        }
    }

    private void checkForDuplicate(String key, Map<String, RepoDetailsType> duplicateName, String type) {
        if (type.startsWith(LOCAL_REPOSITORIES)) {
            checkForDuplicate(key, duplicateName, LOCAL);
        } else if (type.startsWith(VIRTUAL_REPOSITORIES)) {
            checkForDuplicate(key, duplicateName, VIRTUAL);
        } else if (type.startsWith(FEDERATED_REPOSITORIES)) {
            checkForDuplicate(key, duplicateName, FEDERATED);
        } else if (type.startsWith(REMOTE_REPOSITORIES)) {
            checkForDuplicate(key, duplicateName, REMOTE);
        } else if (type.startsWith(RELEASE_BUNDLE_REPOSITORIES)) {
            checkForDuplicate(key, duplicateName, RELEASE_BUNDLE);
        } else {
            log.error("Repo details type - {} , for key - {} does not exist in the possible RepoDetailsType", type,
                    key);
            throw new ConfigurationException(
                    "Repo details type - " + type + " , for key - " + key +
                            " does not exist in the possible RepoDetailsType");
        }
    }

    private void validateRepoNames(Map<String, Map<String, Object>> main) {
        main.forEach((key, value) -> RepositoryNameYamlConfigurationValidator.validateRepoName(
                key, objectToRepoType(value.get("type")).getType()));
    }

    private void checkNoEmptyRepoConfigIsPassed(DataDiff<?> e) {
        if (e.getFieldName().split(YML_INHERITANCE_MARKER).length == 2 && e.getNewValue() == null) {
            throw new IllegalArgumentException("No values under " + e.getFieldName().split(
                    YML_INHERITANCE_MARKER)[0]);
        }
    }

    private Optional<RepoConfig> createRepoByRepoType(RepoConfig repoConfig, Map<String, Object> stringObjectMap) {
        CentralConfigDescriptor descriptor = ContextHelper.get().getCentralConfig().getDescriptor();
        if (LOCAL.equals(repoConfig.getRepoType())) {
            return Optional.of(LocalRepoConfigMapper.mapLocalRepoConfig(
                    (LocalRepoConfig) repoConfig, stringObjectMap, descriptor));
        } else if (REMOTE.equals(repoConfig.getRepoType())) {
            return Optional.of(RemoteRepoConfigMapper.mapRemoteRepoConfig(
                    (RemoteRepoConfig) repoConfig, stringObjectMap, descriptor));
        } else if (VIRTUAL.equals(repoConfig.getRepoType())) {
            return Optional.of(VirtualRepoConfigMapper.mapVirtualRepoConfig(
                    (VirtualRepoConfig) repoConfig, stringObjectMap, descriptor));
        } else if (FEDERATED.equals(repoConfig.getRepoType())) {
            return Optional.of(FederatedRepoConfigMapper.mapFederatedRepoConfig(
                    (FederatedRepoConfig) repoConfig, stringObjectMap, descriptor));
        }
        return Optional.empty();
    }

    private void addRepoConfigByType(Map<String, RepoConfig> repoConfigs, String key, String type, Object
            possibleRepoType) {
        try {
            RepoType repoType = objectToRepoType(possibleRepoType);
            if (type.startsWith(LOCAL_REPOSITORIES)) {
                repoConfigs.put(key, LocalRepoConfig.create(key, repoType));
            } else if (type.startsWith(REMOTE_REPOSITORIES)) {
                repoConfigs.put(key, RemoteRepoConfig.create(key, repoType));
            } else if (type.startsWith(VIRTUAL_REPOSITORIES)) {
                repoConfigs.put(key, VirtualRepoConfig.create(key, repoType));
            } else if (type.startsWith(FEDERATED_REPOSITORIES)) {
                repoConfigs.put(key, FederatedRepoConfig.create(key, repoType));
            }
        } catch (ClassCastException e) {
            throw new IllegalArgumentException("No values under type");
        }
    }

    private RepoType objectToRepoType(Object repoType) {
        return repoType != null ? RepoType.fromType((String) repoType) : RepoType.Generic;
    }

    @Override
    public void reloadRepoConfigurations(RepoConfigPropagationModel model) {
        if (model == null || model.isEmpty()) {
            log.debug("Reload repo configuration: nothing to handle");
            return;
        }
        RepoConfigPropagationModelFilter repoConfigPropagationModelFilter =
                new RepoConfigPropagationModelFilter(repoConfigCache.getCachedBasicRepoConfigs(),
                        storageService, ConstantValues.repoConfigChangeVerifyOnGetAllRepoKeysSinceSize.getInt());
        repoConfigPropagationModelFilter.filter(model);

        Map<String, BasicRepoConfig> repoKeysDatesToCreateUpdate = model.getRepoConfigCreateOrUpdate().stream()
                .collect(toMap(RepoConfigReload::getRepoKey, RepoConfigReload::toBasic));
        repoConfigCache.addAndUpdateRepoConfigKeys(repoKeysDatesToCreateUpdate);

        Set<String> repoKeysToDelete = model.getRepoConfigDelete().stream()
                .map(RepoConfigReload::getRepoKey)
                .collect(Collectors.toSet());
        repoConfigCache.deleteRepoConfigsByKeys(repoKeysToDelete);

        Map<RepoDetailsType, List<RepoConfigReload>> reposToReload =
                Stream.of(model.getRepoConfigCreateOrUpdate(), model.getRepoConfigDelete())
                        .flatMap(Collection::stream)
                        .collect(groupingBy(RepoConfigReload::getRepoDetailsType));

        ContextHelper.get().beanForType(RepositoryConfigService.class)
                .callOnRepoConfigReloadAsync(reposToReload);
    }

    @Override
    public boolean existsByType(RepoDetailsType repoType, String repoKey) {
        return findRepositoriesByType(repoType).stream()
                .filter(repoConfig -> !isInvalid(repoConfig.getKey()))
                .anyMatch(repoConfig -> repoConfig.getKey().equals(repoKey));
    }

    @Override
    public List<VirtualRepoConfig> getVirtualReposContainingRepo(RepoConfig repoConfig, int depth) {
        RepoConfig config = repoConfig;
        if (repoConfig instanceof LocalCacheRepoConfig cacheRepoConfig) {
            //VirtualRepoResolver does not directly support local cache repos, so if the items config is a cache,
            //We extract the caches remote repo, and use it instead
            config = cacheRepoConfig.getRemoteRepoConfig();
        }
        List<VirtualRepoConfig> reposToDisplay = new ArrayList<>();
        List<VirtualRepoConfig> virtualRepos = getVirtualRepoConfigs();
        for (VirtualRepoConfig virtualRepo : virtualRepos) {
            VirtualRepositoryConfigResolverImpl resolver = new VirtualRepositoryConfigResolverImpl(this);
            LinkedHashSet<RepoDetailsType> repoDetailsTypes =
                    CollectionUtils.newLinkedHashSet(LOCAL, FEDERATED, REMOTE, VIRTUAL);
            Set<String> resolvedRepoConfigs =
                    resolver.getRepoConfigs(virtualRepo.getKey(), repoDetailsTypes, depth).stream()
                            .map(RepoConfig::getKey)
                            .collect(Collectors.toSet());
            if (resolvedRepoConfigs.contains(config.getKey())) {
                reposToDisplay.add(virtualRepo);
            }
        }
        return reposToDisplay;
    }

    @Override
    public List<VirtualRepoConfig> getVirtualReposContainingRepo(RepoConfig repoConfig) {
        return getVirtualReposContainingRepo(repoConfig, Integer.MAX_VALUE);
    }

    @Override
    public void exportTo(ExportSettings settings) {
        MutableStatusHolder status = settings.getStatusHolder();
        status.status("Exporting repositories config...", log);
        writeRepositoriesConfigToFile(settings.getBaseDir(), getAllRepoConfigsSkipCache());
    }

    @Override
    public void writeRepositoriesConfigToFile(File baseDir, Collection<RepoConfig> allRepositoryConfigs) {
        File destFile = new File(baseDir, ArtifactoryHome.ARTIFACTORY_REPO_CONFIG_FILE);
        AggregatedReposConfig aggregatedReposConfig = new AggregatedReposConfig(allRepositoryConfigs);
        JsonUtils.getInstance().writeValue(destFile, aggregatedReposConfig, true);
    }

    @Override
    public void importFrom(ImportSettings settings) {
        MutableStatusHolder status = settings.getStatusHolder();
        status.status("Importing repositories config...", log);
        File newReposConfigFile = new File(settings.getBaseDir(), ArtifactoryHome.ARTIFACTORY_REPO_CONFIG_FILE);
        if (!newReposConfigFile.exists()) {
            log.info("Repositories configuration file {} does not exist, repositories will not be imported.",
                    newReposConfigFile);
            return;
        }
        status.status("Reloading repositories configuration from " + newReposConfigFile, log);
        AggregatedReposConfig aggregatedReposConfig =
                JsonUtils.getInstance().readValue(newReposConfigFile, AggregatedReposConfig.class);
        importRepoConfigs(aggregatedReposConfig.getAll(), true, true);
        status.status("Repositories configuration reloaded from " + newReposConfigFile, log);
    }

    @Override
    public void importRepoConfigs(List<? extends RepoConfig> repoConfigsToImport, boolean reload,
            boolean propagateChanges) {
        repoConfigsToImport = convertImportedRepoConfigs(repoConfigsToImport);
        List<String> oldRepoConfigsKeys =
                new ArrayList<>(getAllBasicRepoConfigs().keySet());
        bulkUpsertRepositoryConfigs(repoConfigsToImport, reload, true);
        oldRepoConfigsKeys.removeAll(repoConfigsToImport.stream().map(RepoConfig::getKey).toList());
        bulkDeleteRepositoryConfigs(oldRepoConfigsKeys, reload, propagateChanges, true);
    }

    @Override
    public RepoType getRepoType(String repoKey) {
        if (StringUtils.isBlank(repoKey)) {
            return null;
        }
        if (repoKey.endsWith(REMOTE_CACHE_SUFFIX)) {
            repoKey = StringUtils.removeEnd(repoKey, REMOTE_CACHE_SUFFIX);
        }
        RepoConfig repoConfigByKey = getRepoConfigByKey(repoKey);
        return repoConfigByKey == null ? null : repoConfigByKey.getPackageType();
    }

    // TODO Move to interceptors
    private void validate(Collection<? extends RepoConfig> repoConfigs) {
        ContextHelper.get().beanForType(ProjectAdminRepoConfigValidator.class).validate(repoConfigs);
        ContextHelper.get().beanForType(ArtifactoryVersionValidator.class).validate();
        configValidationService.validate(repoConfigs);
    }

    /**
     * @return list of repo configs directly from the storage and not through the cache
     */
    public Set<RepoConfig> getAllRepoConfigsSkipCache() {
        return storageService.getAllRepositories().stream()
                .map(this::dbInfoBlobToRepositoryConfig)
                .collect(toSet());
    }

    private List<RepoConfig> convertImportedRepoConfigs(List<? extends RepoConfig> repoConfigs) {
        Set<byte[]> repoInfoBlobs = repoConfigs.stream()
                .map(repoConfig -> JsonUtils.getInstance().valueToByteArray(repoConfig))
                .collect(Collectors.toSet());
        RepositoriesConverterManager repositoriesConverterManager = new RepositoriesConverterManager(subConverters);
        return repoInfoBlobs.stream()
                .map(repositoriesConverterManager::convert)
                .toList();
    }

    private boolean isInvalid(String repoKey) {
        return StringUtils.isBlank(repoKey) || isTrash(repoKey) || isSupportBundle(repoKey) || isCache(repoKey);
    }

    private void insertRemoteRepoConfig(RemoteRepoConfig remoteRepoConfig) {
        remoteRepoConfig.processUrl();
        CentralConfigService centralConfig = ContextHelper.get().getCentralConfig();
        MutableCentralConfigDescriptor mutableDescriptor = centralConfig.getMutableDescriptor();
        boolean addToBackups = mutableDescriptor.conditionallyAddToBackups(remoteRepoConfig.getKey());
        if (addToBackups) {
            centralConfig.saveEditedDescriptorAndReload(mutableDescriptor);
        }
        insertRepositoryInternal(remoteRepoConfig);
    }

    private void updateRemoteRepoConfig(RemoteRepoConfig remoteRepoConfig) {
        remoteRepoConfig.processUrl();
        updateRepositoryInternal(remoteRepoConfig);
    }

    private List<RepoConfig> getRepoConfigsByTypes(Set<RepoDetailsType> types) {
        List<RepoConfig> allRepos = new ArrayList<>();
        for (RepoDetailsType type : types) {
            allRepos.addAll(findRepositoriesByType(type));
        }
        return allRepos;
    }

    @VisibleForTesting
    public void setCache(RepoConfigCache repoConfigCache) {
        this.repoConfigCache = repoConfigCache;
    }

    @NotNull
    private List<VirtualRepoConfig> getVirtualRepoConfigs() {
        return findRepositoriesByType(RepoDetailsType.VIRTUAL).stream()
                .map(VirtualRepoConfig.class::cast)
                .toList();
    }

    @NotNull
    private RepoConfigChange getRepoConfigCreateChange(RepoConfig repositoryConfig) {
        return new RepoConfigChange(repositoryConfig.getKey(),
                REPO_CREATE, repositoryConfig, null);
    }

    @NotNull
    private RepoConfigChange getRepoConfigUpdateChange(RepoConfig repositoryConfig, RepoConfig oldRepoConfig) {
        return new RepoConfigChange(repositoryConfig.getKey(),
                RepoConfigChange.ConfigAction.REPO_UPDATE, repositoryConfig, oldRepoConfig);
    }

    @NotNull
    private RepoConfigChange getRepoConfigDeleteChange(RepoConfig oldRepoConfig) {
        return new RepoConfigChange(oldRepoConfig.getKey(),
                RepoConfigChange.ConfigAction.REPO_DELETE, null, oldRepoConfig);
    }

    @NotNull
    private Set<DbRepositoryConfig> getDbRepoConfigsByAction(
            Map<RepoDetailsType, List<RepoConfigChange>> mutableRepoConfigChangesByType,
            RepoConfigChange.ConfigAction configAction) {
        return mutableRepoConfigChangesByType.values().stream()
                .flatMap(Collection::stream)
                .filter(change -> configAction.equals(change.getConfigAction()))
                .map(change -> repositoryConfigMapper
                        .repositoryConfigToDbRepositoryConfig(change.getNewRepoConfig()))
                .collect(Collectors.toSet());
    }

    @VisibleForTesting
    void dereferenceRepositoryFromConfigDescriptor(List<RepoConfig> oldRepoConfigs) {
        CentralConfigService centralConfigService = ContextHelper.get().getCentralConfig();
        CentralConfigDescriptor oldDescriptor = centralConfigService.getMutableDescriptor();
        MutableCentralConfigDescriptor mutableDescriptor = centralConfigService.getMutableDescriptor();
        oldRepoConfigs.forEach(mutableDescriptor::removeReferences);
        DiffResult diffs = ArtifactoryConfigDiffUtil.findDiff(oldDescriptor, mutableDescriptor);
        if (!diffs.getDiffs().isEmpty()) {
            centralConfigService.saveEditedDescriptorAndReload(mutableDescriptor);
        }
    }

    private List<VirtualRepoConfig> deleteReferenceFromVirtualRepos(List<RepoConfig> removedRepos, boolean isImport) {
        // remove from any virtual repository
        List<VirtualRepoConfig> virtualRepoConfigs = findRepositoriesByType(VIRTUAL, isImport).stream()
                .map(VirtualRepoConfig.class::cast)
                .toList();
        if (virtualRepoConfigs.isEmpty()) {
            return List.of();
        }
        List<VirtualRepoConfig> virtualRepoConfigsToUpdate = new ArrayList<>();
        for (VirtualRepoConfig virtualRepoConfig : virtualRepoConfigs) {
            removedRepos.forEach(removedRepo -> {
                if (removeRefsFromVirtualRepoConfig(removedRepo, virtualRepoConfig)) {
                    virtualRepoConfigsToUpdate.add(virtualRepoConfig);
                }
            });

        }
        return virtualRepoConfigsToUpdate;
    }

    private boolean removeRefsFromVirtualRepoConfig(RepoConfig removedRepo, VirtualRepoConfig virtualRepoConfig) {
        boolean virtualConfigChanged = virtualRepoConfig.removeRepository(removedRepo.getKey());
        if (removedRepo instanceof LocalConfig &&
                virtualRepoConfig.getRepoTypeConfig().getDefaultDeploymentRepoRef() != null &&
                virtualRepoConfig.getRepoTypeConfig().getDefaultDeploymentRepoRef()
                        .equals(removedRepo.getKey())) {
            virtualRepoConfig.getRepoTypeConfig().setDefaultDeploymentRepoRef(null);
            virtualConfigChanged = true;
        }
        ExternalDependenciesConfig externalDependencies =
                ExternalDependenciesPackageTypeConfigMapper.INSTANCE.map(virtualRepoConfig.getPackageTypeConfig());
        if (removedRepo instanceof RemoteRepoConfig && externalDependencies != null &&
                removedRepo.getKey().equals(externalDependencies.getRemoteRepoRef())) {
            removeExternalDependencies(virtualRepoConfig);
            virtualConfigChanged = true;
        }
        return virtualConfigChanged;
    }

    private void removeExternalDependencies(RepoConfig repoConfig) {
        repoConfig.getPackageTypeConfig()
                .remove(ExternalDependenciesPackageTypeConfigMapper.EXTERNAL_DEPENDENCIES_ENABLED);
        repoConfig.getPackageTypeConfig()
                .remove(ExternalDependenciesPackageTypeConfigMapper.EXTERNAL_DEPENDENCIES_PATTERNS);
        repoConfig.getPackageTypeConfig()
                .remove(ExternalDependenciesPackageTypeConfigMapper.EXTERNAL_DEPENDENCIES_REMOTE_REPO_REF);
    }

    private void flushRepoConfigChanges(Map<RepoDetailsType, List<RepoConfigChange>> mutableRepoConfigChangesByType,
            boolean propagateChanges) {
        mutableRepoConfigChangesByType.values()
                .stream()
                .flatMap(List::stream)
                .filter(change -> !change.getConfigAction().equals(REPO_DELETE))
                .filter(change -> change.getNewRepoConfig() != null)
                .forEach(configChange -> configChange.getNewRepoConfig().setModification(System.currentTimeMillis()));

        Set<DbRepositoryConfig> reposToCreate = getDbRepoConfigsByAction(mutableRepoConfigChangesByType, REPO_CREATE);
        Set<DbRepositoryConfig> reposToUpdate = getDbRepoConfigsByAction(mutableRepoConfigChangesByType, REPO_UPDATE);
        Set<String> reposToDelete = mutableRepoConfigChangesByType.values().stream()
                .flatMap(Collection::stream)
                .filter(change -> REPO_DELETE.equals(change.getConfigAction()))
                .map(RepoConfigChange::getRepoKey)
                .collect(Collectors.toSet());
        storageService.flushBulkChanges(reposToCreate, reposToUpdate, reposToDelete);
        if (!reposToCreate.isEmpty()) {
            repoConfigCache.addRepoConfigKeys(dbInfoBlobListToRepoKeyAndModificationMap(reposToCreate));
        }
        if (!reposToUpdate.isEmpty()) {
            repoConfigCache.updateRepoConfigKeys(dbInfoBlobListToRepoKeyAndModificationMap(reposToUpdate));
        }
        if (!reposToDelete.isEmpty()) {
            repoConfigCache.deleteRepoConfigsByKeys(reposToDelete);
        }
        if (propagateChanges) {
            propagateCreateRepoConfig(mutableRepoConfigChangesByType);
        }
        writeMetricsInfoToMetricsThreadLocal(mutableRepoConfigChangesByType);
    }

    private Map<String, RepoConfig> getAllRepoConfigsWithoutCache() {
        List<DbRepositoryConfig> dbRepoConfigs = storageService.getAllRepositories();
        return dbInfoBlobListToRepoConfigsSet(dbRepoConfigs).stream()
                .collect(Collectors.toMap(RepoConfig::getKey, Function.identity()));
    }

    private RepoConfig getRepoByKeyWithoutCache(String repoKey) {
        DbRepositoryConfig dbRepositoryConfig = storageService.findRepositoryByKey(repoKey);
        if (dbRepositoryConfig != null) {
            return repositoryConfigMapper.dbInfoBlobToRepositoryConfig(dbRepositoryConfig);
        }
        return null;
    }

    private Map<String, RepoConfig> getRepoConfigsByKeysWithoutCache(Set<String> repoKeys) {
        Set<DbRepositoryConfig> dbRepoConfigs = storageService.findRepositoryConfigsByKeys(repoKeys);
        return dbInfoBlobListToRepoConfigsSet(dbRepoConfigs).stream()
                .collect(Collectors.toMap(RepoConfig::getKey, Function.identity()));
    }

    private Set<RepoConfig> dbInfoBlobListToRepoConfigsSet(Collection<DbRepositoryConfig> dbInfoBlobs) {
        return dbInfoBlobs.stream()
                .map(dbRepo -> repositoryConfigMapper.dbInfoBlobToRepositoryConfig(dbRepo))
                .collect(toSet());
    }

    private Map<String, BasicRepoConfig> dbInfoBlobListToRepoKeyAndModificationMap(
            Collection<DbRepositoryConfig> dbInfoBlobs) {
        return dbInfoBlobListToRepoConfigsSet(dbInfoBlobs).stream()
                .collect(toMap(RepoConfig::getKey, RepoConfig::toBasic));
    }

    private void flushRepoConfigChange(RepoConfig repositoryConfig, RepoConfigChange repoConfigChange) {
        RepoDetailsType type = repositoryConfig.getRepoType();
        Map<RepoDetailsType, List<RepoConfigChange>> mutableRepoConfigChangesByType =
                new EnumMap<>(RepoDetailsType.class);
        mutableRepoConfigChangesByType.put(type, Lists.newLinkedList(List.of(repoConfigChange)));
        interceptors.forEach(interceptor -> interceptor.onBeforeSave(mutableRepoConfigChangesByType));
        flushRepoConfigChanges(mutableRepoConfigChangesByType, true);
        callOnRepoConfigReload(convertToRepoConfigReloadByType(mutableRepoConfigChangesByType));
        interceptors.forEach(interceptor -> interceptor.onAfterSave(mutableRepoConfigChangesByType));
    }

    private void createDefaultRepoConfigsIfDoesNotExist() {
        Authentication originalAuthentication = SecurityContextHolder.getContext().getAuthentication();
        try {
            SecurityContextHolder.getContext().setAuthentication(new SystemAuthenticationToken());
            List<LocalRepoConfig> reposToCreate = getLocalRepoConfigsToCreate();
            bulkInsertRepositoryConfigs(reposToCreate, true, true, false);
        } finally {
            SecurityContextHolder.getContext().setAuthentication(originalAuthentication);
        }
    }

    /**
     * Generate Local repo configs to create on first start.
     * If {@link ConstantValues#skipExampleRepoCreation} is set, example-repo-local will not get created.
     */
    @NonNull
    private List<LocalRepoConfig> getLocalRepoConfigsToCreate() {
        LocalRepoConfig buildInfoRepoConfig =
                new BuildInfoRepo().buildRepoConfig(ContextHelper.get().getCentralConfig().getDescriptor());
        LocalRepoConfig exampleLocalRepoConfig = generateExampleRepoLocalConfig();
        List<LocalRepoConfig> reposToCreate;
        if (ConstantValues.skipExampleRepoCreation.getBoolean()) {
            reposToCreate = List.of(buildInfoRepoConfig);
        } else {
            reposToCreate = List.of(buildInfoRepoConfig, exampleLocalRepoConfig);
        }
        return reposToCreate;
    }

    private LocalRepoConfig generateExampleRepoLocalConfig() {
        LocalRepoConfig exampleLocalRepoConfig = LocalRepoConfig.create(EXAMPLE_REPO_KEY, RepoType.Generic);
        exampleLocalRepoConfig.getBaseConfig().setDescription("Example artifactory repository");
        exampleLocalRepoConfig.getRepoTypeConfig().setHandleReleases(true);
        exampleLocalRepoConfig.getRepoTypeConfig().setHandleSnapshots(true);
        return exampleLocalRepoConfig;
    }

    private void insertRepositoryInternal(RepoConfig repositoryConfig) {
        RepoConfigChange repoConfigChange = getRepoConfigCreateChange(repositoryConfig);
        flushRepoConfigChange(repositoryConfig, repoConfigChange);
    }

    private void updateRepositoryInternal(RepoConfig repositoryConfig) {
        RepoConfig oldRepoConfig = getRepoConfigByKey(repositoryConfig.getKey());
        if (oldRepoConfig == null) {
            throw new NotFoundException(
                    String.format("Cannot update repository '%s': repository config does not exist.",
                            repositoryConfig.getKey()));
        }
        RepoConfigChange repoConfigChange = getRepoConfigUpdateChange(repositoryConfig, oldRepoConfig);
        flushRepoConfigChange(repositoryConfig, repoConfigChange);
    }

    private void setBackups(List<String> repoKeys) {
        CentralConfigService centralConfig = ContextHelper.get().getCentralConfig();
        MutableCentralConfigDescriptor mutableDescriptor = centralConfig.getMutableDescriptor();
        List<String> addToBackups =
                repoKeys.stream().filter(mutableDescriptor::conditionallyAddToBackups).toList();
        if (!addToBackups.isEmpty()) {
            centralConfig.saveEditedDescriptorAndReload(mutableDescriptor);
        }
    }

    private void propagateCreateRepoConfig(Map<RepoDetailsType, List<RepoConfigChange>> repoConfigChangeByType) {
        RepoConfigPropagationModel propagationModel = new RepoConfigPropagationModel();
        List<RepoConfigReload> repoConfigReloads = repoConfigChangeByType.values().stream().flatMap(Collection::stream)
                .map(this::repoConfigChangeToRepoConfigReload).toList();
        propagationModel.setRepoConfigCreateOrUpdate(
                filterRepoConfigChangeByAction(repoConfigReloads, Set.of(REPO_CREATE, REPO_UPDATE)));
        propagationModel.setRepoConfigDelete(filterRepoConfigChangeByAction(repoConfigReloads, Set.of(REPO_DELETE)));
        if (propagationModel.isEmpty()) {
            return;
        }
        ContextHelper.get().beanForType(AddonsManager.class).addonByType(HaAddon.class)
                .propagateRepoConfigsChanges(propagationModel);
    }

    private Set<RepoConfigReload> filterRepoConfigChangeByAction(List<RepoConfigReload> repoConfigReloads,
            Set<RepoConfigChange.ConfigAction> actions) {
        return repoConfigReloads.stream()
                .filter(repo -> actions.contains(repo.getConfigAction()))
                .collect(Collectors.toSet());
    }

    private boolean isPomCleanUpPolicyChanged(RepoConfig newRepoConfig, RepoConfig oldRepoConfig) {
        // TODO[t.r] - consider reading the values differently
        MavenConfiguration newMavenConfig = getDeserializedType(MavenConfiguration.class, newRepoConfig);
        MavenConfiguration oldMavenConfig = getDeserializedType(MavenConfiguration.class, oldRepoConfig);
        PomCleanupPolicy newPolicy = newMavenConfig.getPomRepositoryReferencesCleanupPolicy();
        PomCleanupPolicy oldPolicy = oldMavenConfig.getPomRepositoryReferencesCleanupPolicy();
        return !newPolicy.equals(oldPolicy);
    }

    private boolean isRemoteRepoConfigChanged(RemoteRepoConfig newConfig, RemoteRepoConfig oldConfig) {
        return newConfig.getRepoTypeConfig().getSocketTimeoutMillis() !=
                (oldConfig.getRepoTypeConfig().getSocketTimeoutMillis()) ||
                !StringUtils.equals(newConfig.getRepoTypeConfig().getUrl(), oldConfig.getRepoTypeConfig().getUrl());
    }

    /**
     * Writes relevant data to {@link MetricsThreadLocal}. This data is later used by MetricsListenerService.
     *
     * @param typeToChanges Map of config changes to process.
     */
    void writeMetricsInfoToMetricsThreadLocal(Map<RepoDetailsType, List<RepoConfigChange>> typeToChanges) {
        List<MetricsInfo> metricsInfos = new ArrayList<>();
        typeToChanges.forEach((repoDetailsType, repoConfigChanges) -> repoConfigChanges
                .forEach(repoConfigChange -> metricsInfos.add(
                        MetricsInfo.builder().repoClass(repoDetailsType.typeNameLowercase())
                                .repoType(repoConfigChange.getConfigAction() == REPO_CREATE ?
                                        repoConfigChange.getNewRepoConfig().getPackageType().getType() :
                                        repoConfigChange.getOldRepoConfig().getPackageType().getType())
                                .action(getRepoActionName(repoConfigChange)).build())));
        MetricsThreadLocal.setMetricsInfoList(metricsInfos);
    }

    private String getRepoActionName(RepoConfigChange repoConfigChange) {
        switch (repoConfigChange.getConfigAction()) {
            case REPO_CREATE:
                return "REPO_CREATED";
            case REPO_UPDATE:
                return "REPO_UPDATED";
            case REPO_DELETE:
                return "REPO_DELETED";
            default:
                String actionName = repoConfigChange.getConfigAction().name();
                log.error("Failed to convert {} config action. Action not found", actionName);
                return null;
        }
    }

    @Override
    public Map<RepoDetailsType, List<RepoConfigChange>> apply(Set<? extends RepoConfig> repoConfigs) {
        Collection<RepoConfig> existingRepos = getExistingRepoConfigs(repoConfigs.stream().toList(), false);
        return convertRepoConfigsToCreateUpdateChanges(repoConfigs.stream().toList(), existingRepos);
    }

    /**
     * Convert repo configs to {@link RepoConfigChange} and group by {@link RepoDetailsType}.
     */
    public static Map<RepoDetailsType, List<RepoConfigChange>> convertRepoConfigsToCreateUpdateChanges(
            Collection<? extends RepoConfig> repoConfigsToClassify, Collection<RepoConfig> existingRepoConfigs) {
        Map<String, RepoConfig> existingRepoConfigsMap = existingRepoConfigs.stream()
                .collect(toMap(RepoConfig::getKey, Function.identity()));
        return repoConfigsToClassify.stream()
                .map(repo -> {
                    RepoConfig existingRepo = existingRepoConfigsMap.get(repo.getKey());
                    RepoConfigChange.ConfigAction configAction = existingRepo == null ? REPO_CREATE : REPO_UPDATE;
                    return new RepoConfigChange(repo.getKey(), configAction, repo, existingRepo);
                })
                .collect(groupingBy(change -> change.getNewRepoConfig().getRepoType()));
    }

    /**
     * Convert repo configs to {@link RepoConfigChange} and group by {@link RepoDetailsType}.
     */
    public static Map<RepoDetailsType, List<RepoConfigChange>> convertRepoConfigsToUpdateChanges(
            List<? extends RepoConfig> repoConfigsToClassify, Collection<RepoConfig> existingRepoConfigs) {
        Map<String, RepoConfig> existingRepoConfigsMap = existingRepoConfigs.stream()
                .collect(toMap(RepoConfig::getKey, Function.identity()));
        return repoConfigsToClassify.stream()
                .map(repo -> {
                    RepoConfig existingRepo = existingRepoConfigsMap.get(repo.getKey());
                    if (existingRepo != null) {
                        return new RepoConfigChange(repo.getKey(), REPO_UPDATE, repo, existingRepo);
                    }
                    return null;
                }).filter(Objects::nonNull)
                .collect(groupingBy(change -> change.getNewRepoConfig().getRepoType()));
    }


    private Collection<RepoConfig> getExistingRepoConfigs(Collection<? extends RepoConfig> repositoryConfigs,
            boolean isImport) {
        Set<String> repoKeys = repositoryConfigs.stream().map(RepoConfig::getKey).collect(Collectors.toSet());
        return isImport ? getRepoConfigsByKeysWithoutCache(repoKeys).values() : getReposByKeysFromCache(repoKeys);
    }

    private Collection<RepoConfig> getReposByKeysFromCache(Set<String> repoKeys) {
        Set<String> requestedRepoKeys = new HashSet<>(repoKeys);
        Set<String> allRepoKeys = repoConfigCache.getAllRepoKeys();
        requestedRepoKeys.retainAll(allRepoKeys);
        return requestedRepoKeys.isEmpty() ? List.of() : repoConfigCache.getRepositoryConfigsByKeys(requestedRepoKeys);
    }

    private Collection<RepoConfig> findRepositoriesByType(RepoDetailsType repoType, boolean isImport) {
        if (isImport) {
            String typeAsString = repoConfigAttributesMapper.repoDetailTypeToLowercaseRepoTypeName(repoType);
            return dbInfoBlobListToRepoConfigsSet(storageService.findRepositoriesByDetailsType(typeAsString));
        }
        Set<BasicRepoConfig> allBasicRepoConfigs = repoConfigCache.getAllBasicConfigs();
        Set<String> filteredRepoKeys = allBasicRepoConfigs.stream()
                .filter(repoConfig -> repoType.equals(repoConfig.getRepoType()))
                .map(BasicRepoConfig::getRepoKey)
                .collect(toSet());
        return new ArrayList<>(repoConfigCache.getRepositoryConfigsByKeys(filteredRepoKeys));
    }

    private RepoConfig dbInfoBlobToRepositoryConfig(DbRepositoryConfig dbRepositoryConfig) {
        return Optional.of(dbRepositoryConfig)
                .map(repositoryConfigMapper::dbInfoBlobToRepositoryConfig)
                .orElse(null);
    }

    private void callOnRepoConfigReloadAsync(RepoConfigPropagationModel model) {
        if (model == null) {
            return;
        }
        ContextHelper.get().beanForType(RepositoryConfigService.class).callOnRepoConfigReloadAsync(
                Stream.of(model.getRepoConfigCreateOrUpdate(), model.getRepoConfigDelete())
                        .flatMap(Collection::stream)
                        .collect(groupingBy(RepoConfigReload::getRepoDetailsType)));
    }

    private Map<String, BasicRepoConfig> getAllBasicRepoConfigs() {
        return storageService.getAllBasicRepoConfigs().stream()
                .map(entity -> new BasicRepoConfig(entity.getKey(), entity.getModified(),
                        repoConfigAttributesMapper.repoTypeNameToRepoDetailsType(entity.getType()),
                        repoConfigAttributesMapper.typeStringToPackageType(entity.getPackageType())))
                .collect(toMap(BasicRepoConfig::getRepoKey, Function.identity()));
    }
}
