package org.artifactory.security.access.providers;

import lombok.extern.slf4j.Slf4j;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.security.AnonymousAuthenticationToken;
import org.artifactory.security.RealmAwareAuthenticationProvider;
import org.artifactory.security.SimpleUser;
import org.artifactory.security.UserGroupInfo;
import org.artifactory.security.UserInfo;
import org.artifactory.security.access.AccessResponseAuthentication;
import org.artifactory.security.access.AccessService;
import org.artifactory.security.access.util.AccessAuthProviderUtils;
import org.artifactory.util.SecurityUtils;
import org.jfrog.access.client.AccessClientHttpException;
import org.jfrog.access.client.auth.LoginResult;
import org.jfrog.access.client.auth.UserPrincipal;
import org.jfrog.access.model.Realm;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.authentication.AuthenticationServiceException;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.AuthenticationException;
import org.springframework.stereotype.Component;

import java.util.Objects;
import java.util.Set;

/**
 * @author Mor Merhav
 */
@Slf4j
@Component
public class AccessInternalAuthenticationProvider implements RealmAwareAuthenticationProvider {

    private final AccessService accessService;
    private final AuthorizationService authorizationService;

    @Autowired
    public AccessInternalAuthenticationProvider(AccessService accessService,
            AuthorizationService authorizationService) {
        this.accessService = accessService;
        this.authorizationService = authorizationService;
    }

    @Override
    public String getRealm() {
        return Realm.INTERNAL.getName();
    }

    @Override
    public void addExternalGroups(String username, Set<UserGroupInfo> groups) {
        // Noop
    }

    @Override
    public boolean userExists(String userName) {
        return false; // Not called for internal
    }

    @Override
    public Authentication authenticate(Authentication authentication) throws AuthenticationException {
        String username = authentication.getName();
        String password = (String) authentication.getCredentials();

        SecurityUtils.forceEnableBasicAuthIfNeeded();

        if (SecurityUtils.isAnyExternalAuthEnabled() &&
                !accessService.isBasicAuthEnabled() && !UserInfo.ANONYMOUS.equals(username)) {
            log.debug("Attempt to log in with disabled basic authentication");
            throw new AuthenticationServiceException("Basic authentication is disabled");
        }

        // RTDEV-24978 - Do not call Access if anon access is disabled
        if (!authorizationService.isAnonAccessEnabled() && UserInfo.ANONYMOUS.equals(authentication.getName())) {
            return new AnonymousAuthenticationToken();
        }
        log.debug("Internal authentication is being performed via Access service.");

        try {/**/
            LoginResult loginResult = accessService.ensureAuth(() -> accessService.getGrpcAuthClient()
                    .authenticate(username, password));
            SimpleUser simpleUser = AccessAuthProviderUtils.toArtifactoryUser((UserPrincipal) Objects.requireNonNull(
                    loginResult.getPrincipal()), password);
            return new AccessResponseAuthentication(true, simpleUser);
        } catch (AccessClientHttpException e) {
            // Case we got response from Access server and the user can't authenticate
            AccessAuthProviderUtils.handleAccessClientHttpException(authentication, e);
            throw e;
        }
    }

    @Override
    public boolean supports(Class<?> authentication) {
        return accessService.isRealmAuthenticationEnabled(Realm.INTERNAL.getName());
    }
}
