package org.artifactory.servicetrust.cold;

import lombok.AccessLevel;
import lombok.NoArgsConstructor;
import lombok.NonNull;
import org.apache.commons.lang.StringUtils;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.retention.ArtifactRetentionAddon;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.api.security.access.GenericTokenSpec;
import org.artifactory.api.security.access.TokenSpec;
import org.artifactory.descriptor.authentication.AuthenticationToken;
import org.artifactory.servicetrust.ExchangeTokenRequest;
import org.artifactory.servicetrust.ExchangeTokenResponse;
import org.artifactory.servicetrust.ServiceTrust;
import org.artifactory.servicetrust.ServiceTrustException;
import org.artifactory.servicetrust.ServiceTrustToken;
import org.artifactory.servicetrust.ServiceTrustTokenData;
import org.artifactory.servicetrust.ServiceTrustUrls;
import org.artifactory.servicetrust.ServiceTrustUseCase;
import org.artifactory.servicetrust.TokenExtension;
import org.artifactory.servicetrust.TokenSpecOutboundService;
import org.artifactory.servicetrust.TokenSpecServiceInboundBase;
import org.jfrog.access.common.ServiceId;
import org.jfrog.access.token.JwtAccessToken;

import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Optional;
import java.util.stream.Stream;

import static org.artifactory.addon.retention.ArtifactRetentionConstants.INSTANCES_ALREADY_PAIRED;
import static org.artifactory.servicetrust.ServiceTrustService.INTERNAL_SCOPE_PREFIX;

/**
 * @author volodymyrm
 */
@NoArgsConstructor(access = AccessLevel.PRIVATE)
public class ArtifactoryColdTokenSpecService extends TokenSpecServiceInboundBase
        implements TokenSpecOutboundService {

    public static final ArtifactoryColdTokenSpecService ARTIFACTORY_COLD_TOKEN_SPEC_SERVICE =
            new ArtifactoryColdTokenSpecService();

    @Override
    public TokenSpec<GenericTokenSpec> createPairingTokenSpec(@NonNull ServiceTrustTokenData tokenData,
            @NonNull ServiceTrust trust,
            @NonNull ServiceTrustUrls urls) {
        verifyIsColdStorageEnabled();
        return super.createPairingTokenSpec(tokenData, trust, urls);
    }

    @NonNull
    @Override
    protected String getUseCasePathForScope(@NonNull String scope) {
        String useCasePath = getUseCaseFromPairScope(Collections.singletonList(scope));
        if (ServiceTrustUseCase.ARTIFACTORY_COLD.matches(useCasePath)) {
            useCasePath = generateIceUseCaseWithNamespace(useCasePath);
        }
        return useCasePath;
    }

    @Override
    public TokenSpec<GenericTokenSpec> createMasterTokenSpec(@NonNull ServiceTrustTokenData tokenData,
            @NonNull String useCasePath, @NonNull ServiceTrustUrls urls) {
        verifyIsColdStorageEnabled();
        return super.createMasterTokenSpec(tokenData, useCasePath, urls);
    }

    @Override
    public TokenSpec<GenericTokenSpec> createExchangeTokenSpec(ServiceId artifactoryServiceId, String useCasePath) {
        verifyIsColdStorageEnabled();
        return super.createExchangeTokenSpec(artifactoryServiceId, useCasePath);
    }

    @Override
    public ExchangeTokenRequest createExchangeTokenRequest(String subject, JwtAccessToken jwt) {
        verifyIsWarmStorageEnabled();
        String updatedUseCasePath = getUseCaseFromPairScope(jwt.getScope());
        return new ExchangeTokenRequest(subject, updatedUseCasePath);
    }

    @Override
    public void storeMasterTokenData(ExchangeTokenResponse response, TokenExtension extension) {
        verifyIsWarmStorageEnabled();
        assertInstanceNotPaired();
        String scope = response.getScope();
        assertScopeExists(scope);
        String[] tokens = scope.split(" ");
        String tokenWithNamespace = findFirstRetentionToken(Arrays.stream(tokens),
                "Master token scope with the 'artifactory-cold' scope does not exist");
        String namespace = tokenWithNamespace
                .replace(ServiceTrustUseCase.ARTIFACTORY_COLD.getName() + "/", "")
                .replace(ServiceTrustUseCase.ARTIFACTORY_COLD.getName(), "")
                .replace(INTERNAL_SCOPE_PREFIX, "");

        if (StringUtils.isEmpty(namespace)) {
            throw new IllegalArgumentException("Can not parse the master token scope to get a cold namespace: " +
                    "namespace is empty");
        }
        getArtifactRetentionAddon().saveNamespaceAndMasterToken(extension.getBaseUrl(), extension.getExchangeUrl(),
                namespace, response.getTokenId(), response.getAccessToken(), response.getTokenType(),
                response.getExpiresIn(), response.getScope());
    }

    private void assertInstanceNotPaired() {
        if (!getArtifactRetentionAddon().getColdStoragePairAuthenticationTokens().isEmpty()) {
            throw new ArtifactRetentionBindException(INSTANCES_ALREADY_PAIRED);
        }
    }

    private void assertScopeExists(String scope) {
        if (StringUtils.isBlank(scope)) {
            throw new IllegalArgumentException("Master token does not have scope");
        }
    }

    @Override
    public void updateMasterTokenData(ExchangeTokenResponse response) {
        verifyIsWarmStorageEnabled();
        getArtifactRetentionAddon().updateMasterToken(response.getTokenId(), response.getAccessToken(),
                response.getTokenType(), response.getExpiresIn(), response.getScope());
    }

    @Override
    public ExchangeTokenRequest createMasterTokenExchangeRefreshRequest(@NonNull String subject) {
        verifyIsWarmStorageEnabled();
        String namespace = getArtifactRetentionAddon().getJfrogColdStorageArchivingNamespace();
        if (StringUtils.isEmpty(namespace)) {
            throw new IllegalStateException("Can not create exchange request to refresh master token: " +
                    "namespace does not exist");
        }
        return new ExchangeTokenRequest(subject,
                String.join("/", ServiceTrustUseCase.ARTIFACTORY_COLD.getName(), namespace));
    }

    @Override
    public ExchangeTokenRequest createScopedTokenExchangeRequest(@NonNull String subject, String... scopes) {
        verifyIsWarmStorageEnabled();
        String namespace = getArtifactRetentionAddon().getJfrogColdStorageArchivingNamespace();
        if (StringUtils.isEmpty(namespace)) {
            throw new ServiceTrustException("Can not create exchange request to refresh master token: " +
                    "namespace does not exist");
        }
        String target =
                String.format("%s/%s:%s", ServiceTrustUseCase.ARTIFACTORY_COLD.getName(), namespace, LIMITED_SCOPE);
        return new ExchangeTokenRequest(subject, target);
    }

    @Override
    public Optional<String> getAuthenticationToken(String baseUrl) {
        verifyIsWarmStorageEnabled();
        return getArtifactRetentionAddon().getJfrogColdStorageAuthenticationToken(baseUrl)
                .map(AuthenticationToken::getAccessToken);
    }

    @Override
    public Collection<ServiceTrustToken> getPairedInstances() {
        return getArtifactRetentionAddon().getColdStoragePairAuthenticationTokens().stream()
                .map(this::toServiceTrustToken).toList();
    }

    @Override
    public void deleteAuthenticationToken(String baseUrl) {
        verifyIsWarmStorageEnabled();
        getArtifactRetentionAddon().deleteJfrogColdStorageMasterToken(baseUrl);
    }

    /**
     * Extract a use case with namespace 'artifactory-cold/namespace' from scopes
     */
    @Override
    @NonNull
    protected String getUseCaseFromPairScope(@NonNull List<String> scope) {
        assertScopesExist(scope);
        return findFirstRetentionToken(scope.stream(), "Pair's token does not have retention related scope");
    }

    private String findFirstRetentionToken(@NonNull Stream<String> tokens, String exceptionMessage) {
        return tokens
                .filter(t -> t.contains(ServiceTrustUseCase.ARTIFACTORY_COLD.getName()))
                .findFirst()
                .map(this::getSimpleScope)
                .orElseThrow(() -> new IllegalArgumentException(exceptionMessage));
    }

    private void verifyIsColdStorageEnabled() {
        ArtifactRetentionAddon artifactRetentionAddon = getArtifactRetentionAddon();
        if (!artifactRetentionAddon.isColdInstance()) {
            throw new IllegalStateException("Cold Storage mode is disabled");
        }
    }

    private void verifyIsWarmStorageEnabled() {
        ArtifactRetentionAddon artifactRetentionAddon = getArtifactRetentionAddon();
        if (artifactRetentionAddon.isColdInstance()) {
            throw new IllegalStateException("Instance running in the Cold Storage mode");
        }
    }

    private ArtifactRetentionAddon getArtifactRetentionAddon() {
        return ContextHelper.get().beanForType(AddonsManager.class)
                .addonByType(ArtifactRetentionAddon.class);
    }

    /**
     * Generates use case with namespace for the retention feature. It should look like 'artifactory-cold/namespace'
     */
    @NonNull
    private String generateIceUseCaseWithNamespace(@NonNull String useCaseWithoutNamespace) {
        ArtifactRetentionAddon artifactRetentionAddon = getArtifactRetentionAddon();
        String namespace = artifactRetentionAddon.generateJFrogColdStorageNamespace();
        return String.join("/", useCaseWithoutNamespace, namespace);
    }
}