package org.artifactory.servicetrust.curation;

import lombok.AccessLevel;
import lombok.NoArgsConstructor;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.curation.CurationAddon;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.api.security.access.GenericTokenSpec;
import org.artifactory.api.security.access.TokenSpec;
import org.artifactory.servicetrust.ExchangeTokenRequest;
import org.artifactory.servicetrust.ExchangeTokenResponse;
import org.artifactory.servicetrust.ServiceTrust;
import org.artifactory.servicetrust.ServiceTrustToken;
import org.artifactory.servicetrust.ServiceTrustTokenData;
import org.artifactory.servicetrust.ServiceTrustUrls;
import org.artifactory.servicetrust.ServiceTrustUseCase;
import org.artifactory.servicetrust.TokenExtension;
import org.artifactory.servicetrust.TokenSpecOutboundService;
import org.artifactory.servicetrust.TokenSpecServiceInboundBase;
import org.jfrog.access.common.ServiceId;
import org.jfrog.access.token.JwtAccessToken;
import org.jfrog.common.ClockUtils;
import org.jfrog.common.JsonUtils;

import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.concurrent.TimeUnit;

import static org.artifactory.servicetrust.ServiceTrustService.INTERNAL_SCOPE_PREFIX;
import static org.artifactory.servicetrust.curation.CurationConstants.INSTANCES_ALREADY_PAIRED;

/**
 * @author Tomer Slutsky
 */
@Slf4j
@NoArgsConstructor(access = AccessLevel.PRIVATE)
public class CurationTokenSpecService extends TokenSpecServiceInboundBase implements TokenSpecOutboundService {

    public static final CurationTokenSpecService CURATION_TOKEN_SPEC_SERVICE =
            new CurationTokenSpecService();

    @Override
    public TokenSpec<GenericTokenSpec> createPairingTokenSpec(@NonNull ServiceTrustTokenData tokenData,
            @NonNull ServiceTrust trust,
            @NonNull ServiceTrustUrls urls) {
        assertCurationEnabled();
        return super.createPairingTokenSpec(tokenData, trust, urls);
    }

    @Override
    public TokenSpec<GenericTokenSpec> createMasterTokenSpec(@NonNull ServiceTrustTokenData tokenData,
            @NonNull String useCasePath, @NonNull ServiceTrustUrls urls) {
        assertCurationEnabled();
        return super.createMasterTokenSpec(tokenData, useCasePath, urls);
    }

    @Override
    public void storeMasterTokenData(ExchangeTokenResponse response, TokenExtension extension) {
        assertCurationEnabledWithTrust();
        assertInstanceNotPaired();
        String scope = response.getScope();
        String[] tokens = scope.split(" ");
        String tokenWithNamespace = getUseCaseFromPairScope(Arrays.asList(tokens));
        String namespace = tokenWithNamespace
                .replace(ServiceTrustUseCase.CURATION.getName() + "/", "")
                .replace(ServiceTrustUseCase.CURATION.getName(), "")
                .replace(INTERNAL_SCOPE_PREFIX, "");
        CurationSettings settings = CurationSettings.builder()
                .jfrogServiceTrustToken(buildCurationMasterToken(response)
                        .baseUrl(extension.getBaseUrl())
                        .exchangeUrl(extension.getExchangeUrl())
                        .build())
                .jfrogCurationNamespace(namespace)
                .enabled(false)
                .build();
        getCurationAddon().saveSettings(JsonUtils.getInstance().writeValueAsString(settings));
    }

    @Override
    public void updateMasterTokenData(ExchangeTokenResponse response) {
        assertCurationEnabledWithTrust();
        getCurationAddon().getCurationSettings()
                .map(settings -> JsonUtils.getInstance().readValue(settings, CurationSettings.class))
                .map(settings -> settings.toBuilder()
                        .jfrogServiceTrustToken(buildCurationMasterToken(response)
                                .baseUrl(settings.jfrogServiceTrustToken().baseUrl())
                                .exchangeUrl(settings.jfrogServiceTrustToken().exchangeUrl()).build()).build())
                .ifPresent(settings -> getCurationAddon().saveSettings(
                        JsonUtils.getInstance().writeValueAsString(settings)));
    }

    private ServiceTrustToken.ServiceTrustTokenBuilder buildCurationMasterToken(ExchangeTokenResponse response) {
        return ServiceTrustToken.builder()
                .accessToken(encrypt(response.getAccessToken()))
                .lastRefreshDateMillis(ClockUtils.epochMillis())
                .expirationInMillis(TimeUnit.SECONDS.toMillis(response.getExpiresIn()))
                .tokenType(encrypt(response.getTokenType()))
                .scope(encrypt(response.getScope()))
                .tokenId(encrypt(response.getTokenId()));
    }

    @Override
    public TokenSpec<GenericTokenSpec> createExchangeTokenSpec(ServiceId artifactoryServiceId, String useCasePath) {
        assertCurationEnabled();
        return super.createExchangeTokenSpec(artifactoryServiceId, useCasePath);
    }

    @Override
    public ExchangeTokenRequest createExchangeTokenRequest(String subject, JwtAccessToken jwt) {
        assertCurationEnabledWithTrust();
        String updatedUseCasePath = getUseCaseFromPairScope(jwt.getScope());
        return new ExchangeTokenRequest(subject, updatedUseCasePath);
    }

    @Override
    public ExchangeTokenRequest createMasterTokenExchangeRefreshRequest(String subject) {
        assertCurationEnabledWithTrust();
        String namespace = getCurationAddon().getNamespace();
        return new ExchangeTokenRequest(subject,
                String.join("/", ServiceTrustUseCase.CURATION.getName(), namespace));
    }

    @Override
    public ExchangeTokenRequest createScopedTokenExchangeRequest(String subject, String... scopes) {
        assertCurationEnabledWithTrust();
        String namespace = getCurationAddon().getNamespace();
        String target =
                String.format("%s/%s:%s", ServiceTrustUseCase.CURATION.getName(), namespace, LIMITED_SCOPE);
        return new ExchangeTokenRequest(subject, target);
    }

    @Override
    public Optional<String> getAuthenticationToken(String baseUrl) {
        assertCurationEnabledWithTrust();
        return getCurationAddon().getPairedToken();
    }

    @Override
    public void deleteAuthenticationToken(String baseUrl) {
        assertCurationEnabledWithTrust();
        getCurationAddon().getCurationSettings()
                .map(settings -> JsonUtils.getInstance().readValue(settings, CurationSettings.class))
                .filter(settings -> Objects.equals(settings.jfrogServiceTrustToken().baseUrl(), baseUrl))
                .ifPresent(settings -> getCurationAddon().saveSettings(JsonUtils.getInstance()
                        .writeValueAsString(settings.toBuilder().jfrogServiceTrustToken(null).build())));
    }

    @Override
    public Collection<ServiceTrustToken> getPairedInstances() {
        return getCurationAddon().getPairedInstance().stream()
                .map(url -> ServiceTrustToken.builder().baseUrl(url).build()).toList();
    }

    @NonNull
    @Override
    protected String getUseCasePathForScope(@NonNull String scope) {
        String useCasePath = getUseCaseFromPairScope(Collections.singletonList(scope));
        return ServiceTrustUseCase.CURATION.matches(useCasePath) ? generateWithNamespace(useCasePath) : useCasePath;
    }

    @Override
    @NonNull
    protected String getUseCaseFromPairScope(@NonNull List<String> scopes) {
        assertScopesExist(scopes);
        return scopes.stream()
                .filter(scope -> scope.contains(ServiceTrustUseCase.CURATION.getName()))
                .findFirst()
                .map(this::getSimpleScope)
                .orElseThrow(() -> new IllegalArgumentException("Pair's token does not have a curation related scope"));
    }

    private void assertCurationEnabled() {
        if (!getCurationAddon().isCurationAsService()) {
            throw new IllegalStateException("Curation is not enabled on this instance");
        }
    }

    private void assertCurationEnabledWithTrust() {
        if (!getCurationAddon().isCurationWithTrustEnabled()) {
            throw new IllegalStateException("Curation is not enabled on this instance");
        }
    }

    private String generateWithNamespace(@NonNull String useCaseWithoutNamespace) {
        CurationAddon curationAddon = getCurationAddon();
        return String.join("/", useCaseWithoutNamespace, curationAddon.generateNamespace());
    }

    private CurationAddon getCurationAddon() {
        return ContextHelper.get().beanForType(AddonsManager.class)
                .addonByType(CurationAddon.class);
    }

    private void assertInstanceNotPaired() {
        if (getCurationAddon().getPairedToken().isPresent()) {
            throw new IllegalStateException(INSTANCES_ALREADY_PAIRED);
        }
    }
}
