package org.artifactory.hqc.aop;

import org.aopalliance.intercept.MethodInvocation;
import org.apache.commons.lang3.RandomStringUtils;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.hqc.annotation.HeavyQueryCache;
import org.artifactory.hqc.config.HeavyQueryCacheConfig;
import org.artifactory.hqc.enums.HeavyQueryCacheType;
import org.artifactory.hqc.service.HeavyQueryCacheService;
import org.artifactory.spring.ArtifactoryApplicationContext;
import org.artifactory.test.ArtifactoryHomeBoundJUnitTest;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;

import java.util.Arrays;
import java.util.concurrent.Callable;
import java.util.stream.Stream;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

/**
 * @author Andrey Tashlitsky
 */
class HeavyQueryCacheAdviceTest extends ArtifactoryHomeBoundJUnitTest {

    private static final String MIN_CALLABLE_DURATION_MILLIS_PROPERTY =
            "artifactory.hqc.total.artifact.count.min.callable.duration.millis";
    private static final String EXPIRY_FACTOR_PROPERTY = "artifactory.hqc.total.artifact.count.expiry.factor";
    private static final String EXPIRY_UPPER_BOUND_PROPERTY = "artifactory.hqc.total.artifact.count.expiry.upper.bound";
    private static final String EXPIRY_LOWER_BOUND_PROPERTY = "artifactory.hqc.total.artifact.count.expiry.lower.bound";
    private final HeavyQueryCacheAdvice heavyQueryCacheAdvice = new HeavyQueryCacheAdvice();
    @Mock
    private ArtifactoryApplicationContext applicationContext;
    @Mock
    private HeavyQueryCacheService hqcService;
    @Mock
    private MethodInvocation methodInvocation;
    private AutoCloseable autoCloseable;

    @BeforeEach
    public void beforeEach() {
        autoCloseable = MockitoAnnotations.openMocks(this);
        when(applicationContext.beanForType(HeavyQueryCacheService.class)).thenReturn(hqcService);
        Stream.of(MIN_CALLABLE_DURATION_MILLIS_PROPERTY, EXPIRY_FACTOR_PROPERTY, EXPIRY_UPPER_BOUND_PROPERTY,
                EXPIRY_LOWER_BOUND_PROPERTY).forEach(System::clearProperty);
        ArtifactoryContextThreadBinder.bind(applicationContext);
    }

    @AfterEach
    public void afterEach() throws Exception {
        ArtifactoryContextThreadBinder.unbind();
        autoCloseable.close();
    }

    @Test
    void testGetName() {
        int size = 254 - HeavyQueryCacheType.TOTAL_ARTIFACT_COUNT.toString().length() - 2;
        String randomLongString = RandomStringUtils.randomAlphanumeric(size);
        when(methodInvocation.getArguments()).thenReturn(new Object[] {randomLongString, 5});
        String name = heavyQueryCacheAdvice.getName(methodInvocation, HeavyQueryCacheType.TOTAL_ARTIFACT_COUNT);
        String expectedName = randomLongString + "_" + 5;
        assertEquals(expectedName, name);
    }

    @Test
    void testGetNameTooLong() {
        int size = 255 - HeavyQueryCacheType.TOTAL_ARTIFACT_COUNT.toString().length() - 2;
        String randomLongString = RandomStringUtils.randomAlphanumeric(size);
        when(methodInvocation.getArguments()).thenReturn(new Object[] {randomLongString, 5});
        String name = heavyQueryCacheAdvice.getName(methodInvocation, HeavyQueryCacheType.TOTAL_ARTIFACT_COUNT);
        String expectedName = String.valueOf(Arrays.hashCode(new Object[] {randomLongString, 5}));
        assertEquals(expectedName, name);
    }

    @Test
    void testParamsToConfigDefaultValues() throws Throwable {
        HeavyQueryCache hqc = Dummy.class.getDeclaredMethod("m1").getAnnotation(HeavyQueryCache.class);
        HeavyQueryCacheConfig heavyQueryCacheConfig = heavyQueryCacheAdvice.paramsToConfig(hqc);
        assertEquals(1000, heavyQueryCacheConfig.getMinCallableDurationMillis());
        assertEquals(1024, heavyQueryCacheConfig.getMaxBlobSizeKb());
        assertEquals(100, heavyQueryCacheConfig.getExpiryFactor());
        assertEquals(Integer.MAX_VALUE, heavyQueryCacheConfig.getExpiryUpperBound());
        assertEquals(Integer.MIN_VALUE, heavyQueryCacheConfig.getExpiryLowerBound());
    }

    @Test
    void testParamsToConfigValues() throws Throwable {
        HeavyQueryCache hqc = Dummy.class.getDeclaredMethod("m2").getAnnotation(HeavyQueryCache.class);
        HeavyQueryCacheConfig heavyQueryCacheConfig = heavyQueryCacheAdvice.paramsToConfig(hqc);
        assertEquals(0, heavyQueryCacheConfig.getMinCallableDurationMillis());
        assertEquals(2048, heavyQueryCacheConfig.getMaxBlobSizeKb());
        assertEquals(123, heavyQueryCacheConfig.getExpiryFactor());
        assertEquals(10_000, heavyQueryCacheConfig.getExpiryUpperBound());
        assertEquals(0, heavyQueryCacheConfig.getExpiryLowerBound());
    }

    @Test
    void testParamsToConfigOverriddenValues() throws Throwable {
        System.setProperty(MIN_CALLABLE_DURATION_MILLIS_PROPERTY, "7");
        System.setProperty(EXPIRY_FACTOR_PROPERTY, "456");
        System.setProperty(EXPIRY_UPPER_BOUND_PROPERTY, "999");
        System.setProperty(EXPIRY_LOWER_BOUND_PROPERTY, "99");
        HeavyQueryCache hqc = Dummy.class.getDeclaredMethod("m2").getAnnotation(HeavyQueryCache.class);
        HeavyQueryCacheConfig heavyQueryCacheConfig = heavyQueryCacheAdvice.paramsToConfig(hqc);
        assertEquals(7, heavyQueryCacheConfig.getMinCallableDurationMillis());
        assertEquals(2048, heavyQueryCacheConfig.getMaxBlobSizeKb());
        assertEquals(456, heavyQueryCacheConfig.getExpiryFactor());
        assertEquals(999, heavyQueryCacheConfig.getExpiryUpperBound());
        assertEquals(99, heavyQueryCacheConfig.getExpiryLowerBound());
    }

    @Test
    void testInvoke() throws Throwable {
        Object o = new Object();
        when(methodInvocation.getArguments()).thenReturn(new Object[] {"cool", 42, o});
        when(methodInvocation.getMethod()).thenReturn(Dummy.class.getDeclaredMethod("m2"));
        when(methodInvocation.proceed()).thenReturn(1);
        heavyQueryCacheAdvice.invoke(methodInvocation);
        HeavyQueryCacheConfig hqcConfig = HeavyQueryCacheConfig.builder()
                .minCallableDurationMillis(0)
                .maxBlobSizeKb(2048)
                .expiryFactor(123)
                .expiryUpperBound(10_000)
                .expiryLowerBound(0)
                .build();
        ArgumentCaptor<Callable<Integer>> callable = ArgumentCaptor.forClass(CallableInteger.class);
        verify(hqcService, times(1)).get(
                callable.capture(),
                eq(int.class),
                eq("cool_42_" + o),
                eq(HeavyQueryCacheType.TOTAL_ARTIFACT_COUNT),
                eq(hqcConfig)
        );
        assertEquals(1, callable.getValue().call());
        verify(methodInvocation, times(1)).proceed();
    }

    private interface Dummy {

        @HeavyQueryCache(hqcType = HeavyQueryCacheType.TOTAL_ARTIFACT_COUNT)
        int m1();

        @HeavyQueryCache(
                hqcType = HeavyQueryCacheType.TOTAL_ARTIFACT_COUNT,
                minCallableDurationMillis = 0,
                maxBlobSizeKb = 2048,
                expiryFactor = 123,
                expiryUpperBound = 10_000,
                expiryLowerBound = 0
        )
        int m2();
    }

    private static abstract class CallableInteger implements Callable<Integer> {

    }
}
