package org.artifactory.ph;

import org.artifactory.common.repo.PackageType;
import com.jfrog.ph.acl.exception.PackageException;
import com.jfrog.ph.acl.service.PackageSpecificationAdapter;
import org.artifactory.descriptor.repo.RepoType;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.mockito.Mock;
import org.mockito.Mockito;

import java.util.stream.Stream;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.never;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.mockito.MockitoAnnotations.openMocks;

class PackageSpecificationImplTest {

    private PackageSpecificationImpl packageIndexChecker;

    @Mock
    private PackageSpecificationAdapter packageSpecificationAdapter;

    private AutoCloseable mocks;


    @BeforeEach
    void setup() {
        mocks = openMocks(this);
        packageIndexChecker = new PackageSpecificationImpl(packageSpecificationAdapter);
    }

    @AfterEach
    void teardown() throws Exception {
        mocks.close();
    }

    @Test
    void testIsExpirable() throws PackageException {
        when(packageSpecificationAdapter.isExpirable(any(), any())).thenReturn(false);
        when(packageSpecificationAdapter.isSupportedPackage(any(PackageType.class))).thenReturn(true);

        Assertions.assertFalse(packageIndexChecker.isExpirable(RepoType.CRAN, "repoKey", "a/b/c"));
        verify(packageSpecificationAdapter, times(1)).isExpirable(PackageType.CRAN, "a/b/c");
        Mockito.reset(packageSpecificationAdapter);
        when(packageSpecificationAdapter.isExpirable(any(), any())).thenReturn(true);
        when(packageSpecificationAdapter.isSupportedPackage(any(PackageType.class))).thenReturn(true);
        Assertions.assertTrue(packageIndexChecker.isExpirable(RepoType.CRAN, "repoKey", "a/b/c"));
        verify(packageSpecificationAdapter, times(1)).isExpirable(PackageType.CRAN, "a/b/c");
    }

    @Test
    void testIsLocalGenerated() throws PackageException {
        when(packageSpecificationAdapter.isLocalGenerated(any(), any())).thenReturn(false);
        when(packageSpecificationAdapter.isSupportedPackage(any(PackageType.class))).thenReturn(true);
        Assertions.assertFalse(packageIndexChecker.isLocalGenerated(RepoType.CRAN, "repoKey", "a/b/c"));
        verify(packageSpecificationAdapter, times(1)).isLocalGenerated(PackageType.CRAN, "a/b/c");
        Mockito.reset(packageSpecificationAdapter);
        when(packageSpecificationAdapter.isLocalGenerated(any(), any())).thenReturn(true);
        when(packageSpecificationAdapter.isSupportedPackage(any(PackageType.class))).thenReturn(true);
        Assertions.assertTrue(packageIndexChecker.isLocalGenerated(RepoType.CRAN, "repoKey", "a/b/c"));
        verify(packageSpecificationAdapter, times(1)).isLocalGenerated(PackageType.CRAN, "a/b/c");
    }

    @Test
    void testShouldExcludePropertyFromMirroring() throws PackageException {
        when(packageSpecificationAdapter.isLocallyGeneratedProperty(PackageType.CRAN, "repoKey")).thenReturn(false);
        when(packageSpecificationAdapter.isSupportedPackage(any(PackageType.class))).thenReturn(true);
        Assertions.assertFalse(packageIndexChecker.shouldExcludePropertyFromMirroring(RepoType.CRAN, "repoKey"));
        verify(packageSpecificationAdapter, times(1)).isLocallyGeneratedProperty(PackageType.CRAN, "repoKey");
        Mockito.reset(packageSpecificationAdapter);
        when(packageSpecificationAdapter.isLocallyGeneratedProperty(PackageType.CRAN, "repoKey")).thenReturn(true);
        when(packageSpecificationAdapter.isSupportedPackage(any(PackageType.class))).thenReturn(true);
        Assertions.assertTrue(packageIndexChecker.shouldExcludePropertyFromMirroring(RepoType.CRAN, "repoKey"));
        verify(packageSpecificationAdapter, times(1)).isLocallyGeneratedProperty(PackageType.CRAN, "repoKey");
    }

    @Test
    void testIsPackageArtifact() throws PackageException {
        when(packageSpecificationAdapter.isSupportedPackage(any(String.class))).thenReturn(true);
        Assertions.assertFalse(packageIndexChecker.isPackageArtifact("repoKey", "path"));
        verify(packageSpecificationAdapter, times(1)).isPackageArtifact("repoKey", "path");
        Mockito.reset(packageSpecificationAdapter);

        when(packageSpecificationAdapter.isSupportedPackage(any(String.class))).thenReturn(true);
        when(packageSpecificationAdapter.isPackageArtifact("repoKey", "path")).thenReturn(true);
        Assertions.assertTrue(packageIndexChecker.isPackageArtifact("repoKey", "path"));
        verify(packageSpecificationAdapter, times(1)).isPackageArtifact("repoKey", "path");
    }

    @ParameterizedTest
    @MethodSource("testIsPackageArtifactFallbackProvider")
    void testIsPackageArtifactFallback(RepoType repoType, String path, boolean expectedResult)
            throws PackageException {
        when(packageSpecificationAdapter.isSupportedPackage(any(String.class))).thenReturn(false);
        Assertions.assertEquals(expectedResult, packageIndexChecker.isPackageArtifact("repoKey", path, repoType));
        verify(packageSpecificationAdapter, never()).isPackageArtifact("repoKey", path);
    }

    private static Stream<Arguments> testIsPackageArtifactFallbackProvider() {
        return Stream.of(Arguments.of(RepoType.Npm, "path.tgz", true),
                Arguments.of(RepoType.Npm, "path.zip", false),
                Arguments.of(RepoType.NuGet, "path.tgz", false),
                Arguments.of(RepoType.NuGet, "path.nupkg", true)
        );
    }

}