package org.artifactory.queue;

import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.HaAddon;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.common.property.ArtifactorySystemProperties;
import org.artifactory.event.queue.QueueStatus;
import org.artifactory.queue.services.PersistentQueueCursorServiceImpl;
import org.artifactory.storage.db.event.entity.enums.NodeEventCursorStatus;
import org.artifactory.storage.db.event.entity.enums.NodeEventCursorType;
import org.artifactory.storage.db.event.model.NodeEventCursor;
import org.artifactory.storage.db.event.service.NodeEventCursorService;
import org.artifactory.storage.event.EventsService;
import org.jfrog.common.ClockUtils;
import org.jfrog.common.concurrency.ConflictsGuard;
import org.joda.time.DateTime;
import org.mockito.ArgumentCaptor;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.stream.IntStream;

import static org.artifactory.event.queue.QueueCursorType.FEDERATED_REPLICATION;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyLong;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.mockito.MockitoAnnotations.openMocks;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertNotNull;


public class PersistentQueueCursorServiceImplTest {

    @Mock
    private NodeEventCursorService nodeEventCursorService;

    @Mock
    private EventsService eventsService;

    @Mock
    private AddonsManager addonsManager;

    @Mock
    private ArtifactoryHome home;

    @Mock
    private ArtifactorySystemProperties props;

    @Mock
    private HaAddon haAddon;

    @Mock
    private ConflictsGuard conflictsGuard;

    private PersistentQueueCursorServiceImpl cursorService;

    @BeforeMethod
    private void setup() {
        openMocks(this);

        ArtifactoryHome.bind(home);
        when(home.getArtifactoryProperties()).thenReturn(props);
        when(props.getLongProperty(eq(ConstantValues.persistentQueueRepoMinEventCacheExpireSec))).thenReturn(0L);
        when(props.getLongProperty(eq(ConstantValues.lastEventIdCacheTimeoutSec))).thenReturn(30L);
        when(addonsManager.addonByType(eq(HaAddon.class))).thenReturn(haAddon);
        when(haAddon.getConflictsGuard(anyString())).thenReturn(conflictsGuard);

        cursorService = new PersistentQueueCursorServiceImpl(nodeEventCursorService, eventsService, addonsManager);

    }

    @AfterMethod
    private void teardown() {
        ArtifactoryHome.unbind();
    }

    @Test
    public void testGetStatusNotAvailable() {


        when(nodeEventCursorService.cursorForOperator(eq("cursor1"))).thenReturn(null);

        QueueStatus status = cursorService.getStatus("repo1", "cursor1");

        assertEquals(status, QueueStatus.NOT_AVAILABLE);

    }

    @Test
    public void testCreateMarkerIfNotExist() throws InterruptedException {
        when(nodeEventCursorService.cursorForOperator("curName")).thenReturn(null);
        when(conflictsGuard.tryToLock(anyString(), anyLong(), any())).thenReturn(true);
        when(nodeEventCursorService.cursorForOperator("curName")).thenReturn(null);

        cursorService.createMarkerIfNotExists("repo", "curName", FEDERATED_REPLICATION);
        ArgumentCaptor<NodeEventCursor> argument = ArgumentCaptor.forClass(NodeEventCursor.class);
        verify(nodeEventCursorService, times(1)).insertCursor(argument.capture());
        assertEquals(argument.getValue().getOperatorId(), "curName");
        assertEquals(-1L, argument.getValue().getEventMarker().longValue());
        assertEquals(argument.getValue().getStatus(), NodeEventCursorStatus.OUT_OF_SYNC);
        assertEquals(argument.getValue().getType(), NodeEventCursorType.FEDERATED_REPLICATION);
    }

    @Test(description = "test copy cursor should copy the marker and the status")
    public void copyCursor(){
        // mocks
        when(nodeEventCursorService.cursorForOperator("curName")).thenReturn(null);
        NodeEventCursor mockNodeEventCursor = mock(NodeEventCursor.class);
        when(mockNodeEventCursor.getEventMarker()).thenReturn(2L);
        when(mockNodeEventCursor.getStatus()).thenReturn(NodeEventCursorStatus.SYNC);

        // run function
        cursorService.copyCursor("repo", "curName", mockNodeEventCursor);

        // test
        ArgumentCaptor<NodeEventCursor> argument = ArgumentCaptor.forClass(NodeEventCursor.class);
        verify(nodeEventCursorService, times(1)).insertCursor(argument.capture());
        assertEquals(argument.getValue().getOperatorId(), "curName");
        assertEquals(2, argument.getValue().getEventMarker().longValue());
        assertEquals(NodeEventCursorStatus.SYNC, argument.getValue().getStatus());
    }

    @Test(description = "test copy cursor should not copy cursor if it already exist")
    public void copyCursorNotFound(){
        // mocks
        NodeEventCursor mockNodeEventCursor = mock(NodeEventCursor.class);
        when(nodeEventCursorService.cursorForOperator("curName")).thenReturn(mockNodeEventCursor);
        when(mockNodeEventCursor.getEventMarker()).thenReturn(2L);
        when(mockNodeEventCursor.getStatus()).thenReturn(NodeEventCursorStatus.SYNC);

        // run function
        cursorService.copyCursor("repo", "curName", mockNodeEventCursor);

        // test
        ArgumentCaptor<NodeEventCursor> argument = ArgumentCaptor.forClass(NodeEventCursor.class);
        verify(nodeEventCursorService, times(0)).insertCursor(argument.capture());
    }
    @Test
    public void testLastEventSupplier() throws InterruptedException {
        when(eventsService.getLastEventLogId()).thenReturn(5L);
        ExecutorService consumerExecutor = Executors.newFixedThreadPool(5);
        List<Callable<Long>> executables = new LinkedList<>();
        IntStream.range(0, 10).forEach(i ->
                executables.add(() -> cursorService.getCachedLatestEventId())
        );

        List<Future<Long>> answers = consumerExecutor.invokeAll(executables);

        answers.forEach(lastEvent -> {
            try {
                assertNotNull(lastEvent.get());
                assertEquals(lastEvent.get().longValue(), 5L);
            } catch (InterruptedException | ExecutionException e) {
                throw new RuntimeException("Failed test for event supplier duo to:", e);
            }
        });
        verify(eventsService, times(1)).getLastEventLogId();
    }

    @DataProvider
    private Object[][] testGetStatusProvider() {
        return new Object[][] {
                {1L, 3L, QueueStatus.DATA_LOST},
                {10L, 5L, QueueStatus.OK},
                {1L, 2L, QueueStatus.OK},
                {2L, 2L, QueueStatus.OK},
        };
    }

    @Test(dataProvider = "testGetStatusProvider")
    public void testGetStatus(long marker, long firstEventId, QueueStatus expectedStatus) {
        NodeEventCursor cursor = Mockito.mock(NodeEventCursor.class);
        when(nodeEventCursorService.cursorForOperator(eq("cursor1"))).thenReturn(cursor);
        when(cursor.getEventMarker()).thenReturn(marker);
        when(eventsService.getFirstEventLogId(eq("repo1"))).thenReturn(firstEventId);
        QueueStatus status = cursorService.getStatus("repo1", "cursor1");
        assertEquals(status, expectedStatus);
    }

    @Test
    void testUpdateOverFlowErrors() {
        cursorService.updateCursorAsOutOfSync("cursor");
        verify(nodeEventCursorService, times(1)).updateCursorWithStatus("cursor", NodeEventCursorStatus.OUT_OF_SYNC);
    }

    @Test
    void testGetCurrentMarkerWithNewCursor() {
        when(nodeEventCursorService.cursorForOperator(anyString())).thenReturn(null);
        NodeEventCursor nodeEventCursor = cursorService.getCursorAndCreateIfNeeded(
                "cursor",
                FEDERATED_REPLICATION
        );
        assertEquals(nodeEventCursor.getEventMarker().longValue(), -1L);
        assertEquals(nodeEventCursor.getOperatorId(), "cursor");
        assertEquals(nodeEventCursor.getStatus(), NodeEventCursorStatus.OUT_OF_SYNC);
    }

    @Test
    void testGetCurrentMarker() {
        when(nodeEventCursorService.cursorForOperator(anyString()))
                .thenReturn(NodeEventCursor.builder()
                        .operatorId("cursor").build());
        NodeEventCursor nodeEventCursor = cursorService.getCursorAndCreateIfNeeded(
                "cursor",
                FEDERATED_REPLICATION
        );

        assertEquals(nodeEventCursor.getOperatorId(), "cursor");
    }

    @Test
    void testUpdateMarkerByTimestamp() {
        ArtifactorySystemProperties properties = mock(ArtifactorySystemProperties.class);
        when(home.getArtifactoryProperties()).thenReturn(properties);
        when(properties.getLongProperty(any())).thenReturn(0L);
        cursorService.updateMarkerByTimestamp("cursor", 1L, 2L);
        verify(nodeEventCursorService, times(1))
                .updateCursorMarker("cursor", 1L, false);
    }

    @Test
    void testUpdateMarkerByTimestampWithNewTime() {
        ArtifactorySystemProperties properties = mock(ArtifactorySystemProperties.class);
        when(home.getArtifactoryProperties()).thenReturn(properties);
        when(properties.getLongProperty(any())).thenReturn(2L);
        long timestamp = new DateTime(ClockUtils.epochMillis()).minusDays(1).getMillis();
        cursorService.updateMarkerByTimestamp("cursor", 1L, timestamp);
        verify(nodeEventCursorService, times(1))
                .updateCursorMarker("cursor", 1L, true);
    }

    @Test
    public void testResetMarker() {
        when(eventsService.getLastEventLogId()).thenReturn(2L);
        cursorService.resetMarker("repo", "cursor");
        verify(nodeEventCursorService, times(1)).updateCursor("cursor", 2L);
    }
}
