/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.repo;

import org.apache.commons.compress.utils.Sets;
import org.apache.commons.io.IOUtils;
import org.apache.http.HttpHeaders;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.utils.DateUtils;
import org.apache.http.message.BasicHeader;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.HaAddon;
import org.artifactory.addon.LayoutsCoreAddon;
import org.artifactory.addon.RestCoreAddon;
import org.artifactory.addon.curation.CurationAddon;
import org.artifactory.addon.plugin.PluginsAddon;
import org.artifactory.addon.xray.XrayAddon;
import org.artifactory.aol.AolService;
import org.artifactory.api.common.BasicStatusHolder;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.repo.ResearchService;
import org.artifactory.api.repo.exception.RepoRejectException;
import org.artifactory.api.request.InternalArtifactoryRequest;
import org.artifactory.api.request.TranslatedArtifactoryRequest;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.checksum.ChecksumInfo;
import org.artifactory.checksum.ChecksumType;
import org.artifactory.checksum.ChecksumsInfo;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.StatusHolder;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.common.property.ArtifactorySystemProperties;
import org.artifactory.descriptor.config.CentralConfigDescriptorImpl;
import org.artifactory.descriptor.repo.HttpRepoDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.fs.RepoResource;
import org.artifactory.io.checksum.ChecksumUtils;
import org.artifactory.md.Properties;
import org.artifactory.model.RemoteRepoConfig;
import org.artifactory.model.xstream.fs.FileInfoImpl;
import org.artifactory.model.xstream.fs.PropertiesImpl;
import org.artifactory.opentracing.ArtifactoryTracer;
import org.artifactory.repo.db.DbCacheRepo;
import org.artifactory.repo.db.DbStoringRepoMixin;
import org.artifactory.repo.http.ConnectionManagersHolder;
import org.artifactory.repo.http.DenyListCacheHolderService;
import org.artifactory.repo.http.HttpConnectionPoolMonitorService;
import org.artifactory.repo.http.IdleConnectionMonitorService;
import org.artifactory.repo.http.IdleConnectionMonitorServiceImpl;
import org.artifactory.repo.remote.interceptor.ExpiryAction;
import org.artifactory.repo.remote.interceptor.RemoteRepoInterceptor;
import org.artifactory.repo.service.InternalRepositoryService;
import org.artifactory.repo.service.deploy.UIBuildArtifactoryRequest;
import org.artifactory.request.InternalRequestContext;
import org.artifactory.request.Request;
import org.artifactory.request.RequestContext;
import org.artifactory.resource.ExpiredRepoResource;
import org.artifactory.resource.FileResource;
import org.artifactory.resource.RemoteRepoResource;
import org.artifactory.resource.RepoResourceInfo;
import org.artifactory.resource.ResourceStreamHandle;
import org.artifactory.resource.ResourceWithChecksumStreamHandle;
import org.artifactory.resource.UnfoundRepoResource;
import org.artifactory.resource.UnfoundRepoResourceReason.Reason;
import org.artifactory.spring.InternalArtifactoryContext;
import org.artifactory.storage.fs.service.FileService;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.artifactory.util.RepoLayoutUtils;
import org.artifactory.version.ArtifactoryServiceVersion;
import org.artifactory.webapp.servlet.HttpArtifactoryRequest;
import org.jfrog.common.concurrency.ConflictsGuard;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;
import org.springframework.test.util.ReflectionTestUtils;
import org.testng.Assert;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.TimeUnit;

import static javax.ws.rs.core.Response.Status.TOO_MANY_REQUESTS;
import static org.artifactory.request.ArtifactoryRequest.CHECKSUM_MD5;
import static org.artifactory.request.ArtifactoryRequest.CHECKSUM_SHA1;
import static org.artifactory.request.ArtifactoryRequest.CHECKSUM_SHA256;
import static org.artifactory.request.ArtifactoryRequest.PARAM_ALTERNATIVE_REMOTE_DOWNLOAD_URL;
import static org.artifactory.request.ArtifactoryRequest.PARAM_ALTERNATIVE_REMOTE_SITE_URL;
import static org.artifactory.request.ArtifactoryRequest.PARAM_REPLICATION_ORIGINATED_DOWNLOAD_REQUEST;
import static org.fest.assertions.Assertions.assertThat;
import static org.mockito.Mockito.any;
import static org.mockito.Mockito.anyBoolean;
import static org.mockito.Mockito.anyString;
import static org.mockito.Mockito.doAnswer;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.doThrow;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.reset;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.testng.AssertJUnit.assertEquals;
import static org.testng.AssertJUnit.assertFalse;
import static org.testng.AssertJUnit.assertNull;
import static org.testng.AssertJUnit.assertTrue;

/**
 * This class tests the behaviour of the allowsDownloadMethod and checks that it returns the proper results on different
 * scenarios
 *
 * @author Noam Tenne
 */
@Test
public class RemoteRepoBaseTest extends ArtifactoryHomeBoundTest {

    @Mock
    private AuthorizationService authService;
    @Mock
    private InternalRepositoryService internalRepoService;
    @Mock
    private ResearchService researchService;
    @Mock
    private InternalArtifactoryContext context;
    @Mock
    private FileService fileService;
    @Mock
    private AolService aolService;
    @Mock
    private AddonsManager addonsManager;
    @Mock
    private LayoutsCoreAddon layoutsCoreAddon;
    @Mock
    private HaAddon haAddon;
    @Mock
    private XrayAddon xrayAddon;
    @Mock
    private CurationAddon curationAddon;
    @Mock
    private RestCoreAddon restCoreAddon;
    @Mock
    private PluginsAddon pluginsAddon;
    @Mock
    private InternalRequestContext requestContext;
    @Mock
    private Request request;
    @Mock
    private CloseableHttpResponse response;
    @Mock
    private HttpRepo.TrafficAwareRemoteResourceStreamHandle trafficAwareRemoteResourceStreamHandle;
    @Mock
    private CentralConfigService configService;

    private HttpRepoDescriptor httpRepoDescriptor = new HttpRepoDescriptor();
    private RemoteRepoConfig remoteRepoConfig;
    private HttpRepo httpRepo;
    private RepoPath repoPath = InternalRepoPathFactory.create("remote-repo-cache", "test/test/1.0/test-1.0.jar");
    private RepoPath nonCachedRepoPath = InternalRepoPathFactory.create("remote-repo", "test/test/1.0/test-1.0.jar");
    private String timeStr = "Tue, 17 Feb 2015 13:15:35 GMT";
    private ChecksumsInfo checksumsInfo;

    @BeforeMethod
    public void setup() {
        remoteRepoConfig = RemoteRepoConfig.create("remoteRepoKey", RepoType.Generic);
        System.setProperty(ConstantValues.artifactoryProductVersion.getPropertyName(),
                ArtifactoryServiceVersion.getCurrent().getVersion());
        MockitoAnnotations.initMocks(this);
        createOriginalChecksum();
        when(context.getAuthorizationService()).thenReturn(authService);
        when(context.beanForType(AddonsManager.class)).thenReturn(addonsManager);
        when(aolService.isAol()).thenReturn(false);
        when(context.beanForType(AolService.class)).thenReturn(aolService);
        when(context.beanForType(HttpConnectionPoolMonitorService.class)).thenReturn(
                mock(HttpConnectionPoolMonitorService.class));
        when(fileService.exists(any())).thenReturn(true);
        when(context.beanForType(FileService.class)).thenReturn(fileService);
        when(addonsManager.addonByType(LayoutsCoreAddon.class)).thenReturn(layoutsCoreAddon);
        when(addonsManager.addonByType(CurationAddon.class)).thenReturn(mock(CurationAddon.class));
        when(addonsManager.addonByType(HaAddon.class)).thenReturn(haAddon);
        when(addonsManager.addonByType(XrayAddon.class)).thenReturn(xrayAddon);
        when(addonsManager.addonByType(CurationAddon.class)).thenReturn(curationAddon);
        when(curationAddon.getPackageStatus(any(), any())).thenReturn(new BasicStatusHolder());
        when(addonsManager.addonByType(RestCoreAddon.class)).thenReturn(restCoreAddon);
        when(addonsManager.addonByType(PluginsAddon.class)).thenReturn(pluginsAddon);
        when(context.beanForType(ArtifactoryTracer.class)).thenReturn(mock(ArtifactoryTracer.class));
        IdleConnectionMonitorServiceImpl idleConnectionMonitorService =
                new IdleConnectionMonitorServiceImpl(mock(ConnectionManagersHolder.class));
        when(context.beanForType(IdleConnectionMonitorService.class)).thenReturn(idleConnectionMonitorService);
        when(context.beanForType(DenyListCacheHolderService.class)).thenReturn(
                mock(DenyListCacheHolderService.class));
        when(requestContext.getRequest()).thenReturn(request);
        when(request.getRepoPath()).thenReturn(repoPath);
        ArtifactoryContextThreadBinder.bind(context);
        CentralConfigDescriptorImpl centralConfigDescriptor = new CentralConfigDescriptorImpl();
        centralConfigDescriptor.setRepoLayouts(List.of(RepoLayoutUtils.MAVEN_2_DEFAULT));
        when(configService.getDescriptor()).thenReturn(centralConfigDescriptor);
        remoteRepoConfig.getBaseConfig().setRepoLayoutRef(RepoLayoutUtils.MAVEN_2_DEFAULT.getName());
        remoteRepoConfig.getRepoTypeConfig().setUrl("http://someUrl");
        httpRepo =
                spy(new HttpRepo(httpRepoDescriptor, remoteRepoConfig, internalRepoService, addonsManager,
                        researchService,  null, null, configService));
    }

    @AfterMethod
    public void tearDown() {
        ArtifactoryContextThreadBinder.unbind();
    }

    private void createOriginalChecksum() {
        String sha1 = "f8237d8959e03355010bb85cc3dc46a46fb31110";
        String sha256 = "0f8eb4b72b6e0c9e88b388eb967b49e067ef1004bf07bffc22c3acb13b43580a";
        String md5 = "a35fe7f7fe8217b4369a0af4244d1fca";

        ChecksumInfo sha1ChecksumInfo = new ChecksumInfo(ChecksumType.sha1, sha1, null);
        ChecksumInfo sha256ChecksumInfo = new ChecksumInfo(ChecksumType.sha256, sha256, null);
        ChecksumInfo md5ChecksumInfo = new ChecksumInfo(ChecksumType.md5, md5, null);
        checksumsInfo = new ChecksumsInfo();
        checksumsInfo.addChecksumInfo(sha1ChecksumInfo);
        checksumsInfo.addChecksumInfo(sha256ChecksumInfo);
        checksumsInfo.addChecksumInfo(md5ChecksumInfo);
        when(trafficAwareRemoteResourceStreamHandle.getResponse())
                .thenReturn(response);
        when(response.getFirstHeader(HttpHeaders.LAST_MODIFIED))
                .thenReturn(new BasicHeader(HttpHeaders.LAST_MODIFIED, "Tue, 17 Feb 2015 13:15:35 GMT"));
        when(response.getFirstHeader(CHECKSUM_SHA1))
                .thenReturn(new BasicHeader(CHECKSUM_SHA1, sha1));
        when(response.getFirstHeader(CHECKSUM_SHA256))
                .thenReturn(new BasicHeader(CHECKSUM_SHA256, sha256));
        when(response.getFirstHeader(CHECKSUM_MD5))
                .thenReturn(new BasicHeader(CHECKSUM_MD5, md5));
        when(response.getFirstHeader(HttpHeaders.CONTENT_LENGTH))
                .thenReturn(new BasicHeader(HttpHeaders.CONTENT_LENGTH, "4"));
        when(response.getFirstHeader(HttpHeaders.ETAG))
                .thenReturn(new BasicHeader(HttpHeaders.ETAG, "etag-1"));

    }

    public void testInternalGetInfoFoundRemoteAndCachedResource() {
        RepoResource remoteResource = new RemoteRepoResource(repoPath, 1234, "etag", 50,
                Sets.newHashSet(), null);
        ExpiredRepoResource cachedResource = new ExpiredRepoResource(new FileResource(
                new FileInfoImpl(repoPath) {{setLastModified(123);}}));
        when(context.beanForType(HttpConnectionPoolMonitorService.class))
                .thenReturn(new HttpConnectionPoolMonitorService());
        httpRepo.init(false);
        HttpRepo httpRepoSpy = spy(this.httpRepo);
        doReturn(cachedResource).when(httpRepoSpy).getInfoFromCache(requestContext);
        doReturn(remoteResource).when(httpRepoSpy)
                .getRemoteResource(any(RequestContext.class), any(RepoPath.class), anyBoolean());

        RepoResource repoResource = httpRepoSpy.internalGetInfo(nonCachedRepoPath, requestContext);
        assertTrue(repoResource instanceof RemoteRepoResource);
        RemoteRepoResource remoteRepoResource = (RemoteRepoResource) repoResource;
        assertEquals(remoteRepoResource.getCorrespondingCachedResource(), cachedResource);
    }

    public void testInternalGetInfoUnfoundResource() {
        RepoResource remoteResource = new UnfoundRepoResource(repoPath, TOO_MANY_REQUESTS.getReasonPhrase(),
                TOO_MANY_REQUESTS.getStatusCode());
        ExpiredRepoResource cachedResource = new ExpiredRepoResource(new FileResource(
                new FileInfoImpl(repoPath) {{setLastModified(123);}}));
        when(context.beanForType(HttpConnectionPoolMonitorService.class))
                .thenReturn(new HttpConnectionPoolMonitorService());
        httpRepo.init(false);
        HttpRepo httpRepoSpy = spy(this.httpRepo);
        doReturn(cachedResource).when(httpRepoSpy).getInfoFromCache(requestContext);
        doReturn(remoteResource).when(httpRepoSpy)
                .getRemoteResource(any(RequestContext.class), any(RepoPath.class), anyBoolean());

        RepoResource repoResource = httpRepoSpy.internalGetInfo(nonCachedRepoPath, requestContext);
        assertTrue(repoResource instanceof ExpiredRepoResource);
        assertEquals(repoResource, cachedResource);
    }

    @Test(dataProvider = "testInternalGetInfoBlockedRemoteDownloadProvider")
    public void testInternalGetInfoBlockedRemoteDownload(boolean foundExpiredInCache) {
        RepoResource notFoundResource = new UnfoundRepoResource(repoPath, TOO_MANY_REQUESTS.getReasonPhrase(),
                TOO_MANY_REQUESTS.getStatusCode());
        RepoResource expiredCachedResource = new ExpiredRepoResource(new FileResource(
                new FileInfoImpl(repoPath) {{setLastModified(123);}}));
        when(context.beanForType(HttpConnectionPoolMonitorService.class))
                .thenReturn(new HttpConnectionPoolMonitorService());
        RemoteRepoInterceptor remoteRepoInterceptor = mock(RemoteRepoInterceptor.class);
        when(remoteRepoInterceptor.isRemoteDownloadAllowed(any(), any())).thenReturn(false);
        when(context.beansForType(RemoteRepoInterceptor.class))
                .thenReturn(Map.of("remoteRepoInterceptor", remoteRepoInterceptor));
        httpRepo.init(false);
        HttpRepo httpRepoSpy = spy(this.httpRepo);
        doReturn(foundExpiredInCache ? expiredCachedResource : notFoundResource).when(httpRepoSpy)
                .getInfoFromCache(requestContext);
        RepoResource repoResource = httpRepoSpy.internalGetInfo(nonCachedRepoPath, requestContext);
        if (foundExpiredInCache) {
            assertEquals(repoResource, expiredCachedResource);
        } else {
            assertTrue(repoResource instanceof UnfoundRepoResource);
        }
    }

    /**
     * Try to download while supplying a path with no "-cache" prefix
     */
    public void testDownloadWithNoCachePrefix() {
        RepoPath repoPathNoPrefix = InternalRepoPathFactory.create("remote-repo", "test/test/1.0/test-1.0.jar");
        expectCanRead(true);
        createCacheRepo();
        StatusHolder statusHolder = tryAllowsDownload(repoPathNoPrefix);
        Assert.assertFalse(statusHolder.isError());
        verifyCanRead();
    }

    /**
     * Try to download when Anonymous mode is enabled
     */
    public void testDownloadWithAnonymousEnabledAndNoReadPermissions() {
        expectCanRead(false);       // don't give read access
        expectXrayBlocked();
        createCacheRepo();   // and enable anon
        StatusHolder statusHolder = tryAllowsDownload(repoPath);
        Assert.assertTrue(statusHolder.isError());
        verifyCanRead();
    }

    private void verifyCanRead() {
        verify(authService).canRead(repoPath);
        reset(authService);
    }

    public void testGetResourceOfflineRepoPath() {
        ArtifactoryHome.get().getArtifactoryProperties()
                .setProperty(ConstantValues.useRepoPathOnRemoteOffline.getPropertyName(), Boolean.toString(true));
        assertEquals("repoPath", httpRepo.getResourceOfflineRepoPath("path", "repoPath"));

        ArtifactoryHome.get().getArtifactoryProperties()
                .setProperty(ConstantValues.useRepoPathOnRemoteOffline.getPropertyName(), Boolean.toString(false));
        assertEquals("path", httpRepo.getResourceOfflineRepoPath("path", "repoPath"));
    }

    /**
     * Try to download when the repo is blacked out
     */
    public void testDownloadWithBlackedOut() {
        remoteRepoConfig.getRepoTypeConfig().setBlackedOut(true);
        createCacheRepo();
        StatusHolder statusHolder = tryAllowsDownload(repoPath);
        Assert.assertEquals(statusHolder.getLastError().getMessage(),
                "The repository '" + nonCachedRepoPath.getRepoKey() + "' is blacked out and cannot serve " +
                        "artifact '" + nonCachedRepoPath + "'.");
        remoteRepoConfig.getRepoTypeConfig().setBlackedOut(false);
    }

    /**
     * Try to download when the repo does not serve releases
     */
    public void testDownloadWithNoReleases() {
        remoteRepoConfig.getRepoTypeConfig().setHandleReleases(false);
        createCacheRepo();
        StatusHolder statusHolder = tryAllowsDownload(repoPath);
        Assert.assertEquals(statusHolder.getLastError().getMessage(),
                "The repository '" + nonCachedRepoPath.getRepoKey() + "' rejected the resolution of an artifact " +
                        "'" + nonCachedRepoPath + "' due to conflict in the snapshot release handling policy.");
        remoteRepoConfig.getRepoTypeConfig().setHandleReleases(true);
    }

    /**
     * Try to download when requested resource is excluded
     */
    public void testDownloadWithExcludes() {
        remoteRepoConfig.getBaseConfig().setExcludesPattern("test/test/1.0/**");
        createCacheRepo();
        StatusHolder statusHolder = tryAllowsDownload(repoPath);
        Assert.assertEquals(statusHolder.getLastError().getMessage(),
                "The repository '" + repoPath.getRepoKey() + "' rejected the artifact '"
                        + repoPath + "' due to its include/exclude pattern settings.");
        remoteRepoConfig.getBaseConfig().setExcludesPattern("");
    }

    /**
     * Verify reading of valid checksum file to test the {@link ChecksumUtils#checksumStringFromStream(InputStream)}
     * method
     */
    public void testReadValidChecksum() throws Exception {
        invokeReadChecksum("valid", "dasfasdf4r234234q32asdfadfasasdfasdf");
    }

    /**
     * Verify reading of a checksum file containing comments to test the
     * {@link ChecksumUtils#checksumStringFromStream(InputStream)}
     * method
     */
    public void testReadCommentChecksum() throws Exception {
        invokeReadChecksum("comment", "asdfaeaef435345435asdf");
    }

    /**
     * Verify reading of a checksum file containing file description to test the {@link
     * ChecksumUtils#checksumStringFromStream(InputStream)} method
     */
    public void testReadDescChecksum() throws Exception {
        invokeReadChecksum("desc", "dasfasdf4r234234q32asdfadfasdf");
    }

    /**
     * Verify reading of an empty checksum file to test the {@link ChecksumUtils#checksumStringFromStream(InputStream)}
     * method
     */
    public void testReadEmptyChecksum() throws Exception {
        invokeReadChecksum("empty", "");
    }

    public void returnCachedResourceResourceIsFound() {
        FileResource cachedResource = mock(FileResource.class);
        when(cachedResource.isFound()).thenReturn(true);
        assertThat(httpRepo.shouldReturnCachedResource(requestContext, cachedResource)).isTrue();
    }

    public void returnCachedResourceWhenPropMismatch() {
        UnfoundRepoResource cachedResource = mock(UnfoundRepoResource.class);
        when(cachedResource.getReason()).thenReturn(Reason.PROPERTY_MISMATCH);
        assertThat(httpRepo.shouldReturnCachedResource(requestContext, cachedResource)).isTrue();
    }

    public void returnCachedResourceWhenRemoteReplicationOriginated() {
        RepoResource cachedResource = mock(RepoResource.class);
        when(requestContext.getRequest()).thenReturn(request);
        when(request.getParameter(PARAM_REPLICATION_ORIGINATED_DOWNLOAD_REQUEST)).thenReturn("true");
        assertThat(httpRepo.shouldReturnCachedResource(requestContext, cachedResource)).isTrue();

        when(request.getParameter(PARAM_REPLICATION_ORIGINATED_DOWNLOAD_REQUEST)).thenReturn(null);
        assertThat(httpRepo.shouldReturnCachedResource(requestContext, cachedResource)).isFalse();
    }

    @Test(dataProvider = "providePathsContainsURL")
    public void testAppendAndGetUrlWithURLInPath(String pathWithURL, String expectedIfAllowed) throws IOException {
        try {
            when(layoutsCoreAddon.translateArtifactPath(any(), any(), anyString())).thenReturn(pathWithURL);
            httpRepo.downloadResource(pathWithURL, requestContext);
            assertThat(true).isFalse();
        } catch (RuntimeException e) {
            assertThat(e.getMessage()).contains(pathWithURL);
        }

        assertThat(httpRepo.appendAndGetUrl(pathWithURL, requestContext))
                .isEqualTo(expectedIfAllowed);

    }

    @Test(dataProvider = "provideValidPaths")
    public void testAppendAndGetUrlValidPath(String validPath) {
        assertThat(httpRepo.appendAndGetUrl(validPath, requestContext)).isEqualTo("http://someUrl/" + validPath);
    }

    @DataProvider
    private Object[][] provideValidPaths() {
        return new Object[][] {
                {"some/path/to/append/"},
                {"some/path/to/append/:properties"},
        };
    }

    @DataProvider
    private Object[][] providePathsContainsURL() {
        return new Object[][] {
                {"http://localhost:8080/some/path/to/append/", "http://localhost:8080/some/path/to/append/"},
                {"http://localhost:8080/some/path/to/append/:properties", "http://someUrl/http://localhost:8080/some" +
                        "/path/to/append/:properties"},
        };
    }

    public void testAlternativeRemoteSiteUrl() {
        Request request = mock(UIBuildArtifactoryRequest.class);
        assertAlternateUrl(request, "http://alternativeSite/");

        request = mock(InternalArtifactoryRequest.class);
        assertAlternateUrl(request, "http://alternativeSite/");

        request = mock(HttpArtifactoryRequest.class);
        assertAlternateUrl(request, "http://someUrl/");
    }

    private void assertAlternateUrl(Request request, String expected) {
        when(requestContext.getRequest()).thenReturn(request);
        when(request.getParameter(PARAM_ALTERNATIVE_REMOTE_SITE_URL)).thenReturn("http://alternativeSite");
        assertThat(httpRepo.appendAndGetUrl("", requestContext)).isEqualTo(expected);
    }

    public void testAlternativeRemoteDownloadUrl() {
        Request request = mock(UIBuildArtifactoryRequest.class);
        assertAlternateDownloadUrl(request, "http://alternativeSite/");

        request = mock(InternalArtifactoryRequest.class);
        assertAlternateDownloadUrl(request, "http://alternativeSite/");

        request = mock(HttpArtifactoryRequest.class);
        assertAlternateDownloadUrl(request, "");
    }

    private void assertAlternateDownloadUrl(Request request, String expected) {
        when(request.getParameter(PARAM_ALTERNATIVE_REMOTE_DOWNLOAD_URL)).thenReturn("http://alternativeSite/");
        assertThat(httpRepo.resolvePathForUrl("", request)).isEqualTo(expected);
    }

    public void testCreateRemoteResource() {
        boolean isSavedResource = ConstantValues.saveGetResource.getBoolean();
        ArtifactoryHome.get().getArtifactoryProperties()
                .setProperty(ConstantValues.saveGetResource.getPropertyName(), Boolean.toString(true));
        long timeMillis = DateUtils.parseDate(timeStr, new String[] {"E, d MMM yyyy HH:mm:ss Z"}).getTime();
        HttpRepo httpRepo1 = spy(httpRepo);
        RemoteRepoResource repoResource =
                httpRepo1.createRemoteResourceFromResponse(repoPath, response, checksumsInfo.getChecksums());
        assertTrue(checksumsInfo.isIdentical(repoResource.getInfo().getChecksumsInfo()));
        assertEquals(timeMillis, repoResource.getLastModified());
        assertEquals(4, repoResource.getSize());
        assertEquals(repoPath, repoResource.getRepoPath());
        assertEquals("etag-1", repoResource.getEtag());
        ArtifactoryHome.get().getArtifactoryProperties()
                .setProperty(ConstantValues.saveGetResource.getPropertyName(), Boolean.toString(isSavedResource));
    }

    public void testReconstructResourceFromGetResponseIfNeeded() {
        Set<ChecksumInfo> remoteChecksums = Sets.newHashSet();
        boolean isSavedResource = ConstantValues.saveGetResource.getBoolean();
        ArtifactoryHome.get().getArtifactoryProperties()
                .setProperty(ConstantValues.saveGetResource.getPropertyName(), Boolean.toString(true));

        RemoteRepoResource repoResource = new RemoteRepoResource(repoPath, 123, "etag", 50, remoteChecksums, null);
        long timeMillis = DateUtils.parseDate(timeStr, new String[] {"E, d MMM yyyy HH:mm:ss Z"}).getTime();

        httpRepo.reconstructResourceFromGetResponseIfNeeded(repoResource, trafficAwareRemoteResourceStreamHandle,
                requestContext.getRequest().getRepoPath());

        assertTrue(checksumsInfo.isIdentical(repoResource.getInfo().getChecksumsInfo()));
        assertEquals(timeMillis, repoResource.getLastModified());
        assertEquals(4, repoResource.getSize());
        assertEquals(repoPath, repoResource.getRepoPath());
        Properties properties = new PropertiesImpl();
        properties.put(DbStoringRepoMixin.ETAG_PROP_KEY, repoResource.getEtag());
        properties = httpRepo.updateEtagInProperties(repoResource, properties);
        assertEquals("etag-1", properties.getFirst("artifactory.internal.etag"));
        ArtifactoryHome.get().getArtifactoryProperties()
                .setProperty(ConstantValues.saveGetResource.getPropertyName(), Boolean.toString(isSavedResource));

    }

    public void testReconstructResourceFromGetResponseNotNeeded() {
        Set<ChecksumInfo> remoteChecksums = Sets.newHashSet();
        boolean isSavedResource = ConstantValues.saveGetResource.getBoolean();
        ArtifactoryHome.get().getArtifactoryProperties()
                .setProperty(ConstantValues.saveGetResource.getPropertyName(), Boolean.toString(false));
        RemoteRepoResource repoResource = new RemoteRepoResource(repoPath, 123, "etag", 50, remoteChecksums, null);

        httpRepo.reconstructResourceFromGetResponseIfNeeded(repoResource, trafficAwareRemoteResourceStreamHandle,
                requestContext.getRequest().getRepoPath());

        assertEquals(remoteChecksums, repoResource.getChecksums());
        assertEquals(123, repoResource.getLastModified());
        assertEquals(50, repoResource.getSize());
        assertEquals(repoPath, repoResource.getRepoPath());
        Properties properties = new PropertiesImpl();
        properties.put(DbStoringRepoMixin.ETAG_PROP_KEY, repoResource.getEtag());
        properties = httpRepo.updateEtagInProperties(repoResource, properties);
        assertEquals("etag", properties.getFirst("artifactory.internal.etag"));
        ArtifactoryHome.get().getArtifactoryProperties()
                .setProperty(ConstantValues.saveGetResource.getPropertyName(), Boolean.toString(isSavedResource));

    }

    @Test
    public void shouldExpireTrueWithRemoteAndExpiredResourcesTest() {
        prepInterceptorMocks(ExpiryAction.DONT_EXPIRE_CONTINUE_NEXT, ExpiryAction.DONT_EXPIRE_CONTINUE_NEXT);
        RemoteRepoResource remoteResource = new RemoteRepoResource(repoPath, 1234, "etag", 50, Sets.newHashSet(), null);
        ExpiredRepoResource cachedResource = new ExpiredRepoResource(
                new FileResource(new FileInfoImpl(repoPath) {{setLastModified(123);}}));
        httpRepo.init(false);
        HttpRepo httpRepoSpy = spy(this.httpRepo);

        assertTrue(httpRepoSpy.shouldExpire(cachedResource, remoteResource));
        verify(httpRepoSpy, times(1)).shouldExpireWithLastModified(cachedResource, remoteResource);
    }

    @Test
    public void shouldExpireFalseWithRemoteAndExpiredResourcesButStoppedByInterceptorTest() {
        prepInterceptorMocks(ExpiryAction.DONT_EXPIRE_CONTINUE_NEXT, ExpiryAction.DONT_EXPIRE_STOP);
        RemoteRepoResource remoteResource = new RemoteRepoResource(repoPath, 1234, "etag", 50, Sets.newHashSet(), null);
        ExpiredRepoResource cachedResource = new ExpiredRepoResource(
                new FileResource(new FileInfoImpl(repoPath) {{setLastModified(123);}}));
        httpRepo.init(false);
        HttpRepo httpRepoSpy = spy(this.httpRepo);

        assertFalse(httpRepoSpy.shouldExpire(cachedResource, remoteResource));
        verify(httpRepoSpy, times(0)).shouldExpireWithLastModified(cachedResource, remoteResource);
    }

    @Test
    public void shouldExpireFalseWithRemoteAndExpiredResourcesButExpiredByInterceptorTest() {
        prepInterceptorMocks(ExpiryAction.DONT_EXPIRE_CONTINUE_NEXT, ExpiryAction.EXPIRE);
        RemoteRepoResource remoteResource = new RemoteRepoResource(repoPath, 1234, "etag", 50, Sets.newHashSet(), null);
        ExpiredRepoResource cachedResource = new ExpiredRepoResource(
                new FileResource(new FileInfoImpl(repoPath) {{setLastModified(123);}}));
        httpRepo.init(false);
        HttpRepo httpRepoSpy = spy(this.httpRepo);

        assertTrue(httpRepoSpy.shouldExpire(cachedResource, remoteResource));
        verify(httpRepoSpy, times(0)).shouldExpireWithLastModified(cachedResource, remoteResource);
    }

    @Test
    public void shouldExpireTrueWithRemoteAndExpiredResourcesTestWithDontExpireStop() {
        HashMap<String, RemoteRepoInterceptor> interceptorsMap =
                prepInterceptorMocks(ExpiryAction.DONT_EXPIRE_STOP, ExpiryAction.DONT_EXPIRE_CONTINUE_NEXT);
        RemoteRepoResource remoteResource = new RemoteRepoResource(repoPath, 1234, "etag", 50, Sets.newHashSet(), null);
        ExpiredRepoResource cachedResource = new ExpiredRepoResource(
                new FileResource(new FileInfoImpl(repoPath) {{setLastModified(123);}}));
        httpRepo.init(false);
        HttpRepo httpRepoSpy = spy(this.httpRepo);

        assertFalse(httpRepoSpy.shouldExpire(cachedResource, remoteResource));
        verify(interceptorsMap.get("1"), times(1)).shouldExpireCachedResource(cachedResource, remoteResource);
        verify(interceptorsMap.get("2"), times(0)).shouldExpireCachedResource(cachedResource, remoteResource);
        verify(httpRepoSpy, times(0)).shouldExpireWithLastModified(cachedResource, remoteResource);
    }

    @Test
    public void shouldExpireTrueWithRemoteAndExpiredResourcesTestWithExpire() {
        HashMap<String, RemoteRepoInterceptor> interceptorsMap =
                prepInterceptorMocks(ExpiryAction.EXPIRE, ExpiryAction.DONT_EXPIRE_CONTINUE_NEXT);
        RemoteRepoResource remoteResource = new RemoteRepoResource(repoPath, 1234, "etag", 50, Sets.newHashSet(), null);
        ExpiredRepoResource cachedResource = new ExpiredRepoResource(
                new FileResource(new FileInfoImpl(repoPath) {{setLastModified(123);}}));
        httpRepo.init(false);
        HttpRepo httpRepoSpy = spy(this.httpRepo);

        assertTrue(httpRepoSpy.shouldExpire(cachedResource, remoteResource));
        verify(interceptorsMap.get("1"), times(1)).shouldExpireCachedResource(cachedResource, remoteResource);
        verify(interceptorsMap.get("2"), times(0)).shouldExpireCachedResource(cachedResource, remoteResource);
        verify(httpRepoSpy, times(0)).shouldExpireWithLastModified(cachedResource, remoteResource);
    }

    @Test
    public void shouldExpireFalseWithRemoteAndExpiredResourcesSameLastModifiedTest() {
        HashMap<String, RemoteRepoInterceptor> interceptorsMap =
                prepInterceptorMocks(ExpiryAction.DONT_EXPIRE_CONTINUE_NEXT, ExpiryAction.DONT_EXPIRE_CONTINUE_NEXT);
        RemoteRepoResource remoteResource = new RemoteRepoResource(repoPath, 1234, "etag", 50, Sets.newHashSet(), null);
        ExpiredRepoResource cachedResource = new ExpiredRepoResource(
                new FileResource(new FileInfoImpl(repoPath) {{setLastModified(1234);}}));
        httpRepo.init(false);
        HttpRepo httpRepoSpy = spy(this.httpRepo);

        assertFalse(httpRepoSpy.shouldExpire(cachedResource, remoteResource));
        verify(interceptorsMap.get("1"), times(1)).shouldExpireCachedResource(cachedResource, remoteResource);
        verify(interceptorsMap.get("2"), times(1)).shouldExpireCachedResource(cachedResource, remoteResource);
        verify(httpRepoSpy, times(1)).shouldExpireWithLastModified(cachedResource, remoteResource);
    }

    @Test
    public void shouldExpireTrueTwoInterceptorsReturnExpireTest() {
        HashMap<String, RemoteRepoInterceptor> interceptorsMap =
                prepInterceptorMocks(ExpiryAction.EXPIRE, ExpiryAction.EXPIRE);
        RemoteRepoResource remoteResource = new RemoteRepoResource(repoPath, 1234, "etag", 50, Sets.newHashSet(), null);
        ExpiredRepoResource cachedResource = new ExpiredRepoResource(
                new FileResource(new FileInfoImpl(repoPath) {{setLastModified(1234);}}));
        httpRepo.init(false);
        HttpRepo httpRepoSpy = spy(this.httpRepo);

        assertTrue(httpRepoSpy.shouldExpire(cachedResource, remoteResource));
        verify(interceptorsMap.get("1"), times(1)).shouldExpireCachedResource(cachedResource, remoteResource);
        verify(interceptorsMap.get("2"), times(0)).shouldExpireCachedResource(cachedResource, remoteResource);
        verify(httpRepoSpy, times(0)).shouldExpireWithLastModified(cachedResource, remoteResource);
    }

    @Test
    public void shouldNotCreateOrGetFolderIfExistsTest() {
        httpRepoDescriptor.setKey("remote-repo");
        DbCacheRepo localCacheRepo = new DbCacheRepo(httpRepo, null, configService);

        DbStoringRepoMixin mixinSpy = spy(new DbStoringRepoMixin(httpRepoDescriptor, null, null));
        Mockito.doReturn(null).when(mixinSpy).createOrGetFolder(any());

        ReflectionTestUtils.setField(localCacheRepo, "mixin", mixinSpy);

        localCacheRepo.init(true);

        verify(fileService, times(0)).exists(any());
        verify(mixinSpy, times(0)).createOrGetFolder(any());
    }

    @Test
    public void shouldCreateOrGetFolderIfNotExistsTest() {
        httpRepoDescriptor.setKey("remote-repo");
        DbCacheRepo localCacheRepo = new DbCacheRepo(httpRepo, null, configService);

        DbStoringRepoMixin mixinSpy = spy(new DbStoringRepoMixin(httpRepoDescriptor, null, null));
        Mockito.doReturn(null).when(mixinSpy).createOrGetFolder(any());

        ReflectionTestUtils.setField(localCacheRepo, "mixin", mixinSpy);

        localCacheRepo.init(false);

        verify(fileService, times(1)).exists(any());
        verify(mixinSpy, times(0)).createOrGetFolder(any());
    }

    @Test
    public void shouldCallFileServiceExistsIfUnknownTest() {
        httpRepoDescriptor.setKey("remote-repo");
        DbCacheRepo localCacheRepo = new DbCacheRepo(httpRepo, null, configService);

        DbStoringRepoMixin mixinSpy = spy(new DbStoringRepoMixin(httpRepoDescriptor, null, null));
        Mockito.doReturn(null).when(mixinSpy).createOrGetFolder(any());

        ReflectionTestUtils.setField(localCacheRepo, "mixin", mixinSpy);
        localCacheRepo.init(false);

        verify(fileService, times(1)).exists(any());
        verify(mixinSpy, times(0)).createOrGetFolder(any());

        when(fileService.exists(any())).thenReturn(false);
        localCacheRepo.init(false);

        verify(fileService, times(2)).exists(any());
        verify(mixinSpy, times(1)).createOrGetFolder(any());
    }

    @Test
    public void testAcquireLock() throws InterruptedException {
        ConflictsGuard conflictGuard = mock(ConflictsGuard.class);
        when(httpRepo.getRemoteDownloaderLockingMap()).thenReturn(conflictGuard);
        doAnswer(invocationOnMock -> true).when(conflictGuard).tryToLock("path/to/lock", 2, TimeUnit.SECONDS);
        assertTrue(httpRepo.acquireLock("path/to/lock", 2));

        doThrow(new InterruptedException()).when(conflictGuard).tryToLock("path/to/lock", 2, TimeUnit.SECONDS);
        assertFalse(httpRepo.acquireLock("path/to/lock", 2));
    }

    @Test
    public void testDownloadAndSave() throws IOException, RepoRejectException {
        InternalRequestContext requestContext = mock(InternalRequestContext.class);
        RepoResource repoResource = mock(RepoResource.class);
        RepoPath repoPath = RepoPathFactory.create("repo", "path");
        RepoResourceInfo repoResourceInfo = new FileInfoImpl(repoPath);
        when(repoResource.getInfo()).thenReturn(repoResourceInfo);
        when(repoResource.getRepoPath()).thenReturn(repoPath);
        when(repoResource.isFound()).thenReturn(true);
        LocalCacheRepo localCacheRepo = mock(LocalCacheRepo.class);
        ReflectionTestUtils.setField(httpRepo, "localCacheRepo", localCacheRepo);
        ExpiredRepoResource expiredRepoResource = new ExpiredRepoResource(repoResource);
        when(localCacheRepo.getInfo(requestContext)).thenReturn(expiredRepoResource);
        ResourceStreamHandle streamHandle = mock(ResourceStreamHandle.class);
        when(localCacheRepo.getResourceStreamHandle(requestContext, expiredRepoResource)).thenReturn(streamHandle);
        Collection<RemoteRepoInterceptor> remoteRepoInterceptors = new ArrayList<>();
        ReflectionTestUtils.setField(httpRepo, "remoteRepoInterceptors", remoteRepoInterceptors);
        doReturn(true).when(httpRepo).shouldExpire(expiredRepoResource, repoResource);
        InternalRepositoryService repositoryService = mock(InternalRepositoryService.class);
        doReturn(repositoryService).when(httpRepo).getRepositoryService();
        doReturn(false).when(httpRepo).acquireLock("repo/path", 60);
        ResourceStreamHandle resourceStreamHandle = httpRepo.downloadAndSave(requestContext, repoResource);
        verify(repoResource, times(1)).isFound();
        assertEquals(resourceStreamHandle, streamHandle);
        verify(localCacheRepo, times(1)).getResourceStreamHandle(requestContext, expiredRepoResource);
    }

    @Test(expectedExceptionsMessageRegExp = ".*Not found", expectedExceptions = FileNotFoundException.class)
    public void testDownloadAndSaveWithException() throws IOException, RepoRejectException {
        InternalRequestContext requestContext = mock(InternalRequestContext.class);
        RepoResource repoResource = mock(RepoResource.class);
        RepoPath repoPath = RepoPathFactory.create("repo", "path");
        RepoResourceInfo repoResourceInfo = new FileInfoImpl(repoPath);
        when(repoResource.getInfo()).thenReturn(repoResourceInfo);
        when(repoResource.getRepoPath()).thenReturn(repoPath);
        when(repoResource.isFound()).thenReturn(true);
        LocalCacheRepo localCacheRepo = mock(LocalCacheRepo.class);
        ReflectionTestUtils.setField(httpRepo, "localCacheRepo", localCacheRepo);
        ExpiredRepoResource expiredRepoResource = new ExpiredRepoResource(repoResource);
        when(localCacheRepo.getInfo(requestContext)).thenReturn(expiredRepoResource);
        ResourceStreamHandle streamHandle = mock(ResourceStreamHandle.class);
        doThrow(new FileNotFoundException("Not found")).when(localCacheRepo)
                .getResourceStreamHandle(requestContext, expiredRepoResource);
        when(localCacheRepo.getResourceStreamHandle(requestContext, expiredRepoResource)).thenReturn(streamHandle);
        Collection<RemoteRepoInterceptor> remoteRepoInterceptors = new ArrayList<>();
        ReflectionTestUtils.setField(httpRepo, "remoteRepoInterceptors", remoteRepoInterceptors);
        doReturn(true).when(httpRepo).shouldExpire(expiredRepoResource, repoResource);
        InternalRepositoryService repositoryService = mock(InternalRepositoryService.class);
        doReturn(repositoryService).when(httpRepo).getRepositoryService();
        doReturn(false).when(httpRepo).acquireLock("repo/path", 60);
        ResourceStreamHandle resourceStreamHandle = httpRepo.downloadAndSave(requestContext, repoResource);
        verify(repoResource, times(1)).isFound();
        assertEquals(resourceStreamHandle, streamHandle);
        verify(localCacheRepo, times(1)).getResourceStreamHandle(requestContext, expiredRepoResource);
    }

    @Test
    public void testDownloadAndSaveNoFallback() throws IOException, RepoRejectException {
        InternalRequestContext requestContext = mock(InternalRequestContext.class);
        RepoResource repoResource = mock(RepoResource.class);
        RepoPath repoPath = RepoPathFactory.create("repo", "path");
        RepoResourceInfo repoResourceInfo = new FileInfoImpl(repoPath);
        when(repoResource.getInfo()).thenReturn(repoResourceInfo);
        when(repoResource.getRepoPath()).thenReturn(repoPath);
        when(repoResource.isFound()).thenReturn(false);
        LocalCacheRepo localCacheRepo = mock(LocalCacheRepo.class);
        ReflectionTestUtils.setField(httpRepo, "localCacheRepo", localCacheRepo);
        ExpiredRepoResource expiredRepoResource = new ExpiredRepoResource(repoResource);
        when(localCacheRepo.getInfo(requestContext)).thenReturn(expiredRepoResource);
        Collection<RemoteRepoInterceptor> remoteRepoInterceptors = new ArrayList<>();
        ReflectionTestUtils.setField(httpRepo, "remoteRepoInterceptors", remoteRepoInterceptors);
        doReturn(true).when(httpRepo).shouldExpire(expiredRepoResource, repoResource);
        InternalRepositoryService repositoryService = mock(InternalRepositoryService.class);
        doReturn(repositoryService).when(httpRepo).getRepositoryService();
        doReturn(false).when(httpRepo).acquireLock("repo/path", 60 * 15);
        ResourceStreamHandle resourceStreamHandle = httpRepo.downloadAndSave(requestContext, repoResource);
        verify(repoResource, times(1)).isFound();
        assertNull(resourceStreamHandle);
        verify(localCacheRepo, times(0)).getResourceStreamHandle(requestContext, expiredRepoResource);
    }

    @Test
    public void testGetHandleFromCacheNotFound() throws IOException, RepoRejectException {
        InternalRequestContext requestContext = mock(InternalRequestContext.class);
        RepoResource remoteResource = mock(FileResource.class);
        RepoResource cachedResource = mock(RepoResource.class);
        String message = "message";
        LocalCacheRepo localCacheRepo = mock(LocalCacheRepo.class);
        ReflectionTestUtils.setField(httpRepo, "localCacheRepo", localCacheRepo);
        assertNull(httpRepo.getHandleFromCache(requestContext, remoteResource, cachedResource, message));
    }

    @Test
    public void testGetHandleFromCacheWithChecksumFromResource() throws IOException, RepoRejectException {
        InternalRequestContext requestContext = mock(InternalRequestContext.class);
        RepoResource remoteResource = mock(RepoResource.class);
        RepoResourceInfo repoResourceInfo = mock(RepoResourceInfo.class);
        ChecksumsInfo remoteChecksumsInfo = new ChecksumsInfo();
        when(repoResourceInfo.getChecksumsInfo()).thenReturn(remoteChecksumsInfo);
        when(remoteResource.getInfo()).thenReturn(repoResourceInfo);
        ExpiredRepoResource expiredRepoResource = new ExpiredRepoResource(remoteResource);
        RepoResource cachedResource = mock(RepoResource.class);
        String message = "message";
        LocalCacheRepo localCacheRepo = mock(LocalCacheRepo.class);
        ReflectionTestUtils.setField(httpRepo, "localCacheRepo", localCacheRepo);
        ChecksumsInfo checksumsInfo = new ChecksumsInfo();
        ChecksumInfo sha1 = new ChecksumInfo(ChecksumType.sha1, "dda2437c5d609ff934bc83d4f5af72c764aaa680",
                "dda2437c5d609ff934bc83d4f5af72c764aaa680");
        checksumsInfo.setChecksums(Set.of(sha1));
        ResourceWithChecksumStreamHandle resourceStreamHandle = new TestResourceWithChecksum(checksumsInfo);
        when(localCacheRepo.getResourceStreamHandle(requestContext, cachedResource)).thenReturn(resourceStreamHandle);
        httpRepo.getHandleFromCache(requestContext, expiredRepoResource, cachedResource, message);
        assertEquals(remoteChecksumsInfo.getSha1(), sha1.getActual());
        verify(localCacheRepo, times(0)).getInfo(requestContext);
    }

    @Test
    public void testGetHandleFromCacheWithoutChecksumFromResource() throws IOException, RepoRejectException {
        InternalRequestContext requestContext = mock(InternalRequestContext.class);
        RepoResource remoteResource = mock(RepoResource.class);
        RepoResourceInfo repoResourceInfo = mock(RepoResourceInfo.class);
        ChecksumsInfo remoteChecksumsInfo = new ChecksumsInfo();
        when(repoResourceInfo.getChecksumsInfo()).thenReturn(remoteChecksumsInfo);
        ChecksumInfo sha1 = new ChecksumInfo(ChecksumType.sha1, "dda2437c5d609ff934bc83d4f5af72c764aaa680",
                "dda2437c5d609ff934bc83d4f5af72c764aaa680");
        remoteChecksumsInfo.setChecksums(Set.of(sha1));

        when(remoteResource.getInfo()).thenReturn(repoResourceInfo);
        ExpiredRepoResource expiredRepoResource = new ExpiredRepoResource(remoteResource);
        RepoResource cachedResource = mock(RepoResource.class);
        String message = "message";
        LocalCacheRepo localCacheRepo = mock(LocalCacheRepo.class);
        ReflectionTestUtils.setField(httpRepo, "localCacheRepo", localCacheRepo);

        ResourceStreamHandle resourceStreamHandle = mock(ResourceStreamHandle.class);
        when(localCacheRepo.getInfo(requestContext)).thenReturn(expiredRepoResource);
        when(localCacheRepo.getResourceStreamHandle(requestContext, cachedResource)).thenReturn(resourceStreamHandle);
        httpRepo.getHandleFromCache(requestContext, expiredRepoResource, cachedResource, message);
        assertEquals(remoteChecksumsInfo.getSha1(), sha1.getActual());
    }

    @Test
    public void testGetInfoPermissionsCheckNotCalledIfMavenAndNotFoundAndConstEnabled() {
        InternalRequestContext requestContext = mock(InternalRequestContext.class);
        RepoResource res = mock(RepoResource.class);
        Request request = mock(Request.class);
        HttpRepoDescriptor descriptor = mock(HttpRepoDescriptor.class);
        HttpConnectionPoolMonitorService connectionPool = new HttpConnectionPoolMonitorService();

        httpRepo.init(false);
        HttpRepo httpRepoSpy = spy(this.httpRepo);

        // perquisites
        when(requestContext.getResourcePath()).thenReturn("colors");
        when(requestContext.getRequest()).thenReturn(request);
        when(context.beanForType(HttpConnectionPoolMonitorService.class)).thenReturn(connectionPool);
        when(restCoreAddon.getDynamicVersionContext(httpRepoSpy, requestContext, true)).thenReturn(requestContext);
        when(httpRepoSpy.getDescriptor()).thenReturn(descriptor);
        doReturn(res).when(httpRepoSpy).internalGetInfo(any(), any());
        doReturn("repoKey").when(httpRepoSpy).getKey();
        doReturn(res).when(httpRepoSpy).getInfoFromCache(requestContext);

        // test essence
        when(descriptor.getType()).thenReturn(RepoType.Maven);
        when(res.isFound()).thenReturn(false);
        ArtifactorySystemProperties artifactoryProperties = ArtifactoryHome.get().getArtifactoryProperties();
        artifactoryProperties.setProperty(ConstantValues.enableMavenNonPreemptiveAuthentication.getPropertyName(),
                "true");

        httpRepoSpy.getInfo(requestContext);
        verify(httpRepoSpy, times(0)).checkDownloadIsAllowed(any());
    }

    @Test
    public void testGetInfoPermissionsCheckCalledIfMavenAndFoundAndConstEnabled() {
        InternalRequestContext requestContext = mock(InternalRequestContext.class);
        RepoResource res = mock(RepoResource.class);
        Request request = mock(Request.class);
        HttpRepoDescriptor descriptor = mock(HttpRepoDescriptor.class);
        RepoPath repoPath = mock(RepoPath.class);
        HttpConnectionPoolMonitorService connectionPool = new HttpConnectionPoolMonitorService();

        httpRepo.init(false);
        HttpRepo httpRepoSpy = spy(this.httpRepo);

        // perquisites
        when(requestContext.getResourcePath()).thenReturn("colors");
        when(requestContext.getRequest()).thenReturn(request);
        when(context.beanForType(HttpConnectionPoolMonitorService.class)).thenReturn(connectionPool);
        when(restCoreAddon.getDynamicVersionContext(httpRepoSpy, requestContext, true)).thenReturn(requestContext);
        when(httpRepoSpy.getDescriptor()).thenReturn(descriptor);
        when(res.getRepoPath()).thenReturn(repoPath);
        when(repoPath.getPath()).thenReturn("colors");
        doReturn(res).when(httpRepoSpy).internalGetInfo(any(), any());
        doReturn("repoKey").when(httpRepoSpy).getKey();
        doReturn(res).when(httpRepoSpy).getInfoFromCache(requestContext);

        // test essence
        when(descriptor.getType()).thenReturn(RepoType.Maven);
        when(res.isFound()).thenReturn(true);
        ArtifactorySystemProperties artifactoryProperties = ArtifactoryHome.get().getArtifactoryProperties();
        artifactoryProperties.setProperty(ConstantValues.enableMavenNonPreemptiveAuthentication.getPropertyName(),
                "true");

        httpRepoSpy.getInfo(requestContext);
        verify(httpRepoSpy, times(1)).checkDownloadIsAllowed(any());
    }

    @Test
    public void testGetInfoPermissionsCheckCalledIfNotMaven() {
        InternalRequestContext requestContext = mock(InternalRequestContext.class);
        RepoResource res = mock(RepoResource.class);
        Request request = mock(Request.class);
        HttpRepoDescriptor descriptor = mock(HttpRepoDescriptor.class);
        HttpConnectionPoolMonitorService connectionPool = new HttpConnectionPoolMonitorService();

        httpRepo.init(false);
        HttpRepo httpRepoSpy = spy(this.httpRepo);

        // perquisites
        when(requestContext.getResourcePath()).thenReturn("colors");
        when(requestContext.getRequest()).thenReturn(request);
        when(context.beanForType(HttpConnectionPoolMonitorService.class)).thenReturn(connectionPool);
        when(restCoreAddon.getDynamicVersionContext(httpRepoSpy, requestContext, true)).thenReturn(requestContext);
        when(httpRepoSpy.getDescriptor()).thenReturn(descriptor);
        doReturn(res).when(httpRepoSpy).internalGetInfo(any(), any());
        doReturn("repoKey").when(httpRepoSpy).getKey();
        doReturn(res).when(httpRepoSpy).getInfoFromCache(requestContext);

        // test essence
        when(descriptor.getType()).thenReturn(RepoType.Docker);

        httpRepoSpy.getInfo(requestContext);
        verify(httpRepoSpy, times(1)).checkDownloadIsAllowed(any());
    }

    @Test
    public void testGetInfoPermissionsCheckCalledIfMavenButConstDisabled() {
        InternalRequestContext requestContext = mock(InternalRequestContext.class);
        RepoResource res = mock(RepoResource.class);
        Request request = mock(Request.class);
        HttpRepoDescriptor descriptor = mock(HttpRepoDescriptor.class);
        RepoPath repoPath = mock(RepoPath.class);
        HttpConnectionPoolMonitorService connectionPool = new HttpConnectionPoolMonitorService();

        httpRepo.init(false);
        HttpRepo httpRepoSpy = spy(this.httpRepo);

        // perquisites
        when(requestContext.getResourcePath()).thenReturn("colors");
        when(requestContext.getRequest()).thenReturn(request);
        when(context.beanForType(HttpConnectionPoolMonitorService.class)).thenReturn(connectionPool);
        when(restCoreAddon.getDynamicVersionContext(httpRepoSpy, requestContext, true)).thenReturn(requestContext);
        when(httpRepoSpy.getDescriptor()).thenReturn(descriptor);
        when(res.getRepoPath()).thenReturn(repoPath);
        when(repoPath.getPath()).thenReturn("colors");
        doReturn(res).when(httpRepoSpy).internalGetInfo(any(), any());
        doReturn("repoKey").when(httpRepoSpy).getKey();
        doReturn(res).when(httpRepoSpy).getInfoFromCache(requestContext);

        // test essence
        when(descriptor.getType()).thenReturn(RepoType.Maven);
        ArtifactorySystemProperties artifactoryProperties = ArtifactoryHome.get().getArtifactoryProperties();
        artifactoryProperties.setProperty(ConstantValues.enableMavenNonPreemptiveAuthentication.getPropertyName(),
                "false");

        httpRepoSpy.getInfo(requestContext);
        verify(httpRepoSpy, times(1)).checkDownloadIsAllowed(any());
    }

    @DataProvider(name = "requestClasses")
    public static Object[][] requestClasses() {
        return new Object[][] {
                {Request.class, false},
                {InternalArtifactoryRequest.class, true},
                {TranslatedArtifactoryRequest.class, true},
        };
    }

    @Test(dataProvider = "requestClasses")
    public void testAllowedQueryParamRewrite(Class<Request> requestClass, boolean queryParamRewriteAllowed) {
        assertEquals(httpRepo
                .isAllowedQueryParamRewrite(mock(requestClass)), queryParamRewriteAllowed);
    }

    private HashMap<String, RemoteRepoInterceptor> prepInterceptorMocks(ExpiryAction expiryAction,
            ExpiryAction expiryAction2) {
        RemoteRepoInterceptor remoteRepoInterceptorMock = mock(RemoteRepoInterceptor.class);
        RemoteRepoInterceptor remoteRepoInterceptorMock2 = mock(RemoteRepoInterceptor.class);
        when(remoteRepoInterceptorMock.shouldExpireCachedResource(any(), any())).thenReturn(expiryAction);
        when(remoteRepoInterceptorMock2.shouldExpireCachedResource(any(), any())).thenReturn(expiryAction2);
        HashMap<String, RemoteRepoInterceptor> remoteInterceptors = new LinkedHashMap<>() {
            {
                put("1", remoteRepoInterceptorMock);
                put("2", remoteRepoInterceptorMock2);
            }
        };
        when(context.beansForType(RemoteRepoInterceptor.class))
                .thenReturn(remoteInterceptors);
        return remoteInterceptors;
    }

    /**
     * Check the different values that are read from the given checksum are as expected
     *
     * @param checksumType  Type of checksum test file to read
     * @param expectedValue Expected checksum value
     */
    private void invokeReadChecksum(String checksumType, String expectedValue) throws Exception {
        InputStream stream = getClass().getResourceAsStream("/org/artifactory/repo/test-" + checksumType + ".md5");
        try {
            String checksum = ChecksumUtils.checksumStringFromStream(stream);
            Assert.assertEquals(expectedValue, checksum, "Incorrect checksum value returned.");
        } finally {
            IOUtils.closeQuietly(stream);
        }
    }

    private void createCacheRepo() {
        ReflectionTestUtils.setField(remoteRepoConfig, "key", "remote-repo");
        DbCacheRepo localCacheRepo = new DbCacheRepo(httpRepo, null, configService);
        ReflectionTestUtils.setField(httpRepo, "localCacheRepo", localCacheRepo);

        DbStoringRepoMixin mixin = mock(DbStoringRepoMixin.class);
        ReflectionTestUtils.setField(localCacheRepo, "mixin", mixin);
        when(mixin.itemExists(any())).thenReturn(true);
        when(mixin.getKey()).thenReturn("remote-repo-cache");
    }

    private void expectCanRead(boolean canRead) {
        when(authService.canRead(repoPath)).thenReturn(canRead);
    }

    private void expectXrayBlocked() {
        when(xrayAddon.isDownloadBlocked(any(RepoPath.class))).thenReturn(false);
    }

    /**
     * Try to download the the given path
     *
     * @param path Repopath to test the download with
     * @return StatusHolder - Contains any errors which might be monitored
     */
    private StatusHolder tryAllowsDownload(RepoPath path) {
        return httpRepo.checkDownloadIsAllowed(path);
    }

    @DataProvider
    public static Object[][] testInternalGetInfoBlockedRemoteDownloadProvider() {
        return new Object[][] {
                {false}, {true}
        };
    }

    private static class TestResourceWithChecksum implements ResourceWithChecksumStreamHandle {
        private ChecksumsInfo checksumsInfo;

        public TestResourceWithChecksum(ChecksumsInfo checksumsInfo) {
            this.checksumsInfo = checksumsInfo;
        }

        @Override
        public Optional<ChecksumsInfo> getChecksumsInfo() {
            return Optional.ofNullable(checksumsInfo);
        }

        @Override
        public InputStream getInputStream() {
            return null;
        }

        @Override
        public long getSize() {
            return 0;
        }

        @Override
        public void close() {

        }
    }
}
