package org.artifactory.repo.interceptor;

import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.npm.NpmAddon;
import org.artifactory.api.properties.PropertiesService;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.fs.FileInfo;
import org.artifactory.md.Properties;
import org.artifactory.model.xstream.fs.FileInfoImpl;
import org.artifactory.model.xstream.fs.PropertiesImpl;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.sapi.fs.VfsItem;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentMatchers;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;

import javax.annotation.Nonnull;
import java.util.Set;

class NpmMetadataInterceptorTest {

    private NpmMetadataInterceptor metadataInterceptor;
    @Mock
    private PropertiesService propertiesService;
    @Mock
    private RepositoryService repositoryService;
    @Mock
    private AddonsManager addonsManager;
    @Mock
    private NpmAddon npmAddon;
    private AutoCloseable autoCloseable;

    @BeforeEach
    void setUp() {
        autoCloseable = MockitoAnnotations.openMocks(this);
        metadataInterceptor = Mockito.spy(new NpmMetadataInterceptor());
        metadataInterceptor.setPropertiesService(propertiesService);
        metadataInterceptor.setRepositoryService(repositoryService);
        metadataInterceptor.setAddonsManager(addonsManager);
        Mockito.when(addonsManager.addonByType(NpmAddon.class)).thenReturn(npmAddon);
        LocalRepoDescriptor localRepoDescriptor = Mockito.mock(LocalRepoDescriptor.class);
        Mockito.when(localRepoDescriptor.getType()).thenReturn(RepoType.Npm);
        Mockito.when(repositoryService.localOrFederatedRepoDescriptorByKey(ArgumentMatchers.anyString()))
                .thenReturn(localRepoDescriptor);
    }

    @AfterEach
    void tearDown() throws Exception {
        autoCloseable.close();
    }

    @Test
    void testAfterCopyWithLatestTag() {
        RepoPath repoPath = RepoPathFactory.create("npm-local/package-2.32.0.tgz");
        FileInfo fileInfo = new FileInfoImpl();
        VfsItem vfsItem = mockVfsItemOfNpmPkg(repoPath, fileInfo, true);

        metadataInterceptor.afterCopy(vfsItem, vfsItem, null, null, null);
        Mockito.verify(propertiesService).removePropertyValues(repoPath, "npm.disttag", Set.of("latest"), true);
        Mockito.verifyNoMoreInteractions(propertiesService);
        Mockito.verify(npmAddon).handleAddAfterCommit(fileInfo);
        Mockito.verifyNoMoreInteractions(npmAddon);
    }

    @Test
    void testAfterMoveWithLatestTag() {
        RepoPath repoPath = RepoPathFactory.create("npm-local/package-2.32.0.tgz");
        FileInfo fileInfo = new FileInfoImpl();
        VfsItem vfsItem = mockVfsItemOfNpmPkg(repoPath, fileInfo, true);

        metadataInterceptor.afterMove(vfsItem, vfsItem, null, null, null);
        Mockito.verify(npmAddon).removeNpmPackage(fileInfo, Set.of());
        Mockito.verify(propertiesService).removePropertyValues(repoPath, "npm.disttag", Set.of("latest"), true);
        Mockito.verifyNoMoreInteractions(propertiesService);
        Mockito.verify(npmAddon).handleAddAfterCommit(fileInfo);
        Mockito.verifyNoMoreInteractions(npmAddon);
    }

    @Test
    void testAfterCopyWhenNoLatestTag() {
        RepoPath repoPath = RepoPathFactory.create("npm-local/package-2.32.0.tgz");
        FileInfo fileInfo = new FileInfoImpl();
        VfsItem vfsItem = mockVfsItemOfNpmPkg(repoPath, fileInfo, false);

        metadataInterceptor.afterCopy(vfsItem, vfsItem, null, null, null);
        Mockito.verifyNoMoreInteractions(propertiesService);
        Mockito.verify(npmAddon).handleAddAfterCommit(fileInfo);
        Mockito.verifyNoMoreInteractions(npmAddon);
    }

    @Nonnull
    private static VfsItem mockVfsItemOfNpmPkg(RepoPath repoPath, FileInfo fileInfo, boolean includeLatestTag) {
        VfsItem vfsItem = Mockito.mock(VfsItem.class);
        Properties properties = new PropertiesImpl();
        Multimap<String, String> npmProps = HashMultimap.create();
        npmProps.put("npm.name", "pkg");
        npmProps.put("npm.version", "1.0.0");
        if (includeLatestTag) {
            npmProps.put("npm.disttag", "latest");
        }
        npmProps.put("npm.disttag", "another_tag");
        properties.putAll(npmProps);
        Mockito.when(vfsItem.getProperties()).thenReturn(properties);
        Mockito.when(vfsItem.getRepoPath()).thenReturn(repoPath);
        Mockito.when(vfsItem.isFile()).thenReturn(true);
        Mockito.when(vfsItem.getInfo()).thenReturn(fileInfo);
        return vfsItem;
    }
}