package org.artifactory.repo.interceptor.config;

import com.jfrog.sysconf.SysConfig;
import com.jfrog.sysconf.SysConfigHelper;
import org.artifactory.api.config.RepositoryConfigService;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.model.RemoteRepoConfig;
import org.artifactory.repo.RepoConfigChange;
import org.artifactory.repo.RepoDetailsType;
import org.jfrog.security.crypto.EncryptionWrapper;
import org.jfrog.security.crypto.result.DecryptionStringResult;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;

import java.io.File;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotEquals;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

/**
 * @author Tamar Rosen
 */
class EncryptRepoConfigInterceptorTest {

    @BeforeAll
    static void setUp(){
        prepareArtifactoryHome();
    }

    @Test
    void testOnBeforeSave() {
        // given
        RemoteRepoConfig newRemoteRepo = RemoteRepoConfig.create("repoKey", RepoType.Generic);
        newRemoteRepo.getRepoTypeConfig().setPassword("old-password");

        // when
        EncryptRepoConfigInterceptor interceptor = new EncryptRepoConfigInterceptor();
        interceptor.onBeforeSave(Map.of(RepoDetailsType.REMOTE, List.of(new RepoConfigChange("remote-repo",
                RepoConfigChange.ConfigAction.REPO_CREATE, newRemoteRepo, null))));

        // then
        assertNotEquals("old-password", newRemoteRepo.getRepoTypeConfig().getPassword());
    }

    @Test
    void testDecrypt() {
        // given
        RemoteRepoConfig remoteRepoConfig = RemoteRepoConfig.create("repoKey", RepoType.Generic);
        String encryptedPassword = "JE2f7DvoFNZ77qa4udE4LxVi8zsZ3q9tEzk5kaYJLgoLJYwdPm";
        remoteRepoConfig.getRepoTypeConfig().setPassword(encryptedPassword);
        RepositoryConfigService repositoryConfigService = mock(RepositoryConfigService.class);
        when(repositoryConfigService.findRepositoriesByType(RepoDetailsType.REMOTE)).thenReturn(
                List.of(remoteRepoConfig));

        // when
        EncryptRepoConfigInterceptor.decrypt(repositoryConfigService);

        // then
        ArgumentCaptor<List> captor = ArgumentCaptor.forClass(List.class);
        verify(repositoryConfigService).bulkUpsertRepositoryConfigs(captor.capture(),eq(true));
        assertEquals(1, captor.getValue().size());
        assertNotEquals("old-password",
                ((RemoteRepoConfig) captor.getValue().get(0)).getRepoTypeConfig().getPassword());
    }

    private static void prepareArtifactoryHome() {
        ArtifactoryHome home = mock(ArtifactoryHome.class);
        ArtifactoryHome.bind(home);
        File key = mock(File.class);
        when(home.getArtifactoryKey()).thenReturn(key);
        when(key.exists()).thenReturn(true);
        EncryptionWrapper encryptionWrapper = mock(EncryptionWrapper.class);
        when(encryptionWrapper.encryptIfNeeded(anyString())).thenReturn("encrypted-password");
        DecryptionStringResult decryptionStringResult = mock(DecryptionStringResult.class);
        when(decryptionStringResult.getDecryptedData()).thenReturn("decrypted-password");
        when(encryptionWrapper.decryptIfNeeded(anyString())).thenReturn(decryptionStringResult);
        when(home.getMasterEncryptionWrapper()).thenReturn(encryptionWrapper);

        SysConfig sysConfig = mock(SysConfig.class);
        SysConfigHelper sysConfigHelper = mock(SysConfigHelper.class);
        when(sysConfigHelper.getMasterKey()).thenReturn(Optional.of(""));
        when(sysConfig.helper()).thenReturn(sysConfigHelper);
        when(home.getSysConfig()).thenReturn(sysConfig);
    }
}