package org.artifactory.schedule;

import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.quartz.JobDataMap;
import org.quartz.JobDetail;
import org.quartz.JobKey;
import org.quartz.Scheduler;
import org.quartz.SimpleScheduleBuilder;
import org.quartz.Trigger;
import org.quartz.TriggerKey;

import java.util.Date;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertNotEquals;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.Mockito.anyString;
import static org.mockito.Mockito.doNothing;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

/**
 * @author Alexei Vainshtein
 */
class QuartzTaskTest {

    private QuartzTask quartzTask;
    private TriggerKey key = new TriggerKey("test", "artifactory");

    @BeforeEach
    void prepareQuartzTask() {
        Trigger trigger = mock(Trigger.class);
        when(trigger.getKey()).thenReturn(key);
        JobDetail jobDetail = mock(JobDetail.class);
        when(jobDetail.getKey()).thenReturn(new JobKey("test"));
        quartzTask = spy(new QuartzTask(HaQuartzTaskTest.Job1.class, trigger, jobDetail));
    }

    @Test
    void testGetScheduler() {
        JobDetail jobDetail = mock(JobDetail.class);
        JobDataMap jobDataMap = mock(JobDataMap.class);
        ArtifactoryContext context = mock(ArtifactoryContext.class);
        ArtifactoryContextThreadBinder.bind(context);
        Scheduler schedulerMock = mock(Scheduler.class);
        QuartzTask mockTask = new QuartzTask(HaQuartzTaskTest.Job1.class, null, jobDetail);
        when(jobDetail.getJobDataMap()).thenReturn(jobDataMap);
        when(jobDataMap.get(anyString())).thenReturn(JobName.HEARTBEAT).thenReturn(JobName.IMPORT);
        when(context.beanForType("hbScheduler", Scheduler.class)).thenReturn(schedulerMock);
        when(context.beanForType(Scheduler.class)).thenReturn(mock(Scheduler.class));
        Scheduler first = mockTask.getScheduler();
        Scheduler second = mockTask.getScheduler();
        assertEquals(first, schedulerMock);
        assertNotEquals(second, schedulerMock);
        ArtifactoryContextThreadBinder.unbind();
    }

    @Test
    void testShouldRescheduleTask() {
        quartzTask.setShouldRescheduleTask(true);
        assertTrue(quartzTask.shouldRescheduleTask());
        quartzTask.setShouldRescheduleTask(false);
        assertFalse(quartzTask.shouldRescheduleTask());
    }

    @Test
    void testRescheduleTask() {
        Date currentTime = new Date(System.currentTimeMillis());
        doNothing().when(quartzTask).cancelTask();
        doNothing().when(quartzTask).scheduleTask();
        quartzTask.setShouldRescheduleTask(true);
        quartzTask.rescheduleTask(1000);
        verify(quartzTask, times(1)).cancelTask();
        verify(quartzTask, times(1)).scheduleTask();
        Trigger trigger = quartzTask.getTrigger();
        assertEquals(trigger.getKey(), key);
        assertTrue(trigger.getScheduleBuilder() instanceof SimpleScheduleBuilder);
        Date startTime = trigger.getStartTime();
        assertTrue(startTime.getTime() >= currentTime.getTime() + 1000);
        assertFalse(quartzTask.shouldRescheduleTask());
    }
}