package org.artifactory.util;


import org.apache.commons.lang.RandomStringUtils;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.Test;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.util.Arrays;

/**
 * Brut force test for the lazy stream loader
 */
class LazyStreamLoaderTest {

    @Test
    void lazyInitialization() throws IOException {
        // Ensure stream is not initialized until call to getStream()
        LazyStreamLoader.StreamProvider test = () -> new StringInputStream("Test");
        LazyStreamLoader lazyStreamLoader = new LazyStreamLoader(test);
        // Ensure stream is initialized after first usage
        int read = lazyStreamLoader.read();
        Assertions.assertEquals('T', read);
    }


    @Test
    void singleBiteRead() throws IOException {
        for (int i = 0; i < 10; i++) {
            for (int j = 1; j < i; j++) {
                byte[] bytes = RandomStringUtils.randomAscii(i).getBytes();
                LazyStreamLoader lazyStreamLoader = new LazyStreamLoader(() -> new ByteArrayInputStream(bytes));
                int read = 0;
                while (read < i) {
                    int data = lazyStreamLoader.read();
                    Assertions.assertEquals(bytes[read], data);
                    read += 1;
                }
            }
        }
    }

    @Test
    void arrayRead() throws IOException {
        for (int i = 0; i < 0; i++) {
            for (int j = 1; j < i; j++) {
                byte[] bytes = RandomStringUtils.randomAscii(i).getBytes();
                LazyStreamLoader lazyStreamLoader = new LazyStreamLoader(() -> new ByteArrayInputStream(bytes));
                byte[] buffer = new byte[j];
                int read = 0;
                while (read < i) {
                    int amount = lazyStreamLoader.read(buffer);
                    Assertions.assertEquals(0, Arrays.compare(bytes, read, read + amount, buffer, 0, amount));
                    read += amount;
                }
            }
        }
    }

    @Test
    void arrayReadWithOffset() throws IOException {
        for (int i = 0; i < 10; i++) {
            for (int j = 1; j < i; j++) {
                byte[] bytes = RandomStringUtils.randomAscii(i).getBytes();
                LazyStreamLoader lazyStreamLoader = new LazyStreamLoader(() -> new ByteArrayInputStream(bytes));
                byte[] buffer = new byte[j + 2];
                int read = 0;
                while (read < i) {
                    int amount = lazyStreamLoader.read(buffer, 2, buffer.length - 2);
                    Assertions.assertEquals(0, Arrays.compare(bytes, read, read + amount, buffer, 2, 2 + amount));
                    read += amount;
                }
            }
        }
    }

    @Test
    void arrayReadAll() throws IOException {
        for (int i = 0; i < 10; i++) {
            byte[] bytes = RandomStringUtils.randomAscii(i).getBytes();
            LazyStreamLoader lazyStreamLoader = new LazyStreamLoader(() -> new ByteArrayInputStream(bytes));
            byte[] result = lazyStreamLoader.readAllBytes();
            Assertions.assertEquals(0, Arrays.compare(bytes, result));
        }
    }

    @Test
    void arrayReadNBites() throws IOException {
        for (int i = 0; i < 10; i++) {
            byte[] bytes = RandomStringUtils.randomAscii(i).getBytes();
            for (int j = 0; j < i; j++) {
                LazyStreamLoader lazyStreamLoader = new LazyStreamLoader(() -> new ByteArrayInputStream(bytes));
                byte[] result = lazyStreamLoader.readNBytes(j);
                Assertions.assertEquals(0, Arrays.compare(bytes, 0, j, result, 0, j));
            }
        }
    }

    @Test
    void arrayReadNBites2() throws IOException {
        for (int i = 0; i < 10; i++) {
            byte[] bytes = RandomStringUtils.randomAscii(i).getBytes();
            for (int j = 0; j < i; j++) {
                LazyStreamLoader lazyStreamLoader = new LazyStreamLoader(() -> new ByteArrayInputStream(bytes));
                byte[] result = new byte[j];
                lazyStreamLoader.readNBytes(result, 0, j);
                Assertions.assertEquals(0, Arrays.compare(bytes, 0, j, result, 0, j));
            }
        }
    }

    @Test
    void skip() throws IOException {
        for (int i = 0; i < 10; i++) {
            byte[] bytes = RandomStringUtils.randomAscii(i).getBytes();
            for (int j = 0; j < i; j++) {
                LazyStreamLoader lazyStreamLoader = new LazyStreamLoader(() -> new ByteArrayInputStream(bytes));
                long skiped = lazyStreamLoader.skip(j);
                byte byteValue = (byte) lazyStreamLoader.read();
                Assertions.assertEquals(bytes[(int) skiped], byteValue);
            }
        }
    }
}