package org.artifactory.api.config;

import org.artifactory.async.Async;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.model.BasicRepoConfig;
import org.artifactory.model.FederatedRepoConfig;
import org.artifactory.model.LocalConfig;
import org.artifactory.model.RealRepoConfig;
import org.artifactory.model.RemoteRepoConfig;
import org.artifactory.model.RepoConfig;
import org.artifactory.model.RepoConfigPropagationModel;
import org.artifactory.model.VirtualRepoConfig;
import org.artifactory.repo.ExtendedRepositoryConfig;
import org.artifactory.repo.RepoConfigReload;
import org.artifactory.repo.RepoDetailsType;
import org.artifactory.sapi.common.ExportSettings;
import org.artifactory.sapi.common.ImportSettings;
import org.jfrog.common.config.diff.DataDiff;

import java.io.File;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.function.Function;

/**
 * @author Tamar Rosen
 */
public interface RepositoryConfigService {

    void insertRepository(RepoConfig repositoryConfig);

    void upsertRepository(RepoConfig repositoryConfig);

    void updateRepository(RepoConfig repositoryConfig);

    /**
     * Removes the repository config with the specified key. Will also remove any references to this
     * repository from config descriptor entities (backup, indexer, replication and reverse proxy) and virtual repos.
     *
     * @param repoKey The key of the repository to remove.
     * @return The removed repository config or null if not found.
     */
    RepoConfig deleteRepository(String repoKey);

    void bulkInsertRepositoryConfigs(List<? extends RepoConfig> repositoryConfigs,
            boolean reload, boolean propagateChanges, boolean doBackups);

    void bulkUpsertRepositoryConfigs(Collection<? extends RepoConfig> repositoryConfigs, boolean reload);

    void callOnRepoConfigReload(Map<RepoDetailsType, List<RepoConfigReload>> typeToChanges);

    @Async
    void callOnRepoConfigReloadAsync(Map<RepoDetailsType, List<RepoConfigReload>> typeToChanges);

    void bulkDeleteRepositoryConfigs(List<String> repositoryConfigKeys, boolean reload, boolean propagateChanges);

    RepoConfig getRepoConfigByKey(String repoKey);

    RealRepoConfig getRealRepoConfigByKey(String repoKey);

    VirtualRepoConfig getVirtualRepoConfigByKey(String repoKey);

    RemoteRepoConfig getRemoteRepoConfigByKey(String repoKey);

    boolean exists(String repoKey);

    boolean existsRepositoryIgnoreCase(String repoKey);

    boolean existsByType(RepoDetailsType repoType, String repoKey);

    Collection<RepoConfig> getAllRepositoryConfigs();

    Collection<String> getAllRepositoryKeys();

    Set<RepoConfig> findRepositoryConfigsByKeys(Set<String> repoKeys);

    LinkedHashSet<RepoConfig> findRepositoryConfigsByKeysInSpecificOrder(VirtualRepoConfig virtualRepoConfig);

    <T> ExtendedRepositoryConfig<T> getExtendedRepositoryByKey(String repoKey, Class<T> type);

    Map<String, String> getPackageConfigByRepoKey(String repoKey);

    /**
     * @param repoKey key of the repository to return
     * @return Local or Federated repository config by repository key
     */
    LocalConfig getLocalOrFederatedRepoConfigByKey(String repoKey);

    FederatedRepoConfig getFederatedRepoConfigByKey(String repoKey);

    List<RepoConfig> findRepositoriesByType(RepoDetailsType repoType);

    List<RepoConfig> findRepositoriesByPackageType(RepoType packageType);

    List<RepoConfig> findRepositoriesByTypeAndPackageType(Set<RepoDetailsType> type, RepoType packageType);

    Set<String> getRepoKeysByTypeAndPackageType(Set<RepoDetailsType> type, RepoType packageType);

    <T> T getPackageConfigByRepoKey(String repoKey, Class<T> type);

    void insertPackageConfigByKey(String repoKey, Map<String, String> packageConfig);

    int countAll();

    int deleteAll(boolean reload, boolean propagateChanges);

    List<VirtualRepoConfig> getVirtualReposContainingRepo(RepoConfig repoConfig, int depth);

    List<VirtualRepoConfig> getVirtualReposContainingRepo(RepoConfig repoConfig);

    List<RepoConfig> getRepoKeysByProjectAndType(Function<String, Boolean> projectFilter,
            Set<RepoDetailsType> types);

    void reloadRepoConfigurations();

    @Async
    void calculateCacheDiffAndReloadAsync(Map<String, BasicRepoConfig> oldBasicRepoConfigs,
            Map<String, BasicRepoConfig> newBasicRepoConfigs);

    void reloadRepoConfigurations(RepoConfigPropagationModel model);

    void importFrom(ImportSettings settings);

    void exportTo(ExportSettings settings);

    void writeRepositoriesConfigToFile(File baseDir, Collection<RepoConfig> allRepositoryConfigs);

    void repoConfigSaveNewData(Set<DataDiff<?>> newDataDiffs);

    void importRepoConfigs(List<? extends RepoConfig> repoConfigsToImport, boolean reload, boolean propagateChanges);

    RepoType getRepoType(String repoKey);

    Set<String> findAllMissingRepoRootPaths();
}
