package org.artifactory.api.request;

import com.google.common.cache.Cache;
import com.google.common.cache.CacheBuilder;
import lombok.val;
import org.artifactory.common.ConstantValues;
import org.artifactory.request.ArtifactoryRequest;
import org.slf4j.Logger;
import org.slf4j.MDC;

import java.io.IOException;

public interface GracefulShutdownAware {

    String UBER_TRACE_ID = "uber-trace-id";

    default boolean canBeShutdown() {
        return true;
    }

    static Cache<String, Object> createTrackingCache() {
        return CacheBuilder.newBuilder()
                .maximumSize(ConstantValues.gracefulShutdownMaxTrackedRequestCount.getLong())
                .build();
    }

    static void track(Process process, ArtifactoryRequest request, ArtifactoryResponse response,
            Cache<String, Object> tracker, Logger log)
            throws IOException {
        val traceId = MDC.get(UBER_TRACE_ID);
        if (traceId != null) {
            tracker.put(traceId, traceId);
        }
        log.debug("Process {} tracking started", traceId);
        try {
            process.execute(request, response);
        } finally {
            if (traceId != null) {
                tracker.invalidate(traceId);
            }
            log.debug("Process {} tracking stopped", traceId);
        }
    }

    static boolean noActiveTracking(Cache<String, Object> tracker) {
        val trackingCount = tracker.size();
        return trackingCount == 0;
    }

    @FunctionalInterface
    interface Process {
        void execute(ArtifactoryRequest request, ArtifactoryResponse response) throws IOException;
    }

}
