package org.artifactory.util;

import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.descriptor.config.CentralConfigDescriptorImpl;
import org.artifactory.repo.federated.FederatedMember;
import org.artifactory.test.ArtifactoryHomeBoundJUnitTest;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.testng.annotations.AfterMethod;

import java.util.stream.Stream;

import static org.artifactory.repo.federated.FederatedMember.BASE_URL_PLACEHOLDER;
import static org.artifactory.util.FederatedMemberUtils.isLocal;
import static org.artifactory.util.FederatedMemberUtils.resolveUrl;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

/**
 * @author Tamar Rosen
 */
class FederatedMemberUtilsTest extends ArtifactoryHomeBoundJUnitTest {

    @Mock
    private ArtifactoryContext context;
    private AutoCloseable autoCloseable;

    @BeforeEach
    void setup() {
        autoCloseable = MockitoAnnotations.openMocks(this);
        ArtifactoryContextThreadBinder.bind(context);
    }

    @AfterMethod
    private void cleanup() throws Exception {
        ArtifactoryContextThreadBinder.unbind();
        autoCloseable.close();
    }

    @ParameterizedTest
    @MethodSource("resolveUrlProvider")
    void testResolveUrl(String memberUrl, String baseUrl, String fedBaseUrl, String expectedResolvedUrl) {
        // given
        when(context.isReady()).thenReturn(true);
        CentralConfigService configService = mock(CentralConfigService.class);
        CentralConfigDescriptorImpl configDescriptor = new CentralConfigDescriptorImpl();
        configDescriptor.setUrlBase(baseUrl);
        configDescriptor.setFederatedRepoUrlBase(fedBaseUrl);
        when(configService.getDescriptor()).thenReturn(configDescriptor);
        when(context.getCentralConfig()).thenReturn(configService);
        FederatedMember member = new FederatedMember();
        member.setUrl(memberUrl);
        // when
        String resolvedUrl = resolveUrl(member);
        // then
        assertEquals(expectedResolvedUrl, resolvedUrl);
    }

    @ParameterizedTest
    @MethodSource("isLocalUrlProvider")
    void testIsLocal(String memberUrl, String key, String baseUrl, boolean expected) {
        FederatedMember member = new FederatedMember();
        member.setUrl(memberUrl);
        assertEquals(expected, isLocal(member, key, baseUrl));
    }

    public static Stream<Arguments> resolveUrlProvider() {
        String BASE_URL = "https://baseurl.com";
        String FED_BASE_URL = "https://fed-baseurl.com";
        return Stream.of(
                Arguments.of(BASE_URL_PLACEHOLDER + "/repo", null, FED_BASE_URL, FED_BASE_URL + "/artifactory/repo"),
                Arguments.of(BASE_URL_PLACEHOLDER + "/repo", BASE_URL, null, BASE_URL + "/artifactory/repo"),
                Arguments.of(BASE_URL_PLACEHOLDER + "/repo", BASE_URL, FED_BASE_URL,
                        FED_BASE_URL + "/artifactory/repo"),
                Arguments.of(BASE_URL_PLACEHOLDER + "/repo", BASE_URL, FED_BASE_URL,
                        FED_BASE_URL + "/artifactory/repo"),
                Arguments.of(BASE_URL_PLACEHOLDER + "/repo", BASE_URL, FED_BASE_URL + "/artifactory",
                        FED_BASE_URL + "/artifactory/repo"),
                Arguments.of(BASE_URL + "/repo", null, null, BASE_URL + "/repo")
        );
    }

    public static Stream<Arguments> isLocalUrlProvider() {
        String BASE_URL = "https://baseurl.com";
        return Stream.of(
                Arguments.of(BASE_URL_PLACEHOLDER + "/repo", "repo", BASE_URL, true),
                Arguments.of(BASE_URL + "/repo", "repo", BASE_URL, true),
                Arguments.of("https://remote-url.com/repo", "repo", BASE_URL, false),
                Arguments.of(BASE_URL_PLACEHOLDER + "/repo", "other-repo", BASE_URL, false)
        );
    }
}