package org.artifactory.model;

import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonProperty;
import lombok.AccessLevel;
import lombok.Getter;
import lombok.Setter;
import org.artifactory.common.config.metadata.MetadataInfo;
import org.artifactory.common.config.metadata.MetadataVersion;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.repo.RepoDetailsType;
import org.artifactory.util.FederatedDefaultPackageConfigSetter;
import org.artifactory.util.ParamValidationUtils;
import org.jfrog.common.config.validation.NotUnicode;

import javax.validation.Valid;
import java.util.HashMap;
import java.util.Map;
import java.util.Objects;

/**
 * Federated repository configurations.
 * The federated repository configurations are similar to a local repository, but it extends the local repository
 * config {@link LocalRepositoryRepoTypeConfig} with additional federation related configurations, see
 * {@link FederatedRepositoryRepoTypeConfig}
 *
 * @author shayb
 */
@Valid
@Getter
@MetadataVersion(metadataVersion = 1)
public class FederatedRepoConfig implements LocalConfig, ProxyRepoConfig {
    @NotUnicode
    @MetadataInfo
    private final String key;
    @Valid
    @Setter(AccessLevel.PRIVATE)
    private BaseConfig baseConfig;
    @Valid
    @Setter(AccessLevel.PRIVATE)
    private FederatedRepositoryRepoTypeConfig repoTypeConfig = new FederatedRepositoryRepoTypeConfig();
    @MetadataInfo(defaultValue = "generic")
    private final RepoType packageType;
    /**
     * This is skipped since it's only mapping of the configuration.
     * Those are only Strings represents actual fields.
     * The info itself fetched from the package configuration itself
     *
     */
    @Setter(AccessLevel.PUBLIC)
    @MetadataInfo(skip = true)
    private Map<String, String> packageTypeConfig = new HashMap<>();
    @Valid
    @Setter(AccessLevel.PRIVATE)
    private SecurityConfig securityConfig = new SecurityConfig();
    @JsonIgnore
    @Setter
    @MetadataInfo(skip = true)
    private long modification;

    @JsonCreator
    protected FederatedRepoConfig(@JsonProperty("key") String repoKey,
            @JsonProperty("packageType") RepoType packageType) {
        this.key = ParamValidationUtils.requireNonBlank(repoKey, "Repository key cannot be blank");
        this.packageType = Objects.requireNonNull(packageType, "Repository package type cannot be null");
        this.baseConfig = new BaseConfig();
        FederatedDefaultPackageConfigSetter.setDefaultValues(packageType, packageTypeConfig);
    }

    /**
     * Factory method enabling exclusive access to instance creation
     *
     * @param repoKey     the repository key
     * @param packageType the repository package type
     * @return Federated repo configuration
     * @throws IllegalArgumentException if {@code repoKey} is blank
     * @throws NullPointerException     if {@code packageType} is null
     */
    public static FederatedRepoConfig create(String repoKey, RepoType packageType) {
        return new FederatedRepoConfig(repoKey, packageType);
    }

    /**
     * Factory method enabling exclusive access to deep copy a federated repo config
     *
     * @param federatedRepoConfig the federated repo config to be copied
     * @return a copy of the federated repo config
     * @throws NullPointerException if {@code federatedRepoConfig} is null
     */
    static FederatedRepoConfig copy(FederatedRepoConfig federatedRepoConfig) {
        Objects.requireNonNull(federatedRepoConfig, "Federated repo config cannot be null");
        FederatedRepoConfig repoCopy =
                new FederatedRepoConfig(federatedRepoConfig.getKey(), federatedRepoConfig.getPackageType());
        repoCopy.setBaseConfig(new BaseConfig(federatedRepoConfig.getBaseConfig()));
        repoCopy.setRepoTypeConfig(new FederatedRepositoryRepoTypeConfig(federatedRepoConfig.getRepoTypeConfig()));
        repoCopy.setPackageTypeConfig(new HashMap<>(federatedRepoConfig.getPackageTypeConfig()));
        repoCopy.setSecurityConfig(new SecurityConfig(federatedRepoConfig.getSecurityConfig()));
        repoCopy.setModification(federatedRepoConfig.getModification());
        return repoCopy;
    }

    @Override
    public boolean isCache() {
        return false;
    }

    @Override
    public void setProxyRef(String proxyKey) {
        getRepoTypeConfig().setProxyRef(proxyKey);
    }

    @Override
    public String getProxyRef() {
        return getRepoTypeConfig().getProxyRef();
    }

    @Override
    public void setDisableProxy(boolean disable) {
        getRepoTypeConfig().setDisableProxy(disable);
    }

    @Override
    public boolean isProxyDisabled() {
        return getRepoTypeConfig().isDisableProxy();
    }

    @Override
    public RepoDetailsType getRepoType() {
        return RepoDetailsType.FEDERATED;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }
        if (!(o instanceof FederatedRepoConfig)) {
            return false;
        }
        FederatedRepoConfig that = (FederatedRepoConfig) o;
        if (!Objects.equals(this.getRepoTypeConfig().getFederatedMembers(),
                that.getRepoTypeConfig().getFederatedMembers())) {
            return false;
        }
        if (!Objects.equals(this.getRepoTypeConfig().getProxyRef(), that.getRepoTypeConfig().getProxyRef())) {
            return false;
        }
        return this.getRepoTypeConfig().getModificationDate() == that.getRepoTypeConfig().getModificationDate() &&
                getKey().equals(that.getKey()) &&
                this.getPackageType() == that.getPackageType();
    }

    @Override
    public int hashCode() {
        return Objects.hash(super.hashCode(), this.getRepoTypeConfig().getFederatedMembers());
    }
    @Override
    public String toString() {
        return key;
    }
}
