package org.artifactory.schedule;

import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.ha.HaCommonAddon;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.storage.spring.ArtifactoryStorageContext;
import org.mockito.Mockito;
import org.quartz.JobExecutionContext;
import org.testng.Assert;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import static org.artifactory.schedule.DummyJob.TASKTOKEN;

/**
 * @author Alexei Vainshtein
 */
public class TaskCallbackTest {

    private ArtifactoryStorageContext context;
    private InternalTask activeTask;
    private TaskService taskService;

    @BeforeMethod
    public void setup() {
        activeTask = Mockito.mock(InternalTask.class);
        context = Mockito.mock(ArtifactoryStorageContext.class);
        taskService = Mockito.mock(TaskService.class);
        Mockito.when(taskService.getActiveTask(Mockito.eq(TASKTOKEN), Mockito.eq(true))).thenReturn(activeTask);
        Mockito.when(context.getTaskService()).thenReturn(taskService);
        ArtifactoryContextThreadBinder.bind(context);
    }

    @AfterMethod
    public void tearDown() {
        ArtifactoryContextThreadBinder.unbind();
    }

    @Test
    public void testBeforeExecuteNoContext() {
        JobExecutionContext jobContext = Mockito.mock(JobExecutionContext.class);
        DummyJob dummyJob = new DummyJob();
        Assert.assertFalse(dummyJob.beforeExecute(jobContext));
    }

    @Test
    public void testBeforeExecuteContextNotReady() {
        Mockito.when(context.isReady()).thenReturn(false);
        JobExecutionContext jobContext = Mockito.mock(JobExecutionContext.class);
        DummyJob dummyJob = new DummyJob();
        Assert.assertFalse(dummyJob.beforeExecute(jobContext));

        Mockito.when(activeTask.isSingleExecution()).thenReturn(true);
        Assert.assertFalse(dummyJob.beforeExecute(jobContext));
        Mockito.verify(activeTask, Mockito.times(1)).setShouldRescheduleTask(Mockito.eq(true));
    }

    @Test
    public void testBeforeExecuteContextReady() {
        Mockito.when(context.isReady()).thenReturn(true);
        AddonsManager addonsManager = Mockito.mock(AddonsManager.class);
        HaCommonAddon haCommonAddon = Mockito.mock(HaCommonAddon.class);
        Mockito.when(haCommonAddon.isHaEnabled()).thenReturn(false);
        Mockito.when(addonsManager.addonByType(HaCommonAddon.class)).thenReturn(haCommonAddon);
        Mockito.when(context.beanForType(AddonsManager.class)).thenReturn(addonsManager);
        JobExecutionContext jobContext = Mockito.mock(JobExecutionContext.class);
        DummyJob dummyJob = new DummyJob();
        Mockito.when(activeTask.started()).thenReturn(true);
        Assert.assertTrue(dummyJob.beforeExecute(jobContext));
    }

    @Test
    public void testAfterExecuteWithReschedule() {
        DummyJob dummyJob = new DummyJob();
        JobExecutionContext jobContext = Mockito.mock(JobExecutionContext.class);
        Assert.assertFalse(dummyJob.beforeExecute(jobContext));
        Mockito.when(activeTask.shouldRescheduleTask()).thenReturn(true);
        dummyJob.afterExecute();
        Mockito.verify(taskService, Mockito.times(1)).rescheduleTask(Mockito.eq(TASKTOKEN));
        Mockito.verify(taskService, Mockito.times(0)).cancelTask(Mockito.eq(TASKTOKEN), Mockito.eq(true));
        Mockito.verify(activeTask, Mockito.times(0)).completed();
    }

    @Test
    public void testAfterExecuteNoReschedule() {
        DummyJob dummyJob = new DummyJob();
        JobExecutionContext jobContext = Mockito.mock(JobExecutionContext.class);
        Assert.assertFalse(dummyJob.beforeExecute(jobContext));
        Mockito.when(activeTask.shouldRescheduleTask()).thenReturn(false);
        Mockito.when(activeTask.isSingleExecution()).thenReturn(true);
        dummyJob.afterExecute();
        Mockito.verify(taskService, Mockito.times(0)).rescheduleTask(Mockito.eq(TASKTOKEN));
        Mockito.verify(taskService, Mockito.times(1)).cancelTask(Mockito.eq(TASKTOKEN), Mockito.eq(true));
        Mockito.verify(activeTask, Mockito.times(1)).completed();
    }
}