package org.artifactory.storage.db.binstore;

import com.jfrog.sysconf.SysConfig;
import io.opentracing.Tracer;
import org.apache.http.impl.client.CloseableHttpClient;
import org.artifactory.addon.AddonType;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.ha.HaCommonAddon;
import org.artifactory.addon.retention.ArtifactRetentionAddon;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.ha.HaNodeProperties;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.entitlement.EntitlementType;
import org.artifactory.opentracing.ArtifactoryTracer;
import org.artifactory.security.access.AccessService;
import org.artifactory.storage.db.InternalDbService;
import org.artifactory.storage.db.servers.service.ArtifactoryServersCommonService;
import org.artifactory.util.HttpClientConfigurator;
import org.jfrog.access.client.AccessClient;
import org.jfrog.client.http.model.ProxyConfig;
import org.jfrog.common.concurrency.ConflictsGuard;
import org.jfrog.security.crypto.JFrogMasterKeyEncrypter;
import org.jfrog.storage.DbType;
import org.jfrog.storage.JdbcHelper;
import org.jfrog.storage.config.BinaryStoreConfigService;
import org.jfrog.storage.config.BinaryStoreConstValues;
import org.jfrog.storage.events.DeletedBinariesEvent;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationEvent;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.context.annotation.Lazy;
import org.springframework.stereotype.Service;

import java.io.File;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import static org.apache.commons.lang.StringUtils.isNotBlank;
import static org.artifactory.state.ArtifactoryServerState.RUNNING;

/**
 * @author TamirH
 */
@Service
public class BinaryStoreConfigServiceImpl implements BinaryStoreConfigService {

    private final JdbcHelper jdbcHelper;
    private final ArtifactoryTracer artifactoryTracer;
    private final ApplicationEventPublisher eventPublisher;
    private AccessService accessService;
    private InternalDbService dbService;

    // Some db tests doesn't initialize the addonManager bean,
    // but do have a dummy context that contains it .
    public AddonsManager getAddonsManager() {
        return ContextHelper.get().beanForType(AddonsManager.class);
    }

    @Autowired
    public BinaryStoreConfigServiceImpl(JdbcHelper jdbcHelper, ArtifactoryTracer artifactoryTracer,
            ApplicationEventPublisher eventPublisher) {
        this.jdbcHelper = jdbcHelper;
        this.artifactoryTracer = artifactoryTracer;
        this.eventPublisher = eventPublisher;
    }

    @Override
    public JdbcHelper getJdbcHelper() {
        return this.jdbcHelper;
    }

    @Override
    public DbType getDbType() {
        return ArtifactoryHome.get().getDBProperties().getDbType();
    }

    public Tracer getTracer() {
        return artifactoryTracer.getTracer();
    }

    @Override
    public File getArtifactoryDataDir() {
        return ArtifactoryHome.get().getDataDir();
    }

    @Override
    public boolean isAol() {
        return getAddonsManager().isAddonSupported(AddonType.AOL);
    }

    @Override
    public boolean isLicenseInstalled() {
        return getAddonsManager().isLicenseInstalled();
    }

    @Override
    public boolean isEntitlementEnforced() {
        return getAddonsManager().isEntitlementEnforced();
    }

    @Override
    public boolean isEntitledForRshProvider() {
        return getAddonsManager().isEntitled(EntitlementType.BINARY_PROVIDER_RSH);
    }

    @Override
    public boolean isEntitledForCustomEndpoint() {
        return getAddonsManager().isEntitled(EntitlementType.BINARY_PROVIDER_CUSTOM_ENDPOINT);
    }

    @Override
    public boolean isEntitledCdn() {
        return getAddonsManager().isEntitled(EntitlementType.CDN);
    }

    @Override
    public ConflictsGuard<String> getConflictsGuard(String type) {
        return getAddonsManager().addonByType(HaCommonAddon.class).getConflictsGuard(type);
    }

    @Override
    public Map<String, String> getArtifactRetentionBinaryHeaders() {
        ArtifactRetentionAddon artifactRetentionAddon = getAddonsManager().addonByType(ArtifactRetentionAddon.class);
        return artifactRetentionAddon.getCreateBinaryHeaders();
    }

    @Override
    public Map<String, String> getArtifactRetentionInitiateBinaryRestoreHeaders() {
        ArtifactRetentionAddon artifactRetentionAddon = getAddonsManager().addonByType(ArtifactRetentionAddon.class);
        return artifactRetentionAddon.getInitiateBinaryRestoreHeaders();
    }

    @Override
    public Map<String, String> getArtifactRetentionIsBinaryRestoredHeaders() {
        ArtifactRetentionAddon artifactRetentionAddon = getAddonsManager().addonByType(ArtifactRetentionAddon.class);
        return artifactRetentionAddon.getIsBinaryRestoredHeaders();
    }

    @Override
    public String getArtifactoryRunningModeName() {
        return getAddonsManager().getArtifactoryRunningMode().name();
    }

    @Override
    public boolean isTrialLicense() {
        return getAddonsManager().isTrialLicense();
    }

    @Override
    public boolean isJCR() {
        return getAddonsManager().getArtifactoryRunningMode().isJCR();
    }

    @Override
    public SysConfig getSysConfig() {
        return ArtifactoryHome.get().getSysConfig();
    }

    @Override
    public File getBinaryStoreGcpCredentialsFile() {
        return ArtifactoryHome.get().getBinaryStoreGcpCredentialsFile();
    }

    @Override
    public CloseableHttpClient createClosableHttpClient(Integer maxConnectionsPerRoute,
            Integer maxConnectionsTotal, Integer connectionTimeout,
            ProxyConfig proxy, boolean withParams) {
        HttpClientConfigurator configurator = new HttpClientConfigurator();
        if (maxConnectionsPerRoute != null) {
            configurator.maxConnectionsPerRoute(maxConnectionsPerRoute);
        }
        if (maxConnectionsTotal != null) {
            configurator.maxTotalConnections(maxConnectionsTotal);
        }
        if (connectionTimeout != null) {
            configurator.connectionTimeout(connectionTimeout);
        }
        if (proxy != null && isNotBlank(proxy.getHost())) {
            configurator.proxy(proxy);
        }
        if (withParams) {
            return configurator.build(true);
        }
        return configurator.build();
    }

    @Override
    public boolean isHaConfigured() {
        return ArtifactoryHome.get().isHaConfigured();
    }

    @Override
    public int getIntlValue(BinaryStoreConstValues value) {
        return ConstantValues.valueOf(value.getPropertyName()).getInt();
    }

    @Override
    public boolean getBoolValue(BinaryStoreConstValues value) {
        return ConstantValues.valueOf(value.getPropertyName()).getBoolean();
    }

    @Override
    public long getLongValue(BinaryStoreConstValues value) {
        return ConstantValues.valueOf(value.getPropertyName()).getLong();
    }

    @Override
    public void publishStorageEvent(ApplicationEvent event) {
        eventPublisher.publishEvent(event);
    }

    @Override
    public String getServerId() {
        return ContextHelper.get().getServerId();
    }

    @Override
    public AccessClient getAccessClient() {
        return this.accessService.getAccessClient();
    }

    @Override
    public Properties getHadNodeProperties() {
        final HaNodeProperties haNodeProperties = ArtifactoryHome.get().getHaNodeProperties();
        return haNodeProperties != null ? haNodeProperties.getProperties() : null;
    }

    @Override
    public JFrogMasterKeyEncrypter getJfrogMasterKey() {
        return ArtifactoryHome.get().getJFrogMasterKeyEncrypter();
    }

    @Override
    public File getArtifactoryEtcDir() {
        return ArtifactoryHome.get().getEtcDir();
    }

    @Override
    public File getArtifactorySecurityDir() {
        return ArtifactoryHome.get().getSecurityDir();
    }

    @Override
    public String getClusterDataDir() {
        final HaNodeProperties haNodeProperties = ArtifactoryHome.get().getHaNodeProperties();
        return haNodeProperties != null ? haNodeProperties.getClusterDataDir() : null;
    }

    @Override
    public long getNextDbId() {
        return this.dbService.nextId();
    }

    @Override
    public void publishDeletedBinaryEvent(List<String> binariesList) {
        this.eventPublisher.publishEvent(new DeletedBinariesEvent(this, binariesList));
    }

    @Override
    public int getNumOfActiveServersInCluster() {
        ArtifactoryServersCommonService artifactoryServersCommonService =
                ContextHelper.get().beanForType(ArtifactoryServersCommonService.class);
        return (int) artifactoryServersCommonService.getActiveMembers().stream()
                .filter(server -> server.getServerState().equals(
                        RUNNING)).count();
    }

    //we have no choice but to create the artifactory services as lazy, as in some circumstances they create
    //cyclic dependency for this config class (usually because of AccessService)
    //binaryStoreConfigService -> dbServiceImpl -> artifactoryEncryptionService (via postInitialize in an if condition)
    // -> accessService (deleted dependency now) -> repositoryService -> securityService -> projectsService
    // -> storageService -> artifactoryBinaryStoreService -> BinaryStoreService -> binaryStoreConfigService

    @Autowired
    @Lazy
    public void setAccessService(AccessService accessService) {
        this.accessService = accessService;
    }

    @Autowired
    @Lazy
    public void setInternalDbService(InternalDbService internalDbService) {
        this.dbService = internalDbService;
    }
}
