package org.artifactory.storage.db;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.Lists;
import com.google.common.collect.Sets;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.config.db.ArtifactoryDbProperties;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.common.storage.db.properties.DbVersionInfo;
import org.artifactory.storage.db.init.ArtifactoryTomcatDataSource;
import org.artifactory.storage.db.init.DbInitializationManager;
import org.artifactory.storage.db.properties.service.ArtifactoryDbPropertiesService;
import org.artifactory.storage.db.util.IdGenerator;
import org.artifactory.storage.db.util.JdbcHelper;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.artifactory.test.ArtifactoryHomeStub;
import org.awaitility.Awaitility;
import org.jfrog.storage.DataSourceHealthException;
import org.jfrog.storage.DbType;
import org.jfrog.storage.JFrogDataSource;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.mockito.stubbing.Answer;
import org.springframework.test.util.ReflectionTestUtils;
import org.testng.Assert;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Ignore;
import org.testng.annotations.Test;

import java.io.IOException;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.List;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.anyString;
import static org.mockito.Mockito.doAnswer;
import static org.mockito.Mockito.doNothing;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.doThrow;
import static org.mockito.Mockito.matches;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.testng.Assert.expectThrows;
import static org.testng.Assert.fail;


/**
 * @author Roman Gurevitch.
 */
@Test
public class DbServiceImplTest extends ArtifactoryHomeBoundTest {

    private DbServiceImpl dbService;

    @Mock
    private JdbcHelper jdbcHelper;
    @Mock
    private JFrogDataSource dataSource;
    @Mock
    private Connection connection;
    @Mock
    private DatabaseMetaData dbMetadata;
    @Mock
    private ArtifactoryTomcatDataSource lockDatasource;
    @Mock
    private Connection lockDbConnection;
    @Mock
    private DatabaseMetaData lockConnectionMetadata;
    @Mock
    private ArtifactoryDbProperties dbProperties;
    @Mock
    private ArtifactoryDbPropertiesService dbPropertiesService;
    @Mock
    private IdGenerator idGenerator;

    private final Set<JFrogDataSource> allDataSources = Sets.newHashSet();

    // the ddl are here to ensure that if the file we are using is changed (conversion ddl), we will not accidentally
    //miss it
    private final List<String> ORACLE_DISTRIBUTED_LOCK_INDEXES = ImmutableList
            .of("CREATE INDEX distributed_locks_owner ON distributed_locks (owner)",
                    "CREATE INDEX distributed_locks_owner_thread ON distributed_locks (owner_thread)");
    private final String ORACLE_DISTRIBUTED_LOCK_SQL = "CREATE TABLE distributed_locks (\n" +
            "  category       VARCHAR(64)    NOT NULL,\n" +
            "  lock_key       VARCHAR(255)   NOT NULL,\n" +
            "  owner          VARCHAR(64)    NOT NULL,\n" +
            "  owner_thread   NUMBER(19, 0)  NOT NULL,\n" +
            "  owner_thread_name     VARCHAR(64)   NOT NULL,\n" +
            "  acquire_time   NUMBER(19, 0)  NOT NULL,\n" +
            "  CONSTRAINT locks_pk PRIMARY KEY (category,lock_key)\n" +
            ")";

    private final List<String> MSSQL_DISTRIBUTED_LOCK_INDEXES = ImmutableList
            .of("CREATE INDEX distributed_locks_owner ON distributed_locks (owner)",
                    "CREATE INDEX distributed_locks_owner_thread ON distributed_locks (owner_thread)");
    private final String MSSQL_DISTRIBUTED_LOCK_SQL = "CREATE TABLE distributed_locks (\n" +
            "  category       VARCHAR(64)   NOT NULL,\n" +
            "  lock_key       VARCHAR(255)  NOT NULL,\n" +
            "  owner          VARCHAR(64)   NOT NULL,\n" +
            "  owner_thread   BIGINT        NOT NULL,\n" +
            "  owner_thread_name     VARCHAR(64)   NOT NULL,\n" +
            "  acquire_time   BIGINT        NOT NULL,\n" +
            "  CONSTRAINT locks_pk PRIMARY KEY (category,lock_key)\n" +
            ")";

    private final List<String> MYSQL_DISTRIBUTED_LOCK_INDEXES = ImmutableList
            .of("CREATE INDEX distributed_locks_owner ON distributed_locks (owner)",
                    "CREATE INDEX distributed_locks_owner_thread ON distributed_locks (owner_thread)");
    private final String MYSQL_DISTRIBUTED_LOCK_SQL = "CREATE TABLE distributed_locks (\n" +
            "  category       VARCHAR(64)   NOT NULL,\n" +
            "  lock_key       VARCHAR(255)      NOT NULL,\n" +
            "  owner          VARCHAR(64)   NOT NULL,\n" +
            "  owner_thread   BIGINT     NOT NULL,\n" +
            "  owner_thread_name     VARCHAR(64)   NOT NULL,\n" +
            "  acquire_time   BIGINT     NOT NULL,\n" +
            "  CONSTRAINT locks_pk PRIMARY KEY (category,lock_key)\n" +
            ")";

    private final List<String> PSQL_DISTRIBUTED_LOCK_INDEXES = ImmutableList
            .of("CREATE INDEX distributed_locks_owner ON distributed_locks (owner)",
                    "CREATE INDEX distributed_locks_owner_thread ON distributed_locks (owner_thread)");
    private final String PSQL_DISTRIBUTED_LOCK_SQL = "CREATE UNLOGGED TABLE distributed_locks (\n" +
            "  category       VARCHAR(64)   NOT NULL,\n" +
            "  lock_key       VARCHAR(255)      NOT NULL,\n" +
            "  owner          VARCHAR(64)   NOT NULL,\n" +
            "  owner_thread   BIGINT     NOT NULL,\n" +
            "  owner_thread_name     VARCHAR(64)   NOT NULL,\n" +
            "  acquire_time   BIGINT     NOT NULL,\n" +
            "  CONSTRAINT locks_pk PRIMARY KEY (category,lock_key)\n" +
            ")";

    @BeforeMethod
    private void beforeMethod() throws IOException {
        MockitoAnnotations.initMocks(this);
        dbService = spy(new DbServiceImpl());
        doReturn(true).when(dbService).isNewInstallation();
        doNothing().when(dbService).broadcastNewDbInstallation();
        doNothing().when(dbService).updateDbPropertiesTable();

        ArtifactoryHomeStub artifactoryHomeStub = new ArtifactoryHomeStub();
        artifactoryHomeStub.getSkipVerifyPrivilegesMarkerFile().createNewFile();
        ArtifactoryHome.bind(artifactoryHomeStub);

        doAnswer(invocation -> dataSource).when(jdbcHelper).getDataSource();

        dbService.waitBetweenInitialConnectionRetry = 1;
        ReflectionTestUtils.setField(dbService, "jdbcHelper", jdbcHelper);
        ReflectionTestUtils.setField(dbService, "dbProperties", dbProperties);
        ReflectionTestUtils.setField(dbService, "dbPropertiesService", dbPropertiesService);
        ReflectionTestUtils.setField(dbService, "idGenerator", idGenerator);
        ReflectionTestUtils.setField(dbService, "allDataSources", allDataSources);

        allDataSources.clear();

        ResultSet rsMock = mock(ResultSet.class);
        try {
            when(rsMock.next()).thenReturn(true);
            when(rsMock.getBoolean(1)).thenReturn(false);
            when(rsMock.getString(1)).thenReturn("Sensitive");
            when(jdbcHelper.executeSelect(anyString())).thenReturn(rsMock);
        } catch (SQLException e) {
            Assert.fail();
        }
    }

    public void testRetryConnection() throws Exception {
        when(dataSource.getConnection()).thenThrow(new SQLException("Testing retry connection"));
        when(jdbcHelper.getDataSource()).thenReturn(dataSource);

        DbInitializationManager dbInitializingInstallationManager = spy(new DbInitializationManager(
                dbProperties, jdbcHelper, lockDatasource));
        expectThrows(IllegalStateException.class, dbInitializingInstallationManager::init);

        verify(dataSource, times(3)).getConnection();
    }

    @Test(dataProvider = "provideDbTypeWithSql")
    public void testDifferentDbForLockingWhenTableNotExists(DbType dbType, String expectedCreateTableSql,
            List<String> expectedIndexes) throws Exception {
        testDifferentDbForLock(false, dbType);
        Statement stmtMock = mock(Statement.class);
        when(lockDbConnection.createStatement()).thenReturn(stmtMock);
        DbInitializationManager dbInitializingInstallationManager = spy(new DbInitializationManager(
                dbProperties, jdbcHelper, lockDatasource));
        dbInitializingInstallationManager.init();

        dbService.initDatabaseService();
        verify(lockConnectionMetadata, times(1)).getTables(any(), any(), matches("distributed_locks"), any());
        verify(lockDatasource, times(1)).getConnection();
        verify(lockDbConnection, times(1)).createStatement();
        verify(stmtMock, times(1)).executeUpdate(expectedCreateTableSql);
        expectedIndexes.forEach(index -> {
            try {
                verify(stmtMock, times(1)).executeUpdate(index);
            } catch (SQLException e) {
                fail("Failed validating index creation.", e);
            }
        });
    }

    /**
     * Verify that dbService is healthy
     * and also verify the cache allows only a single health check per data source based on
     * {@link org.artifactory.common.ConstantValues#dbDatasourceHealthCheckCacheInvalidationIntervalMillis
     */
    @Test
    public void testHealthy() {
        for (int i = 0; i < 10; i++) {
            dbService.health();
        }
        allDataSources.forEach(ds -> verify(ds, times(1)).health());
    }

    /**
     * Verify the cache allows more than a single health check per data source after
     * {@link org.artifactory.common.ConstantValues#dbDatasourceHealthCheckCacheInvalidationIntervalMillis expires
     */
    @Test
    public void testHealthCheckMoreThanOneCheckAfterCacheInterval() {
        try {
            getBound().setProperty(ConstantValues.dbDatasourceHealthCheckCacheInvalidationIntervalMillis, "0");
            for (int i = 0; i < 5; i++) {
                dbService.health();
            }
            allDataSources.forEach(ds -> verify(ds, times(5)).health());
        } finally {
            getBound().setProperty(ConstantValues.dbDatasourceHealthCheckCacheInvalidationIntervalMillis,
                    ConstantValues.dbDatasourceHealthCheckCacheInvalidationIntervalMillis.getDefValue());
        }
    }

    /**
     * Verify the health check locking mechanism - Only a single thread out of a burst is allowed to acquire
     * the lock and send the health checks.
     */
    @Test
    public void testHealthCacheLockMechanism() throws InterruptedException {
        Runnable health = () -> dbService.health();
        ExecutorService executorService = Executors.newFixedThreadPool(5);
        for (int i = 0; i < 5; i++) {
            executorService.submit(health);
        }
        executorService.shutdown();
        boolean termination = executorService.awaitTermination(10, TimeUnit.SECONDS);
        Assert.assertTrue(termination);
        allDataSources.forEach(ds -> verify(ds, times(1)).health());
    }

    /**
     * Verify that DB service is unhealthy, should throw DataSourceHealthException
     *
     */
    @Test(expectedExceptions = DataSourceHealthException.class)
    public void testUnHealthyException() {
        JFrogDataSource dataSource = mock(JFrogDataSource.class);
        doThrow(DataSourceHealthException.class).when(dataSource).health();
        allDataSources.add(dataSource);
        dbService.health();
    }

    @Test(expectedExceptions = DataSourceHealthException.class, expectedExceptionsMessageRegExp = "Datasource is not " +
            "healthy: Interrupted by test")
    public void testThreadInterruptedThrowException() {
        JFrogDataSource dataSource = mock(JFrogDataSource.class);
        doAnswer((Answer<Void>) invocationOnMock -> {
            throw new InterruptedException("Interrupted by test");
        }).when(dataSource).health();
        allDataSources.add(dataSource);
        dbService.health();
    }

    /**
     * When a thread is stuck during health check, other threads from subsequent checks should fail
     */
    @Test
    @Ignore // https://jfrog-int.atlassian.net/browse/RTDEV-29981
    public void testHealthCheckThreadStuckSubsequentThreadsShouldFail() throws InterruptedException {
        try {
            final int NUM_THREADS = 5;
            ExecutorService executorService = Executors.newFixedThreadPool(NUM_THREADS);

            getBound().setProperty(ConstantValues.dbDatasourceHealthCheckCacheInvalidationIntervalMillis, "0");
            getBound().setProperty(ConstantValues.dbDataSourceHealthCheckLockRetrievalTimeoutMillis, "50");

            AtomicInteger healthFailCount = new AtomicInteger(0);

            JFrogDataSource stuckDataSource = mock(JFrogDataSource.class);
            doAnswer((Answer<Void>) invocationOnMock -> {
                System.out.println(
                        "Thread:" + Thread.currentThread().getId() + " Waiting for other steps to timeout..");
                Awaitility.waitAtMost(1, TimeUnit.SECONDS)
                        .until(() -> NUM_THREADS - 1 == healthFailCount.get());
                return null;
            }).when(stuckDataSource).health();
            allDataSources.add(stuckDataSource);

            List<DataSourceHealthException> exceptionsThrown = Lists.newArrayList();

            Runnable health = () -> {
                try {
                    System.out.println("Thread:" + Thread.currentThread().getId() + " calling health()...");
                    dbService.health();
                    System.out.println("Thread:" + Thread.currentThread().getId() + " did not throw exception...");
                } catch (DataSourceHealthException e) {
                    exceptionsThrown.add(e);
                    healthFailCount.incrementAndGet();
                }
            };

            for (int i = 0; i < NUM_THREADS; i++) {
                executorService.submit(health);
            }
            executorService.shutdown();
            boolean termination = executorService.awaitTermination(2, TimeUnit.SECONDS);
            Assert.assertTrue(termination);
            Assert.assertEquals(healthFailCount.get(), NUM_THREADS - 1);
            exceptionsThrown.forEach(e -> Assert.assertEquals(e.getMessage(), "Datasource is not healthy. Failed to " +
                    "acquire lock for 0 seconds"));
        } finally {
            getBound().setProperty(ConstantValues.dbDatasourceHealthCheckCacheInvalidationIntervalMillis,
                    ConstantValues.dbDatasourceHealthCheckCacheInvalidationIntervalMillis.getDefValue());
            getBound().setProperty(ConstantValues.dbDataSourceHealthCheckLockRetrievalTimeoutMillis,
                    ConstantValues.dbDataSourceHealthCheckLockRetrievalTimeoutMillis.getDefValue());
        }
    }

    private void testDifferentDbForLock(boolean mockTableExists, DbType dbType) throws Exception {
        when(jdbcHelper.getDataSource()).thenReturn(dataSource);
        when(dataSource.getConnection()).thenReturn(connection);
        when(dbProperties.getDbType()).thenReturn(dbType);
        mockMetadata(dbMetadata, connection, true);
        when(dbPropertiesService.getDbVersionInfo()).thenReturn(
                new DbVersionInfo(111, "Z", 222, 333));

        // prepare external db lock props
        when(dbProperties.getLockingDbSpecificType()).thenReturn(Optional.of(dbType));
        when(lockDatasource.getConnection()).thenReturn(lockDbConnection);
        mockMetadata(lockConnectionMetadata, lockDbConnection, mockTableExists);
    }

    private void mockMetadata(DatabaseMetaData dbMetadata, Connection connection, boolean mockTableExists)
            throws SQLException {
        when(connection.getMetaData()).thenReturn(dbMetadata);
        when(dbMetadata.getDatabaseProductName()).thenReturn("test");
        when(dbMetadata.getDatabaseProductVersion()).thenReturn("test");
        when(dbMetadata.getDriverName()).thenReturn("test");
        when(dbMetadata.getDriverVersion()).thenReturn("test");
        when(dbMetadata.getURL()).thenReturn("url");
        when(dbMetadata.getUserName()).thenReturn("username");
        when(dbMetadata.getConnection()).thenReturn(connection);
        ResultSet resultSet = mock(ResultSet.class);
        when(dbMetadata.getTables(any(), any(), any(), any())).thenReturn(resultSet);
        when(resultSet.next()).thenReturn(mockTableExists);
        if (mockTableExists) {
            when(resultSet.getString(any())).thenReturn("some table");
        }
    }

    @DataProvider
    private Object[][] provideDbTypeWithSql() {
        return new Object[][] {
                {DbType.ORACLE, ORACLE_DISTRIBUTED_LOCK_SQL, ORACLE_DISTRIBUTED_LOCK_INDEXES},
                {DbType.POSTGRESQL, PSQL_DISTRIBUTED_LOCK_SQL, PSQL_DISTRIBUTED_LOCK_INDEXES},
                {DbType.MSSQL, MSSQL_DISTRIBUTED_LOCK_SQL, MSSQL_DISTRIBUTED_LOCK_INDEXES},
                {DbType.MYSQL, MYSQL_DISTRIBUTED_LOCK_SQL, MYSQL_DISTRIBUTED_LOCK_INDEXES}
        };
    }
}