package org.artifactory.storage.db.binstore.service.cleanup;

import org.artifactory.common.ConstantValues;
import org.artifactory.repo.trash.TrashService;
import org.artifactory.storage.StorageException;
import org.artifactory.storage.binstore.service.ArtifactoryBinaryService;
import org.artifactory.storage.binstore.service.cleanup.BinariesGarbageCollectorService;
import org.artifactory.storage.binstore.service.cleanup.GCService;
import org.artifactory.storage.binstore.service.cleanup.GarbageCollectionDatastoreService;
import org.artifactory.storage.fs.service.ConfigsService;
import org.artifactory.test.ArtifactoryHomeBoundJUnitTest;
import org.jfrog.storage.binstore.exceptions.BinaryDbException;
import org.jfrog.storage.binstore.ifc.GarbageCollectorInfo;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Mockito;
import org.springframework.test.util.ReflectionTestUtils;

import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.IntStream;

import static org.artifactory.storage.fs.service.ConfigsServiceKey.ARTIFACTORY_GC_ITERATION;
import static org.mockito.ArgumentMatchers.anyLong;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.any;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.never;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;

/**
 * @author TamirH
 */
class GCServiceImplTest extends ArtifactoryHomeBoundJUnitTest {
    ArtifactoryBinaryService binaryService;
    BinariesGarbageCollectorService minorGcCollectorService;
    TrashService trashService;
    GarbageCollectionDatastoreService garbageCollectionDatastoreService;
    GCService gcService;
    ConfigsService configsService;

    @BeforeEach
    public void init() {
        binaryService = mock(ArtifactoryBinaryService.class);
        minorGcCollectorService = mock(BinariesGarbageCollectorService.class);
        trashService = mock(TrashService.class);
        when(trashService.isTrashcanEnabled()).thenReturn(true);
        garbageCollectionDatastoreService = mock(GarbageCollectionDatastoreService.class);
        configsService = mock(ConfigsService.class);
        gcService = new GCServiceImpl(binaryService, minorGcCollectorService, trashService,
                garbageCollectionDatastoreService, configsService);
        defaultConfigServiceInit();
    }

    @Test
    void testGcFlow() {
        when(minorGcCollectorService.startMinorGC()).thenReturn(
                new GarbageCollectorInfo("Strategy", 10));
        when(binaryService.runFullGarbageCollect()).thenReturn(
                new GarbageCollectorInfo("Strategy", 10));
        IntStream.range(0, 30).forEach(i -> gcService.startGarbageCollect());
        verify(minorGcCollectorService, times(30))
                .startMinorGC();
        verify(binaryService, times(1)).runFullGarbageCollect();
        verify(garbageCollectionDatastoreService, times(30)).saveLastGCRunInfo(any());
    }

    @Test
    void testGcWithoutResult() {
        IntStream.range(0, 30).forEach(i -> gcService.startGarbageCollect());
        verify(minorGcCollectorService, times(30))
                .startMinorGC();
        verify(binaryService, times(1)).runFullGarbageCollect();
        verify(garbageCollectionDatastoreService, times(0)).saveLastGCRunInfo(any());
    }

    @Test
    void testGcCount() {
        IntStream.range(0, 50).forEach(i -> gcService.startGarbageCollect());
        verify(minorGcCollectorService, times(50))
                .startMinorGC();
        verify(binaryService, times(2)).runFullGarbageCollect();
        verify(garbageCollectionDatastoreService, times(0)).saveLastGCRunInfo(any());
    }

    private void defaultConfigServiceInit() {
        final AtomicInteger counter = new AtomicInteger(0);
        when(configsService.getConfig(eq(ARTIFACTORY_GC_ITERATION))).thenAnswer((i) -> String.valueOf(counter));
        when(configsService.addOrUpdateConfig(eq(ARTIFACTORY_GC_ITERATION), anyString(), anyLong())).thenAnswer(i -> {
            counter.set(Integer.parseInt(i.getArgument(1)));
            return true;
        });
    }

    @Test
    void testGcCountWhileCannotFetchConfig() {
        when(configsService.getConfig(eq(ARTIFACTORY_GC_ITERATION))).thenReturn("1").thenReturn("2")
                .thenThrow(new StorageException("Something went wrong"));
        IntStream.range(0, 50).forEach(i -> gcService.startGarbageCollect());
        verify(minorGcCollectorService, times(50))
                .startMinorGC();
        verify(binaryService, times(2)).runFullGarbageCollect();
        verify(garbageCollectionDatastoreService, times(0)).saveLastGCRunInfo(any());
        verify(configsService, times(2)).addOrUpdateConfig(eq(ARTIFACTORY_GC_ITERATION), anyString(), anyLong());
    }

    @Test
    void testGcNullIteration() {
        when(configsService.getConfig(eq(ARTIFACTORY_GC_ITERATION))).thenReturn(null);
        gcService.startGarbageCollect();
        verify(minorGcCollectorService, times(1))
                .startMinorGC();
        verify(binaryService, never()).runFullGarbageCollect();
        verify(configsService, times(1)).addOrUpdateConfig(eq(ARTIFACTORY_GC_ITERATION), eq("1"), anyLong());
    }

    @Test
    public void testGcIterationFallBack() throws BinaryDbException {
        when(configsService.getConfig(eq(ARTIFACTORY_GC_ITERATION)))
                //gc increment doesn't exist in Configs
                .thenReturn(null)
                .thenReturn("1")
                //couldn't retrieve gc increment from DB
                .thenThrow(RuntimeException.class)
                .thenReturn(ConstantValues.gcSkipFullGcBetweenMinorIterations.getInt() - 1 + "").thenReturn("0");
        when(trashService.isTrashcanEnabled()).thenReturn(true);
        AtomicInteger gcIteration = (AtomicInteger) ReflectionTestUtils.getField(gcService, "fallbackGcIteration");
        gcService.startGarbageCollect();
        verify(configsService).addOrUpdateConfig(eq(ARTIFACTORY_GC_ITERATION), eq("1"), anyLong());
        assertEquals(gcIteration.get(), 1);
        gcService.startGarbageCollect();
        verify(configsService).addOrUpdateConfig(eq(ARTIFACTORY_GC_ITERATION), eq("2"), anyLong());
        assertEquals(gcIteration.get(), 2);
        //expect an exception happened, gcIteration was used instead of configs and configs not updated
        Mockito.clearInvocations(configsService);
        gcService.startGarbageCollect();
        assertEquals(gcIteration.get(), 3);
        verify(configsService, never()).addOrUpdateConfig(eq(ARTIFACTORY_GC_ITERATION), anyString(), anyLong());
        gcService.startGarbageCollect();
        verify(configsService).addOrUpdateConfig(eq(ARTIFACTORY_GC_ITERATION), eq("0"), anyLong());
        assertEquals(gcIteration.get(), 0);
        gcService.startGarbageCollect();
        verify(configsService).addOrUpdateConfig(eq(ARTIFACTORY_GC_ITERATION), eq("1"), anyLong());
        assertEquals(gcIteration.get(), 1);
    }

}