/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.storage.db.fs.itest.service;

import org.apache.commons.lang.StringUtils;
import org.artifactory.storage.db.itest.DbBaseTest;
import org.artifactory.storage.fs.service.ConfigsService;
import org.artifactory.storage.fs.service.ConfigsServiceKey;
import org.springframework.beans.factory.annotation.Autowired;
import org.testng.annotations.Test;

import java.sql.SQLException;

import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertFalse;
import static org.testng.Assert.assertNotNull;
import static org.testng.Assert.assertNull;
import static org.testng.Assert.assertTrue;

/**
 * Tests the {@link org.artifactory.storage.fs.service.ConfigsService}
 *
 * @author Yossi Shaul
 */
public class ConfigsServiceImplTest extends DbBaseTest {

    @Autowired
    private ConfigsService configsService;

    public void addConfig() {
        configsService.addConfig("confName", "confinput", System.currentTimeMillis());
    }

    @Test(dependsOnMethods = "addConfig")
    public void hasConfig() {
        assertTrue(configsService.hasConfig("confName"));
    }

    @Test(dependsOnMethods = "addConfig")
    public void getConfig() throws SQLException {
        String data = configsService.getConfig("confName");
        assertNotNull(data);
        assertEquals(data, "confinput");
    }

    @Test(dependsOnMethods = "getConfig")
    public void updateConfig() {
        configsService.updateConfig("confName", "newdata", System.currentTimeMillis());
        String result = configsService.getConfig("confName");
        assertNotNull(result);
        assertEquals(result, "newdata");
    }

    public void hasConfigNoConfig() {
        assertFalse(configsService.hasConfig("nosuchconfig"));
    }

    public void loadConfigNoConfig() {
        assertNull(configsService.getConfig("nosuchconfig"));
    }

    @Test(expectedExceptions = IllegalStateException.class)
    public void updateConfigNoConfig() {
        configsService.updateConfig("nosuchconfig", "newdata", System.currentTimeMillis());
    }

    @Test(dependsOnMethods = "addConfig", expectedExceptions = IllegalStateException.class)
    public void addConfigExistingConfig() {
        configsService.addConfig("confName", "jojjoh", System.currentTimeMillis());
    }

    public void addOrUpdateConfig() {
        boolean created = configsService.addOrUpdateConfig("artifactory", "somedata", System.currentTimeMillis());
        assertTrue(created);
        assertEquals(configsService.getConfig("artifactory"), "somedata");
        created = configsService.addOrUpdateConfig("artifactory", "newdata", System.currentTimeMillis());
        assertFalse(created);
        assertEquals(configsService.getConfig("artifactory"), "newdata");
    }

    @Test(dependsOnMethods = "addConfig", expectedExceptions = IllegalArgumentException.class,
            expectedExceptionsMessageRegExp = "Can not create a backup: config name can not be empty")
    public void backupConfigWhenConfigNameIsEmpty() {
        configsService.backupConfig("", System.currentTimeMillis());
    }

    @Test
    public void backupConfigWhenConfigDataIsEmpty() {
        String configName = "emptyConfig";
        long lastModified = System.currentTimeMillis();
        configsService.backupConfig(configName, lastModified);

        String backupName = String.join("_", configName, "backup", Long.toString(lastModified));
        assertFalse(configsService.hasConfig(backupName));
    }

    @Test(dependsOnMethods = "addConfig", expectedExceptions = IllegalStateException.class,
            expectedExceptionsMessageRegExp = "Can not create a backup with name .* because it already exist.")
    public void backupConfigWhenBackupIsPresent() {
        String configName = "confName";
        long timestamp = System.currentTimeMillis();
        assertTrue(configsService.hasConfig(configName));

        configsService.backupConfig(configName, timestamp);
        String backupName = String.join("_", configName, "backup", Long.toString(timestamp));
        assertTrue(configsService.hasConfig(backupName));

        configsService.backupConfig(configName, timestamp);
    }

    @Test(dependsOnMethods = "addConfig")
    public void backupConfig() {
        String configName = "confName";
        String configData = "confinput";
        long timestamp = System.currentTimeMillis();
        assertTrue(configsService.hasConfig(configName));

        configsService.backupConfig(configName, timestamp);

        String backupName = String.join("_", configName, "backup", Long.toString(timestamp));
        String backupData = configsService.getConfig(backupName);
        assertFalse(StringUtils.isBlank(backupData));
        assertEquals(backupData, configData);
    }

    @Test
    public void testConfigUsingKeys() {
        ConfigsServiceKey keyToUse = ConfigsServiceKey.ARTIFACTORY_GC_ITERATION;
        assertNull(configsService.getConfig(keyToUse), "checking doesn't exist before starts");
        String value = "10";
        assertTrue(configsService.addOrUpdateConfig(keyToUse, value, System.currentTimeMillis()),
                "add config for 1st time using key");
        String result = configsService.getConfig(keyToUse);
        assertNotNull(result, "verify get using key works");
        assertEquals(value, result, "verify getConfig using ConfigsServiceKey works");
        //update value and try again
        value = "11";
        assertFalse(configsService.addOrUpdateConfig(keyToUse, value, System.currentTimeMillis()),
                "modify value and update");
        String result2 = configsService.getConfig(keyToUse);
        assertNotNull(result2, "verify get using key works");
        assertEquals(value, result2, "verify getConfig using ConfigsServiceKey works");
        //delete record
        configsService.deleteConfig(keyToUse.getKeyName());
        assertNull(configsService.getConfig(keyToUse), "verify doesn't exist after delete");
    }
}
