/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.ui.rest.service.admin.configuration.mail;

import org.apache.commons.lang.StringUtils;
import org.artifactory.api.mail.MailServerConfiguration;
import org.artifactory.api.mail.MailService;
import org.artifactory.common.crypto.CryptoHelper;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.descriptor.mail.MailServerDescriptor;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.RestResponse;
import org.artifactory.rest.common.service.RestService;
import org.artifactory.ui.exception.BadRequestException;
import org.artifactory.ui.rest.model.admin.configuration.mail.MailServer;
import org.artifactory.util.AolUtils;
import org.artifactory.util.EmailException;
import org.artifactory.util.MailUtils;
import org.artifactory.util.UrlValidator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import javax.servlet.http.HttpServletResponse;

import static org.artifactory.ui.utils.RequestUtils.PASSWORD_PLACEHOLDER;

/**
 * @author Chen Keinan
 */
@Component
@Scope(BeanDefinition.SCOPE_PROTOTYPE)
public class TestMailService<T extends MailServer> implements RestService<T> {
    private static final Logger log = LoggerFactory.getLogger(TestMailService.class);

    private MailService mailService;

    @Autowired
    public void setMailService(MailService mailService) {
        this.mailService = mailService;
    }

    @Override
    public void execute(ArtifactoryRestRequest<T> request, RestResponse response) {
        AolUtils.assertNotAol("TestMail");
        // get mail configuration instance
        MailServerConfiguration mailServerConfiguration = getMailServerConfiguration(request);
        // validate configuration data
        if (!validateConfig(mailServerConfiguration)) {
            response.error("Sending a test message requires the configuration to be enabled with " +
                    "defined host and port properties, at least.");
            return;
        }
        validateUrl(mailServerConfiguration.getArtifactoryUrl());
        if (PASSWORD_PLACEHOLDER.equals(mailServerConfiguration.getPassword())) {
            String msg = "Mail server password needs to be re-entered to in order to send a test message";
            log.error(msg);
            response.responseCode(HttpServletResponse.SC_BAD_REQUEST);
            response.error(msg);
            return;
        }
        // send Test mail
        sendMail(mailServerConfiguration, request.getImodel(), response);
    }

    /**
     * get mail server configuration
     *
     * @param artifactoryRequest - encapsulate all date related to request
     * @return instance of mail server configuration data
     */
    private MailServerConfiguration getMailServerConfiguration(ArtifactoryRestRequest artifactoryRequest) {
        MailServerDescriptor serverData = (MailServerDescriptor) artifactoryRequest.getImodel();
        return new MailServerConfiguration(
                serverData.isEnabled(), serverData.getHost(), serverData.getPort(), serverData.getUsername(),
                CryptoHelper.decryptIfNeeded(ArtifactoryHome.get(), serverData.getPassword()), serverData.getFrom(),
                serverData.getSubjectPrefix(),
                serverData.isTls(), serverData.isSsl(), serverData.getArtifactoryUrl());
    }

    /**
     * validate mail server setting
     *
     * @param mailServerConfiguration - mail server configuration
     * @return if true - configuration is valid
     */
    private boolean validateConfig(MailServerConfiguration mailServerConfiguration) {
        boolean hasHost = false;
        boolean hasPort = false;
        if (mailServerConfiguration != null) {
            hasHost = !StringUtils.isEmpty(mailServerConfiguration.getHost());
            hasPort = (mailServerConfiguration.getPort() > 0);
        }
        return hasHost && hasPort && mailServerConfiguration.isEnabled();
    }

    private void validateUrl(String url) {
        UrlValidator urlValidator = new UrlValidator(false);
        try {
            urlValidator.validate(url);
        } catch (UrlValidator.UrlValidationException e) {
            throw new BadRequestException("Artifactory Url is in illegal form.");
        }
    }

    /**
     * send test mail to receipt
     *
     * @param configuration - mail server configuration
     * @param model         - mail server model
     * @param response      - encapsulate all data require for response
     */
    private void sendMail(MailServerConfiguration configuration, MailServer model, RestResponse response) {
        String testRecipient = model.getTestReceipt();
        if (!StringUtils.isEmpty(testRecipient)) {
            try {
                mailService.sendMail(new String[] {testRecipient}, "Test",
                        MailUtils.getTestMessage(configuration.getArtifactoryUrl()).toString(),
                        configuration);
                String confirmMessage = "A test message has been sent successfully to '" + testRecipient + "'";
                response.info(confirmMessage);
                log.info(confirmMessage);
            } catch (EmailException e) {
                String message = e.getMessage();
                if (message == null) {
                    message = "An error occurred while sending an e-mail. See logs for additional information";
                }
                log.error(message, e);
                response.error(message);
            }
        } else {
            response.error("Test recipient field cannot be empty");
        }
    }
}
