/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.ui.rest.service.admin.configuration.mail;

import lombok.extern.slf4j.Slf4j;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.descriptor.config.MutableCentralConfigDescriptor;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.RestResponse;
import org.artifactory.rest.common.service.RestService;
import org.artifactory.ui.exception.BadRequestException;
import org.artifactory.ui.rest.model.admin.configuration.mail.MailServer;
import org.artifactory.util.AolUtils;
import org.artifactory.util.UrlValidator;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import static org.artifactory.ui.utils.RequestUtils.PASSWORD_PLACEHOLDER;

/**
 * @author Chen Keinan
 */
@Component
@Scope(BeanDefinition.SCOPE_PROTOTYPE)
@Slf4j
public class UpdateMailService implements RestService {

    private final static String errorMessage =
            "Mail server password needs to be re-entered to allow changes in Mail server settings.";

    @Autowired
    private CentralConfigService centralConfigService;

    @Override
    public void execute(ArtifactoryRestRequest request, RestResponse response) {
        AolUtils.assertNotAol("UpdateMail");
        saveMailChangesToDescriptor(request, response);
    }

    /**
     * save mail changes to descriptor
     *
     * @param artifactoryRequest - encapsulate artifactory data related  to request
     */
    private void saveMailChangesToDescriptor(ArtifactoryRestRequest artifactoryRequest,
            RestResponse artifactoryResponse) {
        MailServer mailServer = (MailServer) artifactoryRequest.getImodel();
        if (PASSWORD_PLACEHOLDER.equals(mailServer.getPassword())) {
            log.error(errorMessage);
            artifactoryResponse.error(errorMessage);
            return;
        }
        validateUrl(mailServer.getArtifactoryUrl());
        MutableCentralConfigDescriptor configDescriptor = centralConfigService.getMutableDescriptor();
        configDescriptor.setMailServer(mailServer);
        centralConfigService.saveEditedDescriptorAndReload(configDescriptor, mailServer);
        artifactoryResponse.info("Successfully updated Mail server settings");
    }

    private void validateUrl(String url) {
        UrlValidator urlValidator = new UrlValidator(false);
        try {
            urlValidator.validate(url);
        } catch (UrlValidator.UrlValidationException e) {
            throw new BadRequestException("Artifactory Url is in illegal form.");
        }
    }
}
