/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.ui.rest.service.artifacts.browse.generic;

import com.google.common.annotations.VisibleForTesting;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.http.HttpStatus;
import org.artifactory.api.repo.BaseBrowsableItem;
import org.artifactory.api.repo.RootNodesFilterResult;
import org.artifactory.api.repo.exception.FolderExpectedException;
import org.artifactory.api.repo.exception.ItemNotFoundRuntimeException;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.RestResponse;
import org.artifactory.rest.exception.ForbiddenException;
import org.artifactory.ui.rest.model.artifacts.browse.generic.NativeBrowserChild;
import org.artifactory.ui.rest.model.artifacts.browse.generic.NativeBrowserModel;
import org.jetbrains.annotations.Nullable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import javax.servlet.http.HttpServletRequest;
import javax.ws.rs.NotFoundException;
import java.util.List;
import java.util.stream.Collectors;

import static org.artifactory.ui.rest.resource.artifacts.browse.generic.NativeBrowserResource.REPO_PATH;

/**
 * @author BarakH
 */
@Component
@Scope(BeanDefinition.SCOPE_PROTOTYPE)
public class BrowseNativeModelService extends BaseBrowseNativeService {

    private static final Logger log = LoggerFactory.getLogger(BrowseNativeModelService.class);

    @Override
    public void execute(ArtifactoryRestRequest<String> request, RestResponse response) {
        try {
            String path = request.getPathParamByKey(REPO_PATH);
            if (StringUtils.isEmpty(path)) {
                throw new NotFoundException("No repository path has been provided");
            }
            RepoPath repoPath = RepoPathFactory.create(path);
            String repoKey = repoPath.getRepoKey();
            if (!repositoryService.getAllRepoKeys().contains(repoKey)) {
                log.debug("Repository does not exist: {}", repoPath.toPath());
                throw new ForbiddenException("");
            }
            boolean isFolder = isFolderPath(path, request, response);
            if (isFolder) {
                final NativeBrowserModel nativeBrowserModel = buildNativeBrowserData(request, response, repoPath);
                if (nativeBrowserModel == null) {
                    return;
                }
                response.iModel(nativeBrowserModel);
            } else {
                response.iModel(
                        new NativeBrowserModel(repoPath.getRepoKey(), repoPath.getPath(), null, false, false, null));
            }
        } catch (ForbiddenException | NotFoundException e) {
            if (authorizationService.isAnonAccessEnabled() && authorizationService.isAnonymous()) {
                setUnauthorizedResponse(response, e.getMessage());
            } else {
                setNotFoundResponse(response, e.getMessage());
            }
        } catch (Exception e) {
            setNotFoundResponse(response, e.getMessage());
        }
    }

    /**
     * Method defines type of provided repo path. It fetches parent's tree, finds item by name and checks its type.
     * If the provided path can't be verified, for example, it does not exist - it will be treated like a folder.
     */
    @VisibleForTesting
    boolean isFolderPath(String path, ArtifactoryRestRequest<String> request, RestResponse response) {
        if (StringUtils.isBlank(path)) {
            return true;
        }
        if (path.endsWith("/")) {
            return true;
        }
        RepoPath originalRepoPath = RepoPathFactory.create(path);
        RepoPath parentRepoPath = originalRepoPath.getParent();
        if (parentRepoPath == null) {
            return true;
        }
        NativeBrowserModel parentBrowserModel = buildNativeBrowserData(request, response, parentRepoPath);
        if (parentBrowserModel == null || CollectionUtils.isEmpty(parentBrowserModel.getChildren())) {
            return true;
        }
        NativeBrowserChild targetItem = parentBrowserModel.getChildren().stream()
                .filter(child -> child.getName().equals(originalRepoPath.getName()))
                .findFirst()
                .orElse(null);
        if (targetItem == null) {
            return true;
        }
        return targetItem.isFolder();
    }

    @Nullable
    @VisibleForTesting
    NativeBrowserModel buildNativeBrowserData(ArtifactoryRestRequest<String> request, RestResponse response,
            RepoPath repoPath) {
        try {
            RootNodesFilterResult rootNodesFilterResult = new RootNodesFilterResult();
            final List<? extends BaseBrowsableItem> browsableItems =
                    getItems(repoPath, null, rootNodesFilterResult);
            boolean isChallengeSet =
                    setResponseToHttpChallenge(repoPath, rootNodesFilterResult, browsableItems, response, true);
            if (isChallengeSet) {
                return null;
            }
            return transformIntoUiModel(repoPath, request.getServletRequest(), browsableItems);
        } catch (FolderExpectedException e) {
            log.debug("", e);
            // This is a file, not a folder.
            if (!hasPermissionOnPath(repoPath)) {
                throw new ForbiddenException("");
            }
            // The UI will open a download request to the provided repoPath.
            return new NativeBrowserModel(repoPath.getRepoKey(), repoPath.getPath(), null, false, false, null);
        } catch (ItemNotFoundRuntimeException e) {
            if (!repoPath.getPath().contains("!/")) {
                throw e;
            }
            log.debug("Retrieving file from inside the archive", e);
            return new NativeBrowserModel(repoPath.getRepoKey(), repoPath.getPath(), null, false, false, null);
        }
    }

    private NativeBrowserModel transformIntoUiModel(RepoPath repoPath, HttpServletRequest request,
            List<? extends BaseBrowsableItem> browsableItems) {
        final List<NativeBrowserChild> nativeItems = browsableItems.stream()
                .map(item -> new NativeBrowserChild(item.getName(), item.getLastModified(), item.getSize(),
                        item.isFolder(), item.isRemote()))
                .collect(Collectors.toList());

        return new NativeBrowserModel(repoPath.getRepoKey(), repoPath.getPath(), getAddressFooter(request),
                repoPath.isRoot(), true, nativeItems);
    }

    private void setNotFoundResponse(RestResponse response, String message) {
        response.error(message).responseCode(HttpStatus.SC_NOT_FOUND);
        log.debug("{} - {}", message, HttpStatus.SC_NOT_FOUND);
    }

    private void setUnauthorizedResponse(RestResponse response, String message) {
        response.error(message).responseCode(HttpStatus.SC_UNAUTHORIZED);
        log.debug("{} - {}", message, HttpStatus.SC_UNAUTHORIZED);
    }
}

