/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.ui.rest.service.setmeup;

import lombok.NonNull;
import org.apache.commons.lang.StringUtils;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.CoreAddons;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.descriptor.repo.HttpRepoDescriptor;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RemoteRepoDescriptor;
import org.artifactory.descriptor.repo.RepoBaseDescriptor;
import org.artifactory.descriptor.repo.RepoDescriptor;
import org.artifactory.descriptor.repo.VirtualRepoDescriptor;
import org.artifactory.repo.InternalRepoPathFactory;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.system.SystemRepos;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.RestResponse;
import org.artifactory.rest.common.service.RestService;
import org.artifactory.ui.rest.model.setmeup.SetMeUpModel;
import org.artifactory.ui.rest.model.utils.repositories.RepoKeyType;
import org.artifactory.ui.utils.RequestUtils;
import org.artifactory.util.AolUtils;
import org.artifactory.util.HttpUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static org.artifactory.api.request.ArtifactoryHeaders.ORIGINAL_CLIENT_URI;

/**
 * @author chen Keinan
 * @author Lior Hasson
 */
@Component
@Scope(BeanDefinition.SCOPE_PROTOTYPE)
public class GetSetMeUpService implements RestService<Void> {

    @Autowired
    private CentralConfigService centralConfigService;

    @Autowired
    private RepositoryService repositoryService;

    @Autowired
    private AuthorizationService authorizationService;

    @Override
    public void execute(ArtifactoryRestRequest<Void> request, RestResponse response) {
        String projectKey = RequestUtils.getProjectKeyFromRequest(request);
        /// get list of repositories descriptor by user permissions
        List<RepoKeyType> repoKeyTypes = getRepoKeyTypes(projectKey);
        // update set me up model
        SetMeUpModel setMeUpModel = getSetMeUpModel(request, repoKeyTypes);
        response.iModel(setMeUpModel);
    }

    SetMeUpModel getSetMeUpModel(ArtifactoryRestRequest<Void> artifactoryRequest, List<RepoKeyType> repoKeyTypes) {
        SetMeUpModel setMeUpModel = new SetMeUpModel();
        setMeUpModel.setRepoKeyTypes(repoKeyTypes);
        // We want to skip the forward host, as we want the response to be leading straight to Artifactory
        String servletContextUrl =
                HttpUtils.getServletContextUrlArtifactoryPreferred(artifactoryRequest.getServletRequest());
        setMeUpModel.setBaseUrl(servletContextUrl);
        String originalClientBaseUrl = artifactoryRequest.getServletRequest().getHeader(ORIGINAL_CLIENT_URI);
        if (StringUtils.isNotBlank(originalClientBaseUrl)) {
            setMeUpModel.setNpmBaseUrl(originalClientBaseUrl);
        }
        setMeUpModel.setServerId(centralConfigService.getServerName());
        setMeUpModel.setHostname(ContextHelper.get().beanForType(AddonsManager.class).addonByType(CoreAddons.class)
                .getArtifactoryServerName());
        setMeUpModel.setSkipDockerDns(AolUtils.skipDockerDns());
        return setMeUpModel;
    }

    /**
     * get user keys type list
     *
     * @return - list of user keys types
     */
    private List<RepoKeyType> getRepoKeyTypes(@NonNull String projectKey) {
        Set<RepoBaseDescriptor> userRepos = getUserRepos(projectKey);
        List<RepoKeyType> repoKeyTypes = new ArrayList<>();
        userRepos.forEach(userRepo -> {
            if (!userRepo.getKey().endsWith("-cache")) {
                RepoKeyType repoKeyType = new RepoKeyType(userRepo.getType(), userRepo.getKey());
                // update can read or deploy
                updateCanReadOrDeploy(userRepo, repoKeyType);

                if (userRepo instanceof LocalRepoDescriptor) {
                    repoKeyType.setIsLocal(true);
                }
                if (userRepo instanceof RemoteRepoDescriptor) {
                    repoKeyType.setIsRemote(true);
                }
                if (userRepo instanceof VirtualRepoDescriptor virtualRepo) {
                    repoKeyType.setIsVirtual(true);
                    if (virtualRepo.getDefaultDeploymentRepo() != null) {
                        repoKeyType.setIsDefaultDeploymentConfigured(true);
                    }
                }
                repoKeyTypes.add(repoKeyType);
            }
        });
        return repoKeyTypes;
    }

    private void updateCanReadOrDeploy(RepoBaseDescriptor userRepo, RepoKeyType repoKeyType) {
        RepoPath repoPath;
        repoKeyType.setCanRead(false);
        repoKeyType.setCanDeploy(false);

        if (userRepo instanceof HttpRepoDescriptor) {
            repoPath = InternalRepoPathFactory.repoRootPath(userRepo.getKey() + "-cache");
        } else {
            repoPath = InternalRepoPathFactory.repoRootPath(userRepo.getKey());
        }

        if (authorizationService.canRead(repoPath) ||
                authorizationService.userHasPermissionsOnRepositoryRoot(repoPath.getRepoKey())) {
            repoKeyType.setCanRead(true);
        }

        if (userRepo instanceof VirtualRepoDescriptor virtualRepo) {
            LocalRepoDescriptor defaultDeploymentRepo = virtualRepo.getDefaultDeploymentRepo();
            if (defaultDeploymentRepo != null) {
                repoPath = InternalRepoPathFactory.repoRootPath(defaultDeploymentRepo.getKey());
            }
        }

        if (authorizationService.canDeploy(repoPath)) {
            repoKeyType.setCanDeploy(true);
        }
    }

    /**
     * get list of repositories allowed for this user to deploy
     *
     * @return - list of repositories
     */
    private Set<RepoBaseDescriptor> getUserRepos(@NonNull String projectKey) {
        Stream<LocalRepoDescriptor> localDescriptors =
                repositoryService.getLocalFederatedAndCachedRepoDescriptorsByProject(projectKey)
                        .stream()
                        .filter(this::isPermittedRepo);
        Stream<RemoteRepoDescriptor> remoteDescriptors =
                repositoryService.getRemoteRepoDescriptorsByProject(projectKey)
                        .stream()
                        .filter(this::isPermittedRepo);
        Stream<VirtualRepoDescriptor> virtualDescriptors =
                repositoryService.getVirtualRepoDescriptorsByProject(projectKey)
                        .stream()
                        .filter(this::isPermittedRepo);
        return Stream.concat(localDescriptors, Stream.concat(remoteDescriptors, virtualDescriptors))
                .collect(Collectors.toCollection(() -> new TreeSet<>(new RepoComparator())));
    }

    /**
     * filter non permitted repositories
     */
    public boolean isPermittedRepo(RepoDescriptor repoDescriptor) {
        if (repoDescriptor == null) {
            return false;
        }
        return authorizationService.userHasPermissionsOnRepositoryRoot(repoDescriptor.getKey()) &&
                !SystemRepos.SystemRepoConfig.isInvisibleSystemRepo(repoDescriptor.getKey());
    }

    private static class RepoComparator implements Comparator<RepoBaseDescriptor> {
        @Override
        public int compare(RepoBaseDescriptor descriptor1, RepoBaseDescriptor descriptor2) {

            //Local repositories can be either ordinary or caches
            if (descriptor1 instanceof LocalRepoDescriptor localDescriptor1 &&
                    descriptor2 instanceof LocalRepoDescriptor localDescriptor2) {
                boolean repo1IsCache = localDescriptor1.isCache();
                boolean repo2IsCache = localDescriptor2.isCache();

                //Cache repositories should appear in a higher priority
                if (repo1IsCache && !repo2IsCache) {
                    return 1;
                } else if (!repo1IsCache && repo2IsCache) {
                    return -1;
                }
            }
            return descriptor1.getKey().compareTo(descriptor2.getKey());
        }
    }
}
