/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes.actions;

import org.artifactory.addon.AddonsManager;
import org.artifactory.api.build.BuildService;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.common.repo.PackageType;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.descriptor.download.FolderDownloadConfigDescriptor;
import org.artifactory.descriptor.repo.RemoteRepoDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.ph.PackageSpecification;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.RepoPathFactory;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.util.List;

import static org.artifactory.repo.RepoDetailsType.FEDERATED_REPO;
import static org.artifactory.repo.RepoDetailsType.LOCAL_REPO;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertFalse;
import static org.testng.Assert.assertNotNull;
import static org.testng.Assert.assertNull;
import static org.testng.Assert.assertTrue;

/**
 * @author Rotem Kfir
 */
@Test
public class TreeNodeActionsPopulatorTest {

    private static final String TRASH_REPO_TYPE = "trash";

    @Mock
    private TreeNodeActionsPopulator populator;
    @Mock
    private RepositoryService repoService;
    @Mock
    private AuthorizationService authService;
    @Mock
    private CentralConfigService centralConfig;

    @Mock
    private BuildService buildService;

    @Mock
    private AddonsManager addonsManager;
    private PackageSpecification packageSpecification;

    AutoCloseable mocks;

    @BeforeMethod
    public void setup() {
        mocks = MockitoAnnotations.openMocks(this);
        packageSpecification = Mockito.mock(PackageSpecification.class);
        when(packageSpecification.isViewable("repo", "path/PACKAGES", RepoType.CRAN)).thenReturn(true);
        populator = new TreeNodeActionsPopulator();
        populator.setAddonsManager(addonsManager);
        populator.setAuthService(authService);
        populator.setCentralConfig(centralConfig);
        populator.setBuildService(buildService);
        populator.setRepoService(repoService);
        populator.setPackageSpecification(packageSpecification);
    }

    @AfterMethod
    public void teardown() throws Exception {
        mocks.close();
    }

    @Test
    public void testSuccessfulDownloadFolderForAnonymous() {
        enableFolderDownload(true);
        makeTheUserAnonymousWithReadPermissions();
        TabOrAction action = populator.downloadFolder(null);
        assertNotNull(action);
    }

    @Test
    public void testDownloadFolderNotEnabledForAnonymous() {
        enableFolderDownload(false);
        makeTheUserAnonymousWithReadPermissions();
        TabOrAction action = populator.downloadFolder(null);
        assertNull(action);
    }

    @Test
    public void testAdminTrashcanActions() {
        when(authService.isAdmin()).thenReturn(true);
        when(authService.canAccessTrashcan()).thenReturn(true);
        when(authService.canRestoreFromTrashcan(any())).thenReturn(true);

        TabsAndActions tabsAndActions = new TabsAndActions();
        tabsAndActions.setRepoPath("some/repo/path");
        tabsAndActions.setRepoType(TRASH_REPO_TYPE);

        populator.populateForRepository(tabsAndActions);
        List<TabOrAction> actions = tabsAndActions.getActions();
        assertTrue(actions.stream().anyMatch(action -> "Refresh".equals(action.getName())));
        assertTrue(actions.stream().anyMatch(action -> "SearchTrash".equals(action.getName())));
        assertTrue(actions.stream().anyMatch(action -> "EmptyTrash".equals(action.getName())));

        populator.populateForFolder(tabsAndActions);
        actions = tabsAndActions.getActions();
        assertTrue(actions.stream().anyMatch(action -> "Refresh".equals(action.getName())));
        assertTrue(actions.stream().anyMatch(action -> "Restore".equals(action.getName())));
        assertTrue(actions.stream().anyMatch(action -> "DeletePermanently".equals(action.getName())));

        populator.populateForFile(tabsAndActions);
        actions = tabsAndActions.getActions();
        assertTrue(actions.stream().anyMatch(action -> "Restore".equals(action.getName())));
        assertTrue(actions.stream().anyMatch(action -> "DeletePermanently".equals(action.getName())));
    }

    @Test
    public void testNonAdminButCanAccessTrashcanActions() {
        when(authService.isAdmin()).thenReturn(false);
        when(authService.canAccessTrashcan()).thenReturn(true);
        when(authService.canRestoreFromTrashcan(any())).thenReturn(true);

        TabsAndActions tabsAndActions = new TabsAndActions();
        tabsAndActions.setRepoPath("some/repo/path");
        tabsAndActions.setRepoType(TRASH_REPO_TYPE);

        populator.populateForRepository(tabsAndActions);
        List<TabOrAction> actions = tabsAndActions.getActions();
        assertTrue(actions.stream().anyMatch(action -> "Refresh".equals(action.getName())));
        assertTrue(actions.stream().anyMatch(action -> "SearchTrash".equals(action.getName())));
        assertFalse(actions.stream().anyMatch(action -> "EmptyTrash".equals(action.getName()))); // false

        populator.populateForFolder(tabsAndActions);
        actions = tabsAndActions.getActions();
        assertTrue(actions.stream().anyMatch(action -> "Refresh".equals(action.getName())));
        assertTrue(actions.stream().anyMatch(action -> "Restore".equals(action.getName())));
        assertFalse(actions.stream().anyMatch(action -> "DeletePermanently".equals(action.getName()))); // false

        populator.populateForFile(tabsAndActions);
        actions = tabsAndActions.getActions();
        assertTrue(actions.stream().anyMatch(action -> "Restore".equals(action.getName())));
        assertFalse(actions.stream().anyMatch(action -> "DeletePermanently".equals(action.getName()))); // false
    }

    @Test
    public void testNonAdminCanAccessTrashcanNoRestorePermissions() {
        when(authService.canAccessTrashcan()).thenReturn(true);
        when(authService.canRestoreFromTrashcan(any())).thenReturn(false);

        TabsAndActions tabsAndActions = new TabsAndActions();
        tabsAndActions.setRepoPath("some/repo/path");
        tabsAndActions.setRepoType(TRASH_REPO_TYPE);

        populator.populateForRepository(tabsAndActions);
        List<TabOrAction> actions = tabsAndActions.getActions();
        assertTrue(actions.stream().anyMatch(action -> "Refresh".equals(action.getName())));
        assertTrue(actions.stream().anyMatch(action -> "SearchTrash".equals(action.getName())));
        assertFalse(actions.stream().anyMatch(action -> "EmptyTrash".equals(action.getName()))); // false

        populator.populateForFolder(tabsAndActions);
        actions = tabsAndActions.getActions();
        assertTrue(actions.isEmpty());

        populator.populateForFile(tabsAndActions);
        actions = tabsAndActions.getActions();
        assertTrue(actions.isEmpty());
    }

    @Test
    public void testNonAdminNoAccessToTrashcanNoRestorePermissions() {
        when(authService.canAccessTrashcan()).thenReturn(false);

        TabsAndActions tabsAndActions = new TabsAndActions();
        tabsAndActions.setRepoPath("some/repo/path");
        tabsAndActions.setRepoType(TRASH_REPO_TYPE);

        populator.populateForRepository(tabsAndActions);
        List<TabOrAction> actions = tabsAndActions.getActions();
        assertTrue(actions.isEmpty());

        populator.populateForFolder(tabsAndActions);
        actions = tabsAndActions.getActions();
        assertTrue(actions.isEmpty());

        populator.populateForFile(tabsAndActions);
        actions = tabsAndActions.getActions();
        assertTrue(actions.isEmpty());
    }

    @Test(dataProvider = "getOptions")
    public void testUserCanDownloadFromRemote(boolean userCanDeploy, boolean fileExistInCache,
            boolean remoteCacheEnabled, boolean downloadShouldBeAvailable) {
        setupRepoService(userCanDeploy, fileExistInCache, remoteCacheEnabled);

        TabsAndActions tabsAndActions = new TabsAndActions();
        tabsAndActions.setRepoType("remote");
        tabsAndActions.setRepoPath("test-remote/item");

        populator.populateForFile(tabsAndActions);

        if (downloadShouldBeAvailable) {
            assertEquals(tabsAndActions.getActions().size(), 1);
            assertEquals(tabsAndActions.getActions().get(0).getName(), "Download");
        } else {
            assertEquals(tabsAndActions.getActions().size(), 0);
        }
    }

    @Test
    public void whenAlibabaEnvironment() {
        try {
            System.setProperty("shared.system.cloudEnv", "alibaba");

            List<TabOrAction> actions = populateActions();

            assertFalse(actions.stream().anyMatch(action -> "View".equals(action.getName())),
                    "view action is not in list");
        } finally {
            System.clearProperty("shared.system.cloudEnv");
        }
    }

    @Test
    public void whenNotAlibabaEnvironment() {
        List<TabOrAction> actions = populateActions();

        assertTrue(actions.stream().anyMatch(action -> "View".equals(action.getName())), "view action is in list");
    }

    @Test(dataProvider = "localAndFederatedPackagesReindex")
    public void testLocalAndFederatedPackagesReindex(PackageType packageType, String repoType,
            boolean shouldHaveReindex, boolean canManage) {
        TabsAndActions tabsAndActions = new TabsAndActions();
        tabsAndActions.setRepoType(repoType);
        tabsAndActions.setRepoPath(String.format("%s-%s/item", packageType.name().toLowerCase(), repoType));
        tabsAndActions.setRepoPkgType(RepoType.fromType(packageType.name()));
        enableFolderDownload(false);
        makeTheUserAnonymousWithReadPermissions();
        when(packageSpecification
                .hasReindex(org.artifactory.common.repo.RepoType.LOCAL, tabsAndActions.getRepoPkgType()))
                .thenReturn(shouldHaveReindex);
        when(authService.canManage(tabsAndActions.getRepoPath())).thenReturn(canManage);
        populator.populateForRepository(tabsAndActions);
        List<TabOrAction> actions = tabsAndActions.getActions();
        if (shouldHaveReindex && canManage) {
            assertTrue(actions.stream().anyMatch(action -> "RecalculateIndex".equals(action.getName())));
        } else {
            assertFalse(actions.stream().anyMatch(action -> "RecalculateIndex".equals(action.getName())));
        }
    }

    @Test
    public void testReleaseBundleV2populateForRepository() {
        TabsAndActions tabsAndActions = new TabsAndActions();
        tabsAndActions.setRepoType(LOCAL_REPO);
        tabsAndActions.setRepoPath(String.format("%s-%s/item", LOCAL_REPO, LOCAL_REPO));
        tabsAndActions.setRepoPkgType(RepoType.ReleaseBundles);
        populator.populateForRepository(tabsAndActions);
        List<TabOrAction> actions = tabsAndActions.getActions();
        assertTrue(actions.stream().anyMatch(action -> "Refresh".equals(action.getName())));
        assertFalse(actions.stream().anyMatch(action -> "Watch".equals(action.getName())));
        assertTrue(actions.stream().anyMatch(action -> "Favorites".equals(action.getName())));
    }

    private List<TabOrAction> populateActions() {
        TabsAndActions tabsAndActions = new TabsAndActions();
        tabsAndActions.setRepoPath("repo/path/PACKAGES");
        tabsAndActions.setRepoPkgType(RepoType.CRAN);
        populator.populateForFile(tabsAndActions);
        return tabsAndActions.getActions();
    }

    private void enableFolderDownload(boolean forAnonymous) {
        FolderDownloadConfigDescriptor folderDescriptor = new FolderDownloadConfigDescriptor();
        folderDescriptor.setEnabled(true);
        folderDescriptor.setEnabledForAnonymous(forAnonymous);
        CentralConfigDescriptor descriptor = Mockito.mock(CentralConfigDescriptor.class);
        when(descriptor.getFolderDownloadConfig()).thenReturn(folderDescriptor);
        when(centralConfig.getDescriptor()).thenReturn(descriptor);
    }

    private void makeTheUserAnonymousWithReadPermissions() {
        when(authService.canRead(any())).thenReturn(true);
        when(authService.isAnonymous()).thenReturn(true);
    }

    private void setupRepoService(boolean userCanDeploy, boolean fileExistInCache, boolean remoteCacheEnabled) {
        RepoPath remotePath = RepoPathFactory.create("test-remote/item");
        RemoteRepoDescriptor remote = new RemoteRepoDescriptor() {
        };
        remote.setKey("test-remote");
        remote.setStoreArtifactsLocally(remoteCacheEnabled);
        when(repoService.remoteRepoDescriptorByKey("test-remote")).thenReturn(remote);
        when(repoService.exists(remotePath)).thenReturn(fileExistInCache);
        when(authService.canDeploy(RepoPathFactory.create("test-remote-cache/item"))).thenReturn(userCanDeploy);
    }

    @DataProvider
    public static Object[][] getOptions() {
        return new Object[][] {
                //userCanDeploy, fileExistInCache, remoteCacheEnabled, downloadShouldBeAvailable
                {false, false, false, true},
                {false, false, true, false},
                {false, true, false, true},
                {false, true, true, true},
                {true, false, false, true},
                {true, false, true, true},
                {true, true, true, true}
        };
    }

    @DataProvider
    public static Object[][] localAndFederatedPackagesReindex() {
        return new Object[][] {
                //packageType, repoType, shouldHaveReindex, canManage
                {PackageType.SWIFT, LOCAL_REPO, true, true},
                {PackageType.SWIFT, FEDERATED_REPO, true, true},
                {PackageType.SWIFT, LOCAL_REPO, true, false},
                {PackageType.SWIFT, FEDERATED_REPO, true, false},
                {PackageType.DOCKER, LOCAL_REPO, false, true},
                {PackageType.DOCKER, FEDERATED_REPO, false, true},
                {PackageType.DOCKER, LOCAL_REPO, false, false},
                {PackageType.DOCKER, FEDERATED_REPO, false, false},
        };

    }
}