package org.artifactory.rest.resource.federation;

import org.apache.http.HttpStatus;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.config.ConfigurationException;
import org.artifactory.federation.FederatedReplaceURLService;
import org.artifactory.util.UrlValidator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;
import org.springframework.web.bind.annotation.RequestBody;
import org.artifactory.model.ReplaceUrlConfig;

import javax.annotation.security.RolesAllowed;
import javax.ws.rs.POST;
import javax.ws.rs.Path;
import javax.ws.rs.core.Response;

import static javax.ws.rs.core.Response.Status.OK;
import static org.apache.commons.lang3.StringUtils.isBlank;

/**
 * @author Daniel simantov
 * Handle baseurl change for federated repositories
 * by updating existing remote members and copying their corresponding
 * queue of the old federated member
 */

@Component
@Scope(BeanDefinition.SCOPE_PROTOTYPE)
@Path("federation")
@RolesAllowed({AuthorizationService.ROLE_ADMIN})
public class FederatedReplaceUrlResource {

    private static final Logger log = LoggerFactory.getLogger(FederatedReplaceUrlResource.class);
    private FederatedReplaceURLService federatedReplaceURLService;

    @Autowired
    public void setFederatedReplaceURLService(FederatedReplaceURLService federatedReplaceURLService) {
        this.federatedReplaceURLService = federatedReplaceURLService;
    }

    /**
     * Handle baseurl change for federated repositories
     * by updating existing remote members and copying their corresponding
     * queue of the old federated member
     */
    @POST
    @Path("replaceUrl")
    @RolesAllowed({AuthorizationService.ROLE_ADMIN})
    public Response replaceUrl(@RequestBody ReplaceUrlConfig replaceURLConfig) {
        try {
            isUrlValid(replaceURLConfig.getNewBaseUrl());
            isUrlValid(replaceURLConfig.getOldBaseUrl());
            log.debug("Received request to replace base url from {} to {}",
                    replaceURLConfig.getOldBaseUrl(), replaceURLConfig.getNewBaseUrl());
            federatedReplaceURLService.replaceUrl(replaceURLConfig.getOldBaseUrl(), replaceURLConfig.getNewBaseUrl());

        } catch (ConfigurationException configurationException) {
            log.error("Replace URL is unavailable for these URLs, Urls incorrect old URL: {}, new URL {}, e: {}",
                    replaceURLConfig.getOldBaseUrl(), replaceURLConfig.getNewBaseUrl(),
                    configurationException.getMessage());
            return Response.status(HttpStatus.SC_BAD_REQUEST)
                    .entity("Replace URL is unavailable for these URLs, Urls incorrect").build();
        } catch (IllegalStateException IllegalStateException) {
            log.error("Replace URL is unavailable for these URLs: old URL:{}, new URL: {}, e: {}",
                    replaceURLConfig.getOldBaseUrl(),replaceURLConfig.getNewBaseUrl(),
                    IllegalStateException.getMessage());
            return Response.status(HttpStatus.SC_BAD_REQUEST)
                    .entity("Replace URL is unavailable for these URL").build();
        }

        return Response.status(OK).entity(getResponse(replaceURLConfig)).build();
    }

    private String getResponse(ReplaceUrlConfig replaceURLConfig) {
        return String.format("Binding URL from URL %s to new %s finished.",
                replaceURLConfig.getOldBaseUrl(), replaceURLConfig.getNewBaseUrl());
    }

    /**
     * Validates URL using UrlValidator.
     *
     * @param urlBase url to validate.
     * @throws ConfigurationException if URL is in illegal form.
     * @see org.artifactory.util.UrlValidator
     */
    private void isUrlValid(String urlBase) {
        if (!isBlank(urlBase)) {
            try {
                new UrlValidator("http", "https").validate(urlBase);
            } catch (UrlValidator.UrlValidationException ex) {
                log.warn("invalid base URL {}", urlBase);
                throw new ConfigurationException("Invalid federated repo URL base", ex);
            }
        }
    }
}

